/**************************************************************************
*                                                                         *
*              National Center for Biotechnology Information              *
*       Bldg. 38A, NIH,  8600 Rockville Pike,  Bethesda, MD  20894        *
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *

Karlin, Samuel and Stephen F. Altschul (1990).  Methods for assessing
the statistical significance of molecular sequence features by using
general scoring schemes.  Proc. Natl. Acad. Sci. USA 87:2264-2268.

Altschul, Stephen F., Warren Gish, Webb Miller, Eugene W. Myers, and
David J. Lipman (1990).  A basic local alignment search tool.
J. Mol. Biol. 215 (in press).
**************************************************************************/
/* (For a better look, set tabstops every 4 columns) */

/* Define the maximum length of the query sequence */
#ifdef BLASTP
#define QUERYLEN_MAX	5000
#define PROC_MAX	8
#define PROC_DEFAULT	8
#define PROC_RESERVE	0
#endif
#ifdef TBLASTN
#define QUERYLEN_MAX	4096
#define PROC_MAX	8
#define PROC_DEFAULT	8
#define PROC_RESERVE	0
#endif
#ifdef BLASTN
#define QUERYLEN_MAX	60000
#define PROC_MAX	8
#define PROC_DEFAULT	8
#define PROC_RESERVE	0
#endif
#ifdef BLASTX
#define QUERYLEN_MAX	10000
#define PROC_MAX	8
#define PROC_DEFAULT	8
#define PROC_RESERVE	0
#endif
#ifdef BLAST3
#define QUERYLEN_MAX	5000
#define PROC_MAX	8
#define PROC_DEFAULT	8
#define PROC_RESERVE	0
#endif

#ifndef QUERYLEN_MAX
#define QUERYLEN_MAX	5000
#define PROC_MAX	1
#define PROC_DEFAULT	1
#endif

#if defined(USE_MPROC) && !defined(MPROC_AVAIL)
#undef USE_MPROC
#endif
#if defined(USE_SHM) && !defined(SYSV_SHM_AVAIL)
#undef USE_SHM
#endif

#ifdef USE_MPROC
#include <ulocks.h>
#include <task.h>
#include <sys/prctl.h>
#endif

#define DFA_USEUNIONS
#include <dfa.h>
#ifdef USE_SHM
#include "shmutil.h"
#endif

/* EXTERN should be defined in the main program module only */
#ifndef EXTERN
#define EXTERN	extern
#else
#ifndef INIT
#define INIT
#endif
#endif

/* define REFINED_STATS for refined calculations of Expects, P-values,
and cutoff scores...
High-scoring alignments found with a short query sequence tend to involve
a large fraction (if not all) of the query sequence.  The expected
length of a high-scoring segment pair significantly reduces the number
of degrees of freedom for producing such high scores with the short
query sequence.  The Expect and Poisson P-values should then be
reduced appropriately.  This is accomplished by subtracting
the expected length (as calculated from the score of the HSP) from the
length of the query sequence before calculating the Expect and P-value.
In a similar manner, the cutoff score used for a search is calculated
from the user's choice (or the default value) for the parameter E.
*/
#define REFINED_STATS

#ifdef OS_UNIX
/* qsort based on the BSD UNIX version provides 2X faster sorting */
#define hsort qsort
#endif

typedef long	SCORE;	/* Datatype for scores and substitution matrix */
#define SCORE_MAX	LONG_MAX
#define SCORE_MIN	LONG_MIN /* Must be a large negative number -- gotcha! */

#ifndef BLASTN
typedef int	DIAG_TYPE;
typedef int	*DIAG_PTR;
#else
/* blastn does things differently, because sequences may be very long. */
/* This structure retains the extents of matches for current sequence */
typedef struct diag_bnd {
		long	diag, pos;
	} DIAG_TYPE;
typedef int	DIAG_PTR;
#define DIAG_MAX	4096
#endif

typedef unsigned char	PACK_TYPE;	/* Compressed nuc. sequence datatype */

#ifdef BLASTX
typedef struct {
		int		frame;
		char	*seq;
		DIAG_TYPE
				*_dlev,			/* extend of discovered matches... */
				*diag_level;	/*    on a given diagonal */
		DIAG_PTR
				diag_stack,		/* list of diagonals with matches */
				diag_top;		/* next diagonal with a match */
		long	len,			/* length of character string that is the seq */
				elen;			/* effective length of the sequence for stats */
		double	xlen;
		SCORE
				maxscore,	/* maximum achievable score */
				highs_exp,	/* expected high score */
				highs_obs,	/* observed high score */
				S,			/* cutoff score for this reading frame */
				X,			/* drop-off score for this reading frame */
				minusX,		/* negative X */
				T;
		double	E,
				K, Lambda, H;
	} seq_data;
#endif

typedef struct hsp {
		struct hsp far	*next;
		Int4	seq_nbr, len, q_pos, s_pos, score;
		union {
			double	P_val;
			double	smallest_P;
		} rawp;
		double	DBP_val;
#if defined(TBLASTN) || defined(BLASTX)
		int		frame;		/* reading frame */
#endif
	} *HSPPtr;


#ifdef USE_MPROC
	/* Multiprocessing context, one for each processor */
typedef struct mpcontext {
		int		id;	/* sub-process id number */
		int		cutoff;
		long
				numhits,
				numhits8,
				numfailed,
				curHSP,
				overcnt,
				totHSP,
				hasHSP,
				current,
				nbr_seqs;
		DIAG_TYPE
				*_dlev,
				*diag_level;
		DIAG_PTR
				diag_stack,
				diag_top;
		char
				*s_seq0,
				*s_seq,
				*d_seq;
		PACK_TYPE	
				*c_seq;
		int	
				s_len,
				d_len,
				c_len,
				frame;
		SCORE	high_score;
		long	*score_hist;
		struct hsp	**HSP;
		HSPPtr	firstHSP;
#ifdef BLASTX
		seq_data	sd[6];
#endif
	} MPContext, *MPContextPtr;
#define MP	mp
#define MPC	mp,
#define MPPROTO	MPContextPtr mpcptr
#define MPPROTOC	MPContextPtr mpcptr,
#define MPDEF	MPContextPtr	mp;
#define MPPTR	mp->
#define M_LOCK	m_lock()
#define M_UNLOCK	m_unlock()
#else /* !USE_MPROC */
/* Dummy definitions */
#define MP
#define MPC
#define MPPROTO
#define MPPROTOC
#define MPDEF
#define MPPTR
#define M_LOCK
#define M_UNLOCK
#endif /* !USE_MPROC */


typedef struct mfile {
		FILE	*fp;
#ifdef USE_SHM
		char far	*sp;
		off_t	offset;
		off_t	filesize;
#endif
	} MFILE;

MFILE far	*mfil_open PROTO((char far *name, char far *mode));
int		mfil_close PROTO((MFILE far *mfp));
char far	*mfil_get PROTO((MFILE far *mfp, size_t len));
size_t	mfil_dup PROTO((MFILE far *mfp, char far *buf, size_t len));
int		mfil_seek PROTO((MFILE far *mfp, long offset, int ptrname));

#ifdef BLAST3
struct record {
		int		score;
		int		scor3;
		int		seq2;
		int		seq3;
		int		start1;
		int		dg2;
		int		dg3;
		int		length;
	};
#define RECORD_MAX	40000
#endif


/* Ambiguous residues are defined using this structure */
typedef struct degen {
		char	residue;	/* ASCII letter for this residue */
		int		ndegen;		/* no. of residues this ASCII letter matches */
		char	*list;		/* null-terminated list of matching letters */
	} Degen, DegenPtr;


/**************************************************************************
*  ALPHASIZE_MAX == max. number of letters permitted to be in the alphabet
**************************************************************************/
#define ALPHASIZE_MAX	32

/************************************************************************
*  ALPHAVAL_MAX == max. permitted numerical value of any letter in the alphabet.
*  For speed, this should be one less than a power of two.
************************************************************************/
#define ALPHAVAL_MAX	31

#define NBPN	2					/* no. of bits per packed nucleotide */

#define TOP_STRAND	1
#define BOTTOM_STRAND	2

#define NT_MAGIC_BYTE	0xfc		/* Magic byte at end of compressed nt db */
#define NSENTINELS	2


int		get_weights PROTO((char *mname, int badletter, int *xltab,
							int *asize, int *minchar, int *maxchar,
							signed char achars[ALPHASIZE_MAX+1],
							signed char order[ALPHAVAL_MAX+1][ALPHASIZE_MAX],
							SCORE *lows, SCORE *highs,
							SCORE maxcost[ALPHAVAL_MAX+1],
							SCORE matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1]));

int		karlin PROTO((SCORE low, SCORE high, double *prob, double *lambda,
					double *K, double *H));
double	stoe PROTO((SCORE S, double Lambda, double K, double H,
					unsigned long dblen, unsigned long queryeff));
SCORE	etos PROTO((double E, double Lambda, double K, double H,
					unsigned long dblen, unsigned long queryeff));
SCORE	ehighs PROTO((double Lambda, double K, double H,
					unsigned long dblen, unsigned long querylen));

void	cutoffs PROTO((SCORE *S, double *E, SCORE *highsexp, double K, double Lambda, double H, Uint4 Neff, Uint4 Meff));

DFA		*bld_table_aa PROTO((char *qseq, int W, SCORE T, int asize,
						SCORE (*matrix)[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1],
						SCORE *maxcost,
						signed char (*suborder)[ALPHAVAL_MAX+1][ALPHASIZE_MAX],
						Uint4 *numwords, Uint4 *numexact
						));

void	print_hist PROTO((SCORE score_hist_dim, long *score_hist,
							SCORE S, SCORE T,
							double lambda, double K, double H,
							long qlen, long dblen, long nbr_seqs));

   /* Functions that detect (and perhaps reject) overlapping HSPs */
void	ckoverlap PROTO((MPPROTOC HSPPtr hp));	/* Deal with overlapping HSPs */
Boolean	ovlap_a PROTO((struct hsp far *hp1, struct hsp far *hp2));
Boolean	ovlap_n PROTO((struct hsp far *hp1, struct hsp far *hp2));
Boolean	ovlap_t PROTO((struct hsp far *hp1, struct hsp far *hp2));
Boolean	ovlap_x PROTO((struct hsp far *hp1, struct hsp far *hp2));
Boolean	ovlap_0 PROTO((struct hsp far *hp1, struct hsp far *hp2));
EXTERN Boolean	(far *ovlap) PROTO((struct hsp far *hp1, struct hsp far *hp2))
#ifdef INIT
#if defined(BLASTP) || defined(BLAST3) || defined(BLASTALL)
	= ovlap_a
#endif
#if defined(BLASTN)
	= ovlap_n
#endif
#if defined(BLASTX)
	= ovlap_x
#endif
#if defined(TBLASTN)
	= ovlap_t
#endif
#endif
	;


void	ghspsave PROTO((MPPROTOC HSPPtr hp));	/* General alloc. and save of an HSP */

int		smaller PROTO((HSPPtr *pki, HSPPtr *pkj));

FILE	*ckopen PROTO((char *fname, char *mode));
char far	*ckalloc PROTO((size_t amt)),	/* storage allocator */
			*ckalloc0 PROTO((size_t amt)),	/*   with zero initialization */
			*ckrealloc PROTO((char far *p, size_t newsize));

Signal_t	sigterm PROTO((int sig)), sigalrm PROTO(());

EXTERN char far	*module;	/* Program name */

EXTERN long	cktime	/* Poll the expended execution time this often (seconds) */
#if defined(CKTIME) && defined(INIT)
	= CKTIME
#endif
	;

EXTERN long	maxtime	/* Maximum permissible execution time (seconds) */
#if defined(MAXTIME) && defined(INIT)
	= MAXTIME
#endif
	;

#define HSP_MAX	500			/* Max. # of HSPs to record (usually per db seq) */

EXTERN long		cutoff;		/* Cutoff score for reporting HSPs */
EXTERN long		overcnt;	/* No. of overlapping HSPs discarded */
EXTERN long		numhits;	/* No. of word hits against database */
EXTERN long		numfailed;	/* No. of failed hit extensions */
EXTERN long		curHSP;		/* No. of HSPs saved with specific db sequence */
EXTERN unsigned long	maxHSP
#ifdef INIT
	= HSP_MAX
#endif
	;
EXTERN long		totHSP;		/* Total no. of HSPs found against the database */
EXTERN long		hasHSP;		/* No. of db sequences with at least one HSP */
#ifdef USE_MPROC
EXTERN int		chunksize;	/* No. of sequences to allocate to a process */
#endif

EXTERN struct hsp far	**HSP;

EXTERN void	(far *hspsave)()
#ifdef INIT
	= ghspsave
#endif
;


EXTERN size_t	hspsize
#ifdef INIT
	= sizeof(struct hsp)
#endif
	;

#define DB_TYPE_PRO	0x78857a4f	/* Magic i.d. for a protein sequence database */
#define DB_TYPE_NUC	0x788325f8	/* Magic i.d. for a nt. sequence database */

#define AAFORMAT	3	/* Latest a.a. database format ID number */
#define NTFORMAT	6	/* Latest nt. database format ID number */

