/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
BLAST3 - Search a protein database for regions that approximately match
regions of a query sequence.  The basic command syntax is

 	BLAST3 database query_sequence

 where:
 	database names a file containing protein sequences, separated by
 		header lines beginning with the character '>'.  A typical
 		sequence entry is:

		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
		RADLIAYLKKATNE

 	query_sequence names a file containing a protein sequence (where the
 		header line is optional).
 
 The BLAST3 command permits optional additional arguments, such as X=50,
 that reset certain internal parameters.  The available parameters are:


Parameters controlling 3-way alignment construction:

	F or f is the highest Expect-value of a reported three-way alignment
		(default is 5.)
	L or l is the lowest score permitted in reporting three-way alignments
		(if specified, it overrides F parameter)
	R or r is the maximum Ratio of the Expect-value for a 3-way alignment
		to the smallest Expect-value of its component pairwise alignments
		(the default value is 2, but the user may very well want to set
		this parameter to 1, 0.5, OR LESS, to avoid what might be copious
		"garbage" output)
	N or n is the highest pairwise Expect-value to use when constructing
		3-way alignments (default is 0.1).  In essence, this defines
		which pairwise alignments are insignificant and should be used
		to find significant 3-way alignments.
	U or u is the highest pairwise score to use when constructing 3-way
		alignments (if specified, it overrides P parameter)


Parameters controlling pairwise alignment construction (initial BLAST search):

	E or e is the estimated number of random HSPs to collect during the
		initial BLAST portion of the search (default is 1000).  This is
		close to the number of diagonals remembered for construction
		of three-way alignment.
	S or s gives the High-scoring Segment Pair (HSP) cutoff score for the
		initial BLAST portion of the search (default depends on E parameter)
	T or t gives the threshold for a word match during the initial BLAST
		portion of the search (default is 13 or 12)
	X or x gives the value for terminating word hit extensions during the
		initial BLAST portion of the search (default is 20)


Parameters controlling both pairwise and three-way alignment construction:

	M or m is the name of a file containing substitution costs
		(default is "PAM120")
	Y or y is the effective query sequence length for estimating statistical
		significance (default is the actual length of the query sequence)
	Z or z is the effective database length for estimating statistical
		significance (default is the actual length of the database)


Miscellaneous parameters:

	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer (i.e., SGI).
	B or b if set to 0, suppresses the reporting of the pairwise
		alignments from the initial BLAST portion of the search.
	D or d if made non-zero, turns on the concise reporting of diagonals
		saved from the initial BLAST portion of the search.
	O or o if set to 0, suppresses the output that normally is sent to stderr
		for monitoring the program's progress.

 Thus a typical command is

	BLAST3 AABANK SEQ S=50 M=PAM250


Relevant literature:

	Karlin & Altschul, "Methods for assessing the statistical significance
	    of molecular sequence features by using general scoring schemes,"
	    Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
	Altschul and  Lipman, "Protein database searches for multiple alignments,"
	    Proc. Natl. Acad. Sci. USA 87 (1990) 5509-5513;
	Altschul, Gish, Miller, Myers and  Lipman, "Basic local alignment search
	    tool," J. Mol. Biol. 214 (1990) in press.

   (Note:  set tabstops every 4 columns for a better look).
*/
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>

#define BLAST3
#define EXTERN
#define CKTIME	20
#define MAXTIME	1800
#include "blast.h"
#include "aabet.h" /* the amino acid alphabet */

#define HEADERLEN_MAX	80	/* maximum length of header line */
#define WORDSIZE_MAX	7	/* maximum length of critical word */

#define DEFAULT_W	4	/* default word size */
#define DEFAULT_ehigh	0.1
#define DEFAULT_elow3	5.
#define DEFAULT_oflag	1
#define DEFAULT_maxdg	1000
#define DEFAULT_Efactor	2.
#define DEFAULT_High	0
#define DEFAULT_showblast	1
#define DEFAULT_showdiags	0
#define DEFAULT_S		0
#define DEFAULT_Low3	0
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T		13	/* default threshold for word hits */
#define DEFAULT_X		20	/* default constant for hit extensions */

char
	header[HEADERLEN_MAX],
	*qname = ">Query",
	qtitle[73],
	*dbtitle,
#ifdef USE_SHM
	*dbbuf,		/* Memory-resident database */
#else
	*seqs, /* Large buffer in which to store all database sequences */
#endif
	*s_seq0, *s_seq, *ss_seq,
	_q_seq[QUERYLEN_MAX+3],
	*q_seq = &_q_seq[1],
	qq_seq[QUERYLEN_MAX+WORDSIZE_MAX],
	M[FILENAME_MAX] = DEFAULT_M; /* name of file containing substitution scores */

unsigned long	q_len;		/* length of query sequence */
int		s_len = -1;	/* length of database sequence */
int		maxnamelen; /* length of longest name of sequence */

long	totdblen;	/* total no. of residues in the database */

unsigned long
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

int
	W = DEFAULT_W,			/* word size */
	W_set,
	S = DEFAULT_S,			/* cutoff for HSPs */
	S_set,
	T = DEFAULT_T,			/* threshold for word hits */
	T_set,
	X = DEFAULT_X,			/* constant for hit extensions */
	X_set,
	minusX = -DEFAULT_X;	/* -X */

double fq[21] = {0, .081, .057, .045, .054, .015, .039, .061, .068, .022, .057,
 		  .093, .056, .025, .040, .049, .068, .058, .013, .032, .067};
double	*prob;
double	fq1[32];
struct record	firstrec[RECORD_MAX], *rec = firstrec;
char	**pts;
char	**names;
long	*lens;
double	Lambda, Lambda3, K, K3, H, H3,
		ehigh = DEFAULT_ehigh,
		elow3 = DEFAULT_elow3,
		Efactor = DEFAULT_Efactor;

unsigned long
		Meff,	/* effective query sequence length */
		Neff;	/* effective database length */

int		oflag = DEFAULT_oflag,
		maxdg = DEFAULT_maxdg,
		showblast = DEFAULT_showblast,
		showdiags = DEFAULT_showdiags,
		nrec;

SCORE
		High = DEFAULT_High,
		Low3 = DEFAULT_Low3,
		lows,
		highs;

long	*score_hist;
SCORE	high_score, score_hist_dim;

/* data structure to hold locations of matches for current database sequence */

DIAG_TYPE
	*_dlev,			/* extent of discovered matches .. */
	*diag_level;	/*     on a given diagonal */

int
	*diag_stack,	/* list of diagonals with matches */
	*diag_top;		/* next diagonal with a match */

HSPPtr firstHSP = NULL;
HSPPtr *HSPlist;

DFA		*dp;	/* the automaton */

static void	initdb PROTO((char far *dbname));
static void	search_aa PROTO((MPPROTO));
static void	get_pvals PROTO((MPPROTO));
static void	print_seq PROTO((HSPPtr hp));
static int	get_seq PROTO((MPPROTO));
static void	dosearch PROTO(());
static void	report();

#ifdef USE_MPROC
MPContext	mpc[PROC_MAX];
int		maxprocs, maxpprocs, numprocs = PROC_DEFAULT;
#else
int		numprocs = 1;
#endif

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3, t4;

int		alpha;				/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static signed char
	/* array of amino acid names */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

SCORE
	maxcost[ALPHAVAL_MAX+1],	/* max cost of substituting for a character */
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];


/* UNIX-specific input routine; database must be pre-processed by SETDB */
static FILE
	*sfile;		/* file pointer for sequences */
static MFILE
	*hfile;		/* file pointer for headers */

static int
	current;		/* number of current sequence */

static off_t
	nbr_seqs,		/* number of database sequences */
	*seq_beg, 		/* starts of sequences */
	*header_beg;		/* starts of headers */

static float star,inc;

#ifndef USE_MPROC
char	*version = "1.1.5 [6-Dec-90]";
#else
char	*version = "1.1.5MP [6-Dec-90]";
#endif


main(argc, argv)
	int		argc;
	char	**argv;
{
	char	*a, o;
	register int	i, j, k;
	register char	*cp;

	sys_times(&t0);

	banner(argv[0], version);

	if (argc < 3)
		fatal("Usage:  %s database query_sequence [options]", module);

	if (!isatty(2))
		oflag = 0;

	while (--argc > 2) {
		if (argv[argc][1] == '=')
			o = argv[argc][0];
		else
			if (argv[argc][0] == '-')
				o = argv[argc][1];
			else
				fatal("argument %d has improper form", argc+1);
		a = argv[argc] + 2;
		o = TOUPPER(o);
		switch (o) {
/* Don't allow user to mess with W -- validity has not been verified recently*/
/***********************************
			case 'W':
				if (sscanf(a, "%d", &W) != 1)
					fatal("Bad value for W.");
				if (W <= 0)
					fatal("Invalid value for W.");
				if (W > WORDSIZE_MAX)
					fatal("W is too big--must be less than %d.", WORDSIZE_MAX+1);
				W_set = 1;
				break;
************************************/
			case 'P':
#ifdef USE_MPROC
				numprocs = atoi(a);
				if (numprocs <= 0)
					fatal("Bad value for P (the number of processors)");
				if (numprocs > PROC_MAX)
					fatal("Too many processors requested (Pmax = %d)", PROC_MAX);
#else
				printf("NOTE:  Parameter P is invalid--\n       this program has not been compiled to use multiprocessing\n");
#endif
				break;
			case 'S':
				if (sscanf(a, "%d", &S) != 1 || S < 1)
					fatal("Bad value for S.");
				S_set = 1;
				break;
			case 'M':
				if (strlen(a) >= FILENAME_MAX)
					fatal("File name is too long.");
				strcpy(M,a);
				break;
			case 'T':
				if (sscanf(a, "%d", &T) != 1 || T < 1)
					fatal("Bad value for T.");
				T_set = 1;
				break;
			case 'X':
				if (sscanf(a, "%d", &X) != 1 || X < 1)
					fatal("Bad value for X.");
				minusX = -X;
				X_set = 1;
				break;
			case 'E':
				if (sscanf(a, "%d", &maxdg) != 1 || maxdg < 1)
					fatal("Bad value for E.");
		    	break;
			case 'U':
				High=atoi(a);
		    	break;
			case 'R':
				if (sscanf(a, "%lf", &Efactor) != 1 || Efactor <= 0.)
					fatal("Bad value for R.");
				break;
			case 'N':
				if (sscanf(a, "%lf", &ehigh) != 1 || ehigh <= 0.)
					fatal("Bad value for N.");
		    	break;
			case 'L':
				if (sscanf(a, "%d", &Low3) != 1 || Low3 < 1)
					fatal("Bad value for L.");
		    	break;
			case 'F':
				if (sscanf(a, "%lf", &elow3) != 1 || elow3 <= 0.)
					fatal("Bad value for F.");
		    	break;
			case 'Y':
				if (sscanf(a, "%d", &Meff) != 1 || Meff == 0)
					fatal("Bad value for Y.");
		    	break;
			case 'Z':
				if (sscanf(a, "%d", &Neff) != 1 || Neff == 0)
					fatal("Bad value for Z.");
		    	break;
			case 'B':
				showblast = atoi(a);
				break;
			case 'D':
				showdiags = atoi(a);
				break;
			case 'O':
				oflag=atoi(a);
			   	break;
			default:
				fatal("Bad option.");
		}
	}

	HSP = (struct hsp **)ckalloc(sizeof(struct hsp)*maxHSP);

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alpha, &minchar, &maxchar, achars,
					order, &lows, &highs, maxcost, matrx);

	get_query(argv[2]);

#ifdef USE_MPROC
	(void) prctl(PR_SETEXITSIG, SIGTERM);

	maxprocs = prctl(PR_MAXPROCS);/* max. processes per user */
	maxpprocs = prctl(PR_MAXPPROCS);/* max. parallel processes */
	numprocs = MIN(numprocs, PROC_MAX);
	numprocs = MAX(1, numprocs);
	numprocs = MIN(numprocs, maxprocs);
	numprocs = MIN(numprocs, maxpprocs - PROC_RESERVE);
	numprocs = MAX(1, numprocs);
#endif

#ifdef CKTIME
	maxtime /= numprocs;
	/* Poll the cpu time that has been consumed every CKTIME seconds */
	(void) signal(SIGALRM, sigalrm);
	(void) alarm(cktime);
#endif

	initdb(argv[1]); /* Open database files */
	(void) fflush(stdout);

	pts = (char **)ckalloc(nbr_seqs*sizeof(*pts));
	names = (char **)ckalloc(nbr_seqs*sizeof(*names));
	lens = (long *)ckalloc(nbr_seqs*sizeof(*lens));

	prob = (double *)ckalloc((3*highs-3*lows+1) * sizeof(*prob));
	for (i=0; i<DIM1(fq1); ++i)
		fq1[i] = 0.0;
	for (i=0; i<q_len; ++i)
		fq1[q_seq[i]] += 1.0;
	for (i=0; i<DIM1(fq1); ++i)
		fq1[i] /= q_len;
	for (i=highs-lows; i>=0; --i)
		prob[i] = 0.0;
	for (i=minchar; i<=maxchar; ++i)
		for (j=1;j<DIM1(fq);++j)
			prob[matrx[i][j]-lows] += fq1[i]*fq[j];
	karlin(lows, highs, prob, &Lambda, &K, &H);
	if (S == 0) {
		S = etos((double)maxdg, Lambda, K, H, Neff, Meff);
		if (S < 24) {
			printf("Calculated cutoff score of %d is too low.  Using 24 instead.\n",
					S);
			S = 24;
		}
	}

	score_hist_dim = etos(0.01, Lambda, K, H, Neff, Meff) + 1;
	score_hist = (long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));

	X = MIN(X, S);
	minusX = -X;


	/*
	The following adjustment to T is tightly linked
	to the PAM120 substitution matrix.
	*/
	if (!T_set && S < 35)
		--T;

	lows *= 3;
	highs *= 3;
	for (i=highs-lows;i>=0;--i)
		prob[i] = 0.0;
	for (i=minchar;i<=maxchar;++i)
		for (j=1;j<DIM1(fq);++j)
			for (k=1;k<DIM1(fq);++k)
				prob[matrx[i][j]+matrx[i][k]+matrx[j][k]-lows] +=
						fq1[i]*fq[j]*fq[k];
	karlin(lows, highs, prob, &Lambda3, &K3, &H3);
	free((char *)prob);

	sys_times(&t1);
	bld_table();		/* build table of critical words */
	if (dfa_close(dp) != 0)
		fatal("error in dfa_close");
	sys_times(&t2);

	if (oflag) {
		fflush(stdout);
		fprintf(stderr, "Lowest segment score remembered: %d.\n", S);
		fprintf(stderr, "....1....2....3....4....5....6....7....8....9....0\n");
		star = inc = (nbr_seqs-1)/50.;
	}

	maxnamelen = strlen(qname);

#ifdef USE_MPROC
	for (i=0; i<numprocs; ++i) {
		hspinit(&mpc[i]);
		initmpc(i, &mpc[i]);
	}
	m_set_procs(numprocs);
	m_fork(dosearch);
	m_kill_procs();
#else
	hspinit();
	dosearch();
#endif

	nrec = rec - firstrec;

	/* Retrieve header lines for matching sequences */
	for (i=0; i<nbr_seqs; ++i) {
		if (pts[i] == 0)
			continue;
		/* get the header line */
		if (mfil_seek(hfile, header_beg[i], 0) == -1)
			fatal("fseek error on header file");
		if ((j = mfil_dup(hfile, header, sizeof(header)-1)) <= 0)
			fatal("error reading header");
		header[j] = NULLB;
		cp = header;
		while (*cp != NULLB && !isspace(*cp) && *cp != '(')
			++cp;
		j = cp - header;
		if (j > maxnamelen)
			maxnamelen = j;
		names[i] = cp = ckalloc(j+1);
		cp[j] = NULLB;
		while (j-- > 0)
			cp[j] = header[j];
	}

	dfanstates = dfa_nstates(dp);
	dfastatesize = dfa_statesize(dp);
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp); /* Deallocate the DFA */

#ifdef USE_MPROC
	{
		register int	i, j;
		register struct hsp	*hp;

		for (i=0; i<numprocs; ++i) {
			if (mpc[i].firstHSP != NULL) {
				if (firstHSP == NULL)
					firstHSP = mpc[i].firstHSP;
				else {
					/* Find the tail of the linked list */
					hp = mpc[i].firstHSP;
					while (hp->next != NULL)
						hp = hp->next;
					hp->next = firstHSP;
					firstHSP = mpc[i].firstHSP;
				}
			}
			hasHSP += mpc[i].hasHSP;
			totHSP += mpc[i].totHSP;
			overcnt += mpc[i].overcnt;
			numhits += mpc[i].numhits;
			numfailed += mpc[i].numfailed;
			if (i > 0)
				for (j=0; j <= score_hist_dim; ++j)
					score_hist[j] += mpc[i].score_hist[j];
		}
	}
#endif

	if (showblast) {
		print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
						Meff, Neff, nbr_seqs);
		print_HSPs();
	}
	putchar('\n');

	sys_times(&t3);
	out3();
	alarm(0);
	sys_times(&t4);

	print_parms();

	print_stats();
}

print_parms()
{
	printf("\nParameters:\n");
	printf("  W = %d, S = %d, T = %d, X = %d, M = %s", W, S, T, X, M);
}

print_stats()
{
	printf("\nStatistics:\n");
	printf("  # of residues in query = %s", lu2s(q_len));
	if (Meff != q_len)
		printf("  (Y = %s)\n", lu2s(Meff));
	else
		putchar('\n');
	printf("  # of neighborhood words = %s\n", ld2s(numwords));
	printf("  # of residues in db = %s", ld2s(totdblen));
	if (Neff != totdblen)
		printf("  (Z = %s)\n", lu2s(Neff));
	else
		putchar('\n');
	printf("  # of word hits against db = %s\n", ld2s(numhits));
	printf("  # of failed hit extensions = %s\n", ld2s(numfailed));
	printf("  # of successful extensions = %s\n", ld2s(totHSP+overcnt));
	printf("  # of HSPs from BLAST run = %s\n", ld2s(totHSP));
	printf("  # of overlapping HSPs discarded = %s\n", ld2s(overcnt));
	printf("  # of sequences in db = %s\n", ld2s(nbr_seqs));
	printf("  # of db sequences with at least one match = %s\n",
		ld2s(hasHSP));

	printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfastatesize*dfanstates, KBYTE)));
	printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	printf("Time to build DFA:  ");
	sys_showtime(&t1, &t2);
#ifdef USE_MPROC
	printf("No. of processors used:  %d\n", numprocs);
#endif
	printf("Time to search database:   ");
	sys_showtime(&t2, &t3);
	printf("Time to find 3-alignments: ");
	sys_showtime(&t3, &t4);
	printf("Total cpu time:            ");
	sys_showtime((Cputime_t *)NULL, &t4);
	exit(0);
}


static void
dosearch()
{
	int		id = 0;
	int		i, j, k;
#ifdef USE_MPROC
	MPDEF

	id = m_get_myid();
	mp = &mpc[id];
	if (id == 0)
		signal(SIGTERM, sigterm);
#endif

	while (get_seq(MP) == 0) {
		if (id == 0 && oflag && MPPTR current>star) {
			fputc('*', stderr);
			fflush(stderr);
			star += inc;
		}
		MPPTR high_score = 0;
		MPPTR curHSP = 0;
		search_aa(MP);	/* search a database sequence */

		if (MPPTR high_score <= score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];

		if (MPPTR curHSP > 0) {
			i = 0;
			k = INT_MIN;
			M_LOCK;
			do {
				j = MPPTR HSP[i]->s_pos - MPPTR HSP[i]->q_pos;
				if (j == k)	/* Remember only one HSP from a given diagonal */
					continue;
				if (rec >= &firstrec[DIM1(firstrec)])
					fatal("Out of storage for diagonals; decrease E parameter,\nor increase RECORD_MAX and recompile");
				rec->dg2 = k = j;
				rec->score = MPPTR HSP[i]->score;
				rec->start1 = MPPTR HSP[i]->q_pos;
				rec->length = MPPTR HSP[i]->len;
				rec->seq2 = MPPTR current;
				rec->seq3 = -1;
				++rec;
			} while (++i < MPPTR curHSP);
			M_UNLOCK;
			pts[MPPTR current] = MPPTR s_seq;
			lens[MPPTR current] = MPPTR s_len;
			get_pvals(MP);
		}
	}
}

#ifdef USE_MPROC
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
	int		id;
	MPContextPtr	mp;
{
	int		len;
	register int	i;

	mp->id = id;

	mp->current = mp->nbr_seqs = -1;

	len = maxlen + q_len + 1;
	mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	mp->diag_level = mp->_dlev + q_len;
	mp->diag_top = mp->diag_stack = (int *)ckalloc(sizeof(*diag_top)*len);
	mp->s_seq0 = ckalloc(maxlen+2);
	*mp->s_seq0++ = NULLB;
	mp->s_seq = mp->s_seq0;
	if (id == 0)
		mp->score_hist = score_hist;
	else
		mp->score_hist = (long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
}
#endif

/* get_query - read the query sequence */
get_query(file_name)
	char	*file_name;
{
	char	*fgets();
	register FILE	*qp;
	register char *p, *q, ch;
		
	qp = ckopen(file_name, "r");
	for (q = q_seq; ; ) {
		p = q;
		if (fgets(p, QUERYLEN_MAX + 2 - (p - q_seq), qp) == NULL) {
			fclose(qp);
			break;
		}
		if (*q == '>')	/* Optional header/title line */
			if (q == q_seq) {
				int		i;
				i = str_len(q);
				if (i > sizeof(qtitle)-1) {
					mem_cpy(qtitle, q, sizeof(qtitle)-4);
					str_cpy(qtitle+sizeof(qtitle)-4, "...");
				}
				else {
					str_cpy(qtitle, q);
					p = str_chr(qtitle, '\n');
					if (p != NULL)
						*p = NULLB;
				}
				p = str_chr(q, ' ');
				if (p == NULL)
					p = str_chr(q, '\t');
				if (p == NULL)
					p = str_chr(q, '(');
				if (p != NULL) {
					*p = NULLB;
					qname = str_dup(q);
				}
				continue;
			}
			else
				break;
		while ((ch = *p++) != '\0') {
			if (aa_atob[ch] < AAID_MAX+1) {
				*q++ = ch;
			}
			else
				if (!isspace(ch))
					if (isprint(ch))
						fatal("Illegal character in query sequence:  ``%c''", ch);
					else
						fatal("Illegal character in query sequence:  #%d", ch);
		}
		if (q - q_seq > QUERYLEN_MAX)
			fatal("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
	*q = '\0';
	q_len = q - q_seq;
	if (q_len < W)
		fatal("query sequence is shorter than %d, the Word length.", W);
	if (qtitle[0] != NULLB)
		(void) printf("Query:  %s\n", qtitle);

	if (Meff == 0)
		Meff = q_len;

	if (str_cnt(q_seq, "ACGTacgt")/(double)q_len > 0.85) {
		printf("\nNote:  the query sequence appears to be nucleic acid when protein is expected.\n\n");
		fflush(stdout);
	}

	q = q_seq;
	while ((ch = *q) != NULLB)
		*q++ = aa_atob[ch];
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	MFILE	*tfile;
	Int4	dbtype, dbformat, i;
	char	*dirname = ".";
	char	database[FILENAME_MAX];
	char	name[FILENAME_MAX];
	int		many, len;
	int		once=0;
	off_t	dbsize;
#ifdef USE_SHM
	int		dbexists;
#endif

	current = -1;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(database, "%s/%s", dirname, dbname);
	else
		STRNCPY(database, dbname, sizeof(database));
	(void) sprintf(name, "%s.ahd", database);
	hfile = mfil_open(name, "r");
	if (hfile == NULL) {
		if ((dirname = getenv("BLASTP")) == NULL)
			dirname = BLASTP_DIR;
		if (++once == 1)
			goto RETRY;
		fatal("Couldn't open database file ``%s''", name);
	}

	(void) sprintf(name, "%s.atb", database);
	tfile = mfil_open(name, "r");
	if (mfil_dup(tfile, (char far *)&dbtype, sizeof(dbtype)) != sizeof(dbtype))
		fatal("Database read error on file", name);
	if (dbtype != DB_TYPE_PRO)
		fatal("``%s'' is not a protein sequence database", dbname);
	if (mfil_dup(tfile, (char *)&dbformat, sizeof(dbformat)) != sizeof(dbformat)
		|| dbformat != AAFORMAT)
			fatal("Database format is an incompatible version; re-run ``setdb''");
 
	if (mfil_dup(tfile, (char far *)&i, sizeof(i)) != sizeof(i)
		|| (dbtitle = mfil_get(tfile, i)) == NULL)              
			fatal("Corrupted database file:  %s", name);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, 1);
	}                           

	if (mfil_dup(tfile, (char *)&nbr_seqs, sizeof(nbr_seqs)) != sizeof(nbr_seqs)
		|| mfil_dup(tfile, (char *)&totdblen, sizeof(totdblen)) != sizeof(totdblen)
		|| mfil_dup(tfile, (char *)&maxlen, sizeof(maxlen)) != sizeof(maxlen))
			fatal("corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) printf("Database:  %s\n", dbtitle);
	(void) printf("           %s sequences; %s total residues.\n",
			ld2s(nbr_seqs), ld2s(totdblen));                       
 
	ss_seq = ckalloc(sizeof(*ss_seq)*(maxlen+1));

	many = nbr_seqs + 1;
	seq_beg = (off_t far *)mfil_get(tfile, sizeof(*seq_beg)*many);
	if (seq_beg == NULL)
		fatal("improper sequence offsets file");
	header_beg = (off_t *)mfil_get(tfile, sizeof(*header_beg)*many);
	if (header_beg == NULL)
		fatal("improper header offsets file");

	(void) sprintf(name, "%s.bsq", database);
#ifdef USE_SHM
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	if ((dbbuf = shm_attfile(name, "r")) == NULL)
#endif
	{
		sfile = fopen(name, "r");
		if (sfile == NULL)
			fatal("Couldn't open database file %s", name);
		dbsize = sys_filesize(name);
		s_seq0 = ckalloc(sizeof(*s_seq)*dbsize);
		(void) printf("Reading database into memory...\n");
		(void) fflush(stdout);
		if (fread((char *)s_seq0, (int)dbsize, 1, sfile) != 1)
			fatal("Database read error.");
	}


	len = maxlen + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;
	diag_top = diag_stack = (int *)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

#ifdef USE_MPROC
	if (numprocs > 1) {
		chunksize = nbr_seqs / (numprocs*50);
		if (numprocs < 4)
			chunksize /= 2;
		chunksize = MAX(chunksize, 1);
	}
	else
		chunksize = nbr_seqs;
#endif
}

/* get_seq - read a database sequence and header line, return 0 on EOF */
static int
get_seq(MP)
	MPDEF
{
	if (++MPPTR current >= MPPTR nbr_seqs)
#ifndef USE_MPROC
		return 1;
#else
	{	/* Allocate a chunk of sequences for this process to scan */
		MPPTR current = m_next() * chunksize;
		if (MPPTR current >= nbr_seqs)
			return 1;
		MPPTR nbr_seqs = MPPTR current + chunksize;
		MPPTR nbr_seqs = MIN(MPPTR nbr_seqs, nbr_seqs);
	}
#endif

#ifdef USE_SHM
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[MPPTR current+1] - seq_beg[MPPTR current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[MPPTR current];
	}
	else
#endif
	{
		MPPTR s_seq = s_seq0 + seq_beg[MPPTR current];
		MPPTR s_len = seq_beg[MPPTR current+1] - seq_beg[MPPTR current] - 1;
	}
	if (MPPTR s_len > maxlen)
		fatal("Database sequence longer than expected maxlen!");
	return 0;
}

/* -----------------------   build table of critical words   -----------------*/

signed char
	wword[WORDSIZE_MAX];	/* candidate for critical word */

SCORE
	best[WORDSIZE_MAX];	/* best[i] = smallest score of wword[0]..wword[i]
				   that can be extended to a critical word */
char	*query;			/* pointer into query sequence */

static
bld_table()
{
	register int i;
	register int	wm1 = W-1;

	dp = dfa_open(NULL, 0, (LETTER *)AAID_MIN, AAID_MAX-AAID_MIN+1, WORDSIZE_MAX,
				ckalloc, 100*KBYTE, free, (DFA_HALTNONALPHAS|DFA_NOXLTAB));
	if (dp == NULL) {
		dfa_perror(dp, module);
		fatal("bld_table");
	}
	for (query = q_seq; query[wm1] != NULLB; query++) {
		best[wm1] = T;
		for (i = wm1; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);
	}
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for wword[0]..wword[i-1] */
enumerate(i, score)
	register int		i;
	register int	score;
{
	/* possible letters by decreasing weight */
	register signed char *o = order[query[i]];
	register int c;
	register int	new_score;

	for (c = 0; c < alpha; ++c) {
		new_score = score + matrx[query[i]][wword[i] = o[c]];
		if (new_score < best[i])
			break;
		if (i < W-1)
			enumerate(i+1, new_score);
		else {
			++numwords;
			if (dfa_add(dp, (LETTER *)wword, W, (mempint)(query + W)) == NULL)
				fatal("error in dfa_add");
		}

	}
}

/*****************************/
#include "searcha.inc"
/*****************************/

int
hsp_score_cmp(h1, h2)
	HSPPtr	*h1, *h2;
{
	return (*h2)->score - (*h1)->score;
}

static void
get_pvals(MP)
	MPDEF
{
	register int i, j;
	register HSPPtr hp;
	register double Q, R, T;

	MPPTR totHSP += MPPTR curHSP;
	++MPPTR hasHSP;
	/* Sort HSPs by score, highest score to lowest */
	if (MPPTR curHSP > 1)
		(void) hsort((char far *)MPPTR HSP, MPPTR curHSP, sizeof(MPPTR HSP[0]), hsp_score_cmp);

	for (i = 0; i < MPPTR curHSP; ++i) {
		hp = MPPTR HSP[i];
		Q = stoe(hp->score, Lambda, K, MPPTR s_len, Meff);
		T = R = 1.0;
		for (j = 1; j <= i; ++j) {
			R *= (Q/j);
			T += R;
		}
		if (Q > 0.01)	/* Is limit approximation applicable? */
			hp->rawp.P_val = 1.0 - exp(-Q)*T;
		else
			hp->rawp.P_val = R * (Q/j);
	}

	/* add HSPs for this sequence to the cumulative list */
	MPPTR HSP[MPPTR curHSP-1]->next = MPPTR firstHSP;
	for (i = MPPTR curHSP-2; i >= 0; --i)
		MPPTR HSP[i]->next = MPPTR HSP[i+1];
	MPPTR firstHSP = MPPTR HSP[0];

	MPPTR curHSP = 0;
}

print_HSPs()
{
	register HSPPtr kp1, kp2;
	register int	i;
	register double	R, minp, minp2, phigh;
	double	xslen;

	if (hasHSP < maxHSP)
		HSPlist = HSP;
	else
		HSPlist = (HSPPtr *)ckalloc(hasHSP*sizeof(HSPPtr));
	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
				kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		firstHSP = kp2;
	}

	/*
		For each sequence, calculate the P_val adjusted for the total
		length of the database (DBP_val); then find the smallest DBP_val
		of its HSPs.
	*/
	for (i = 0; i < hasHSP; ++i) {
		kp1 = HSPlist[i];
		xslen = seq_beg[kp1->seq_nbr+1] - seq_beg[kp1->seq_nbr];
		R = Neff/xslen;
		minp = R * kp1->rawp.P_val;
		if (minp > 0.01) /* Is limit approximation applicable? */
			minp = (double)1.0 - exp(-minp);
		kp1->DBP_val = minp;
		for (kp2 = kp1->next; kp2 != NULL; kp2 = kp2->next) {
			minp2 = R * kp2->rawp.P_val;
			if (minp2 > 0.01)	/* Is limit approximation applicable */
				minp2 = (double)1.0 - exp(-minp2);
			kp2->DBP_val = minp2;
			if (minp2 < minp)
				minp = minp2;
		}
		kp1->rawp.smallest_P = minp;
	}

	/* Sort HSPs by probability */
	(void) hsort((char *)HSPlist, hasHSP, sizeof(HSPlist[0]), smaller);

	/* print the matches for each sequence */
	if (ehigh > 0.01)
		phigh = 1. - exp(-ehigh);
	else
		phigh = ehigh;
	for (i = 0; i < hasHSP && HSPlist[i]->rawp.smallest_P <= phigh; ++i)
		print_seq(HSPlist[i]);
}

static void
print_seq(hp) /* copied from blastp.c */
	HSPPtr	hp;
{
	int		current, len, qpos, spos;
	HSPPtr	kp;
	register char	*cp;

	current = hp->seq_nbr;

	/* get header line */
	if (mfil_seek(hfile, header_beg[current], 0) == -1)
		fatal("seek error on header file");
	len = header_beg[current+1] - header_beg[current];
	if (len > sizeof(header)-1)
		len = sizeof(header)-1;
	if (mfil_dup(hfile, header, len) != len)
		fatal("error reading header");
	header[len] = NULLB;
	(void) printf("\n\n%s\n", header);

	/* get sequence */
#ifdef USE_SHM
	if (dbbuf != NULL) {
		len = seq_beg[current+1] - seq_beg[current] - 1;
		s_seq = dbbuf + seq_beg[current];
	}
	else
#endif
	{
#ifndef USE_MPROC
		len = seq_beg[current+1] - seq_beg[current];
		if (fseek(sfile, seq_beg[current], 0) == -1)
			fatal("fseek error on sequence file");
		if (fread(s_seq, 1, len, sfile) != len)
			fatal("error reading sequence");
		--len;
#else
		s_seq = s_seq0 + seq_beg[current];
		len = seq_beg[current+1] - seq_beg[current] - 1;
#endif
	}
	(void) printf("\tLength = %s\n", ld2s((long)len));
	
	for (kp = hp; kp != NULL; kp = kp->next) {
		qpos = kp->q_pos;
		spos = kp->s_pos;
		len = kp->len;

		/* Extract the aligned subsequences from q_seq & s_seq */

		mem_cpy(qq_seq, q_seq + qpos, len);
		qq_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = qq_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		mem_cpy(ss_seq, s_seq + spos, len);
		ss_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = ss_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		report(ss_seq, qq_seq, spos, qpos, kp->score, len, kp->DBP_val);
		(void) fflush(stdout);
	}
}

/* report - print a match */
static void
report(s, q, spos, qpos, score, len, DBP_val)
	char	*s, *q;
	int		spos, qpos, score, len;
	double	DBP_val;
{
	double	Expect;
	int		numdone = 0;
	int		mlen;

	Expect = stoe(score, Lambda, K, H, Neff, Meff);

	printf("\n  Score = %d, Expect = %#0.2lg, Poisson P = %#0.2g, length = %d\n",
			score, Expect, DBP_val, len);

	while (len > 0) {
		mlen = MIN(len, 60);
		printf("\nQuery: %5d %.60s %d\n",
				qpos + numdone + 1, q, qpos + numdone + mlen);
		printmid(q, s, len);
		printf("Sbjct: %5d %.60s %d\n",
				spos + numdone + 1, s, spos + numdone + mlen);
		q += 60;
		s += 60;
		len -= 60;
		numdone += 60;
	}
}

printmid(s1, s2, len)
	register char   *s1, *s2;
	register int    len;
{
	register int    i;
	register char	ch1, ch2;

	printf("%13s", "");
	if (len > 60)
		len = 60;
	while (len-- > 0) {
		i = matrx[aa_atob[ch1 = *s1++]][aa_atob[ch2 = *s2++]];
		if (i > 0) {
			if (ch1 == ch2)
				putchar(ch1);
			else
				putchar('+');
		}
		else
			putchar(' ');
	}
	putchar('\n');
}
