
/*
* BLASTA - Search a protein database for regions that approximately match
* regions of a query sequence.  The basic command syntax is
*
* 	BLASTA database query_sequence
*
* where:
* 	database names a file containing protein sequences, separated by
* 		header lines beginning with the character '>'.  A typical
* 		sequence entry is:
*
*		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
*		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
*		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
*		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
*		RADLIAYLKKATNE
*
* 	query_sequence names a file containing a protein sequence (where the
* 		header line is optional).
* 
* The BLASTA command permits optional additional arguments, such as X=50,
* that reset certain internal parameters.  The available parameters are:
*
*	M or m gives the name of a file containing substitution costs
*		(the default is "PAM120")
*	K or k gives the karlin cutoff value (the default is 100)
*	T or t gives the threshold for a word match (the default is 20)
*	X or x gives the value for terminating word extensions
*		(the default is 20)
* Thus a typical command is
*
*	BLASTA AABANK SEQ K=50 M=PAM250
*/
#include <ncbi.h>

#define HEADERLEN_MAX	200	/* maximum length of header line */
#define QUERYLEN_MAX	5000	/* maximum length of query sequence */
#define FILENAME_MAX	100	/* maximum length of file name */
#define WORD_SIZE	4	/* length of critical word (do not change!) */

#define DEFAULT_K	100	/* default cutoff for karlins */
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T	20	/* default threshold for word hits */
#define DEFAULT_X	20	/* default constant for hit extensions */

char
	header[HEADERLEN_MAX],
	*s_seq, *ss_seq,
	q_seq[QUERYLEN_MAX+2],
	qq_seq[QUERYLEN_MAX+3],
	M[FILENAME_MAX];	/* name of file containing substitution scores */

int
	haskarlins;	/* number of sequences containing a karlin */
long
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numkarlins,	/* number of karlins found */
	numhits;	/* number of database words in hash table */
int
	K,		/* cutoff for karlins */
	T,		/* threshold for word hits */
	X,		/* constant for hit extensions */
	n_kars;		/* number of karlins found for current sequence */

/* data structure to hold locations of matches for current database sequence */
int
	*_dlev,		/* extent of discovered matches .. */
	*diag_level,	/* on a given diagonal */
	*diag_stack,	/* list of diagonals with matches */
	*diag_top;	/* next diagonal with a match */

typedef struct karlin {
	int seq_nbr, len, q_pos, s_pos, s_hit_pos, score;
	double P_val, smallest_P;
	struct karlin *next;
	} Karlin, *Karlin_ptr;

#define MAX_KARLIN 500	/* limit on number of karlins per database sequence */

Karlin_ptr karl[MAX_KARLIN];

Karlin_ptr first_karlin = NULL;
Karlin_ptr *karl_list;

main(argc, argv)
int argc;
char *argv[];
{
	char *a;

	if (WORD_SIZE != 4)
		fatal("Word size must be 4.");
	if (argc < 3)
		fatalf("%s database query_sequence ...", argv[0]);
	K = DEFAULT_K;
	strcpy(M, DEFAULT_M);
	T = DEFAULT_T;
	X = DEFAULT_X;
	while (--argc > 2) {
		if (argv[argc][1] != '=')
			fatalf("argument %d has improper form", argc+1);
		a = argv[argc] + 2;
		switch (argv[argc][0]) {
			case 'K':
			case 'k':
				K = atoi(a);
				break;
			case 'M':
			case 'm':
				if (strlen(a) >= FILENAME_MAX)
					fatal("File name is too long.");
				strcpy(M,a);
				break;
			case 'T':
			case 't':
				T = atoi(a);
				break;
			case 'X':
			case 'x':
				X = atoi(a);
				break;
			default:
				fatal("Options are K, M, T and X.");
		}
	}
	get_weights();		/* read file of substitution weights */
	get_query(argv[2]);
	bld_table();		/* build table of critical words */

	while (get_seq(argv[1])) {
		n_kars = 0;
		search();	/* search a database sequence */
	}
	print_karlins();

	printf("\n\nStatistics:\n");
	printf("  # of critical words = %d\n", numwords);
	printf("  # of critical hits  = %d\n", numhits);
	printf("  # of matches found  = %d\n", numkarlins);
	printf("  # of db sequences with at least one match = %d\n",
		haskarlins);

	exit (0);
}

/* get_query - read the query sequence */
get_query(file_name)
char *file_name;
{
	FILE *qp, *ckopen();
	char *p, *fgets();
		
	qp = ckopen(file_name, "r");
	for (p = q_seq; ; ) {
		if (fgets(p, QUERYLEN_MAX + 2 - (p - q_seq), qp) == NULL) {
			fclose(qp);
			*p = '\0';
			break;
		}
		if (p == q_seq && *p == '>')	/* optional header line */
			continue;
		while (isupper(*p))
			++p;
		if (*p != '\n' && *p != '\0')
			fatalf("Illegal character %c in query sequence.", *p);
		if (p - q_seq > QUERYLEN_MAX)
			fatalf("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
}

/* UNIX-specific input routine; database must be processed by SETDB */

int
	current,		/* number of current sequence */
	sfile,			/* file descriptor for sequences */
	hfile;			/* file descriptor for headers */
long
	nbr_seqs,		/* number of database sequences */
	*seq_beg, 		/* starts of sequences */
	*header_beg;		/* starts of headers */

/* get_seq - read a database sequence and header line, return 0 on EOF */
int get_seq(file_name)
char *file_name;
{
	static int tfile, init = 1;
	int len, many;
	char name[FILENAME_MAX], *ckalloc();

	if (init) {
		init = 0;
		current = -1;
		sprintf(name, "%s.hed", file_name);
		hfile = open(name, 0);
		sprintf(name, "%s.seq", file_name);
		sfile = open(name, 0);
		sprintf(name, "%s.tab", file_name);
		tfile = open(name, 0);
		if (hfile == -1 || tfile == -1 || sfile == -1)
			fatalf("Cannot open database %s.", file_name);
		if (read(tfile, (char *)&nbr_seqs, sizeof(int)) == -1 ||
		   read(tfile, (char *)&maxlen, sizeof(int)) == -1)
			fatal("corrupted database file");
		s_seq = ckalloc(sizeof(char)*(maxlen+1));
		ss_seq = ckalloc(sizeof(char)*(maxlen+3));
		seq_beg = (long *) ckalloc(sizeof(long)*(nbr_seqs+1));
		header_beg = (long *) ckalloc(sizeof(long)*(nbr_seqs+1));
		many = sizeof(long)*(nbr_seqs+1);
		if (read(tfile, (char *)seq_beg, many) != many ||
		   read(tfile, (char *)header_beg, many) != many)
			fatal("improper compressed database file");
		_dlev = (int *) ckalloc(sizeof(int)*(2*nbr_seqs+1));
		diag_level = _dlev + nbr_seqs;
		diag_top = diag_stack =
			(int *) ckalloc(sizeof(int)*(2*nbr_seqs+1));
	}
	if (++current >= nbr_seqs)
		return 0;
	len = seq_beg[current+1] - seq_beg[current];
	read(sfile, s_seq, len);
	s_seq[len] = '\0';
	return 1;
}

/* -----------------   read file of substitution weights   ----------------*/

/*  Get_weights() reads a file giving weights for amino acid substitutions.
The proper form of the file is illustrated by the following "PAM120" matrix.
The first row gives amino acid abbreviations, which must be single capital
letters.  Weights must be integers between -128 and 127.

  A  R  N  B  D  C  Q  Z  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  X  
  3 -3  0  0  0 -3 -1 -1  0  1 -3 -1 -3 -2 -2 -4  1  1  1 -7 -4  0  0 
 -3  6 -1 -2 -3 -4  1 -1 -3 -4  1 -2 -4  2 -1 -4 -1 -1 -2  1 -6 -3  0 
  0 -1  4  3  2 -5  0  0  1  0  2 -2 -4  1 -3 -4 -2  1  0 -5 -2 -3  0 
  0 -2  3  4  3 -6  0  1  2  0  1 -3 -5  0 -4 -6 -2  0 -1 -7 -4 -3  0 
  0 -3  2  3  5 -7  1  2  3  0  0 -3 -5 -1 -4 -7 -2  0 -1 -8 -5 -3  0 
 -3 -4 -5 -6 -7  9 -7 -7 -7 -5 -4 -3 -7 -7 -6 -6 -3 -1 -3 -8 -1 -2  0 
 -1  1  0  0  1 -7  6  4  2 -3  3 -3 -2  0 -1 -6  0 -2 -2 -6 -5 -3  0 
 -1 -1  0  1  2 -7  4  5  3 -2  1 -3 -3 -1 -3 -6 -1 -2 -2 -7 -5 -3  0 
  0 -3  1  2  3 -7  2  3  5 -1 -1 -3 -4 -1 -4 -6 -1 -1 -2 -8 -4 -3  0 
  1 -4  0  0  0 -5 -3 -2 -1  5 -4 -4 -5 -3 -4 -5 -2  1 -1 -8 -6 -2  0 
 -3  1  2  1  0 -4  3  1 -1 -4  7 -4 -3 -2 -4 -2 -1 -2 -3 -5 -1 -3  0 
 -1 -2 -2 -3 -3 -3 -3 -3 -3 -4 -4  6  1 -2  1  0 -3 -2  0 -7 -2  3  0 
 -3 -4 -4 -5 -5 -7 -2 -3 -4 -5 -3  1  5 -4  3  0 -3 -4 -3 -5 -3  1  0 
 -2  2  1  0 -1 -7  0 -1 -1 -3 -2 -2 -4  5  0 -6 -2 -1 -1 -5 -6 -4  0 
 -2 -1 -3 -4 -4 -6 -1 -3 -4 -4 -4  1  3  0  8 -1 -3 -2 -1 -7 -4  1  0 
 -4 -4 -4 -6 -7 -6 -6 -6 -6 -5 -2  0  0 -6 -1  8 -5 -3 -4 -1  4 -3  0 
  1 -1 -2 -2 -2 -3  0 -1 -1 -2 -1 -3 -3 -2 -3 -5  6  1 -1 -7 -6 -2  0 
  1 -1  1  0  0 -1 -2 -2 -1  1 -2 -2 -4 -1 -2 -3  1  3  2 -2 -3 -2  0 
  1 -2  0 -1 -1 -3 -2 -2 -2 -1 -3  0 -3 -1 -1 -4 -1  2  4 -6 -3  0  0 
 -7  1 -5 -7 -8 -8 -6 -7 -8 -8 -5 -7 -5 -5 -7 -1 -7 -2 -6 12 -1 -8  0 
 -4 -6 -2 -4 -5 -1 -5 -5 -4 -6 -1 -2 -3 -6 -4  4 -6 -3 -3 -1  8 -3  0 
  0 -3 -3 -3 -3 -2 -3 -3 -3 -2 -3  3  1 -4  1 -3 -2 -2  0 -8 -3  5  0 
  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0  0 

*/

int	alpha;			/* alpabet size */
char	achars[26],		/* array of amino acid names */
	maxcost[128],		/* max cost of substituting for a character */
	matrx[128][128],	/* matrx[i][j] = cost of replacing i by j */
	order[128][26];		/* order[i][j] = j-th best char to sub. for i */

get_weights()
{
	FILE *sfile, *ckopen();
	int i, j, score;
	char *m, *o, *p, *q, t;

	sfile = ckopen(M, "r");

	/* read residue names */
	p = achars;
	while ((i = fgetc(sfile)) != '\n')
		if (isupper(i)) {
			if (p - achars >= 26)
				fatalf("Too many names in %s.", M);
			*p++ = i;
		} else if (i != ' ')
			fatalf("%s: residue names must be upper case.", M);
	alpha = p - achars;

	/* for illegal characters, set weight to lowest possible value */
	for (i = 0; i < 128; i++) {
		m = matrx[i];
		for (j = 0; j < 128; j++)
			m[j] = -128;
		maxcost[i] = -128;
	}

	/* read substitution weights for each residue */
	for (i = 0; i < alpha; i++) {
		m = matrx[achars[i]];
		for (j = 0; j < alpha; j++) {
			fscanf(sfile," %d",&score);
			m[achars[j]] = score;
		}
		o = order[achars[i]];
		for (j = 0; j < alpha; j++)
			o[j] = achars[j];
		/* bubble-sort residue names by cost of substitution for i-th
		   residue */
		for (q = o + (alpha-1); q > o; --q)
			for (p = o; p < q; ++p)
				if (m[*p] < m[*q]) {
					t = *p;
					*p = *q;
					*q = t;
				}
		maxcost[achars[i]] = m[*o];
	}

	fclose(sfile);
}

/* -----------------------   build table of critical words   -----------------*/

struct hash_node {
	int ch;			/* fourth letter of word */
	struct match *matches;	/* positions where word hits query sequence */
	struct hash_node *link;	/* next word with same first three letters */
};

struct match {
	int pos;		/* position in query sequence */
	int score;		/* score where word hits */
	struct match *succ;	/* next hit for this word */
};

/* The hash value of a four-letter word is simply an encoding of its first
three letters:
	hval = (word[0] - 'A')*1024 + (word[1] - 'A')*32 + word[2] - 'A';
This is space-inefficient since some hash values are unreachable, but the
hash value is quickly updated as we march along a database sequence.  If s
points to the last letter of the new word, then the new hash value is:
	hval = ((hval & 01777) << 5) + (*s - 'A');
*/

struct hash_node *hashtab[26426];	/* hval of "ZZZ" is 26425 */

char
	word[WORD_SIZE],	/* candidate for critical word */
	best[WORD_SIZE],	/* best[i] = smallest score of word[0]..word[i]
				   that can be extended to a critical word */
	*query;			/* pointer into query sequence */

bld_table()
{
	int i;

	for (query = q_seq; query[3] != '\0'; query++) {
		best[3] = T;
		for (i = 3; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);
	}
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for word[0]..word[i-1] */
enumerate(i, score)
int i, score;
{
	char *o = order[query[i]]; /* possible letters by decreasing weight */
	int c, new_score;

	for (c = 0; c < alpha; ++c) {
		word[i] = o[c];
		new_score = score + matrx[query[i]][word[i]];
		if (new_score < best[i])
			break;
		if (i < WORD_SIZE-1)
			enumerate(i+1, new_score);
		else
			add_word(query - q_seq + WORD_SIZE, new_score);
	}
}
    
/* add_word - add a word to the table of critical words */
add_word(pos, score)
int pos, score;
{
	struct hash_node *h;
	struct match *m;
	long hval;
	char *ckalloc();

	numwords++;
	hval = (word[0] - 'A')*1024 + (word[1] - 'A')*32 + word[2] - 'A';
	for (h = hashtab[hval]; h; h = h->link)
		if (h->ch == word[3])
			break;
	if (h == NULL) {
		h = (struct hash_node *) ckalloc (sizeof(struct hash_node));
		h->link = hashtab[hval];
		hashtab[hval] = h;
		h->ch = word[3];
		h->matches = NULL;
	}
	m = (struct match *) ckalloc (sizeof(struct match));
	m->succ = h->matches;
	h->matches = m;
	m->pos = pos;
	m->score = score;
}

/* -----------------------   search a database sequence   --------------------*/

search()
{
	register struct hash_node *h;
	struct match *m;
	register char *s;
	register long hval;

	s = s_seq;
	if (s[0] == '\0' || s[1] == '\0' || s[2] == '\0')
		return;
	hval = (s[0] - 'A')*1024 + (s[1] - 'A')*32 + s[2] - 'A';
	for (s += 3; *s; ++s) {
		for (h = hashtab[hval]; h; h = h->link)
			if (h->ch == *s) {
				for (m = h->matches; m; m = m->succ)
					karlin(s+1, m->pos, m->score);
				break;
			}
		hval = ((hval & 01777) << 5) + (*s - 'A');
	}

	get_pvals();

	while (diag_top > diag_stack)
		diag_level[*--diag_top] = 0;
}

/* karlin - extend a word-sized hit to a longer match */
karlin(s_pos, q_pos, score)
char *s_pos;
int q_pos, score;
{
	char *s_beg, *q_beg, *q_end, *q, *s;
	int min_sum, max_sum, sum, diag, s_hit_pos = q_pos;
	Karlin *kp;

	numhits++;

	diag = (s_pos-s_seq) - q_pos;
	if (diag_level[diag] > q_pos)
		return;

	/* extend to the left */
	max_sum = sum = 0;
	s = s_pos;
	q_end = q = q_seq + q_pos;
	while (*s != '\0' && *q != '\0' && sum >= max_sum - X)
		if ((sum += matrx[*q++][*s++]) > max_sum) {
			max_sum = sum;
			q_end = q;
		}

	/* extend to the right */
	min_sum = sum = 0;
	s_beg = s = s_pos - WORD_SIZE;
	q_beg = q = (q_seq + q_pos) - WORD_SIZE;
	while (s > s_seq && q > q_seq && sum >= min_sum - X)
		if ((sum += matrx[*--q][*--s]) > min_sum) {
			min_sum = sum;
			s_beg = s;
			q_beg = q;
		}

	score += max_sum + min_sum;
	if (score >= K) {
		++numkarlins;
		kp = karl[n_kars++] = (Karlin *)ckalloc(sizeof(Karlin));
		kp->seq_nbr = current;
		kp->len = q_end - q_beg;
		kp->q_pos = q_beg - q_seq;
		kp->s_pos = s_beg - s_seq;
		kp->s_hit_pos = s_hit_pos;
		kp->score = score;
	}

	if (diag_level[diag] <= 0)
		*diag_top++ = diag;
	diag_level[diag] = (q_end - q_seq) + WORD_SIZE;
}

get_pvals()
{
	int i, j, A, m, n;
	Karlin *kp;
	double lambda = 0.336, K = 0.166;
	double P, Q, R, S, T, exp();

	if (n_kars == 0)
		return;

	++haskarlins;
	/* bubble sort */
	for (i = n_kars-1; i > 0; --i)
		for (j = 0; j < i; ++j)
			if (karl[j]->score < karl[j+1]->score) {
				kp = karl[j];
				karl[j] = karl[j+1];
				karl[j+1] = kp;
			}

	m = strlen(q_seq);
	n = strlen(s_seq);
	for (A = 1; A <= n_kars; ++A) {
		kp = karl[A-1];
		S = kp->score;
		Q = K*m*n*exp(-lambda*S);
		T = R = 1.0;
		for (j = 1; j < A; ++j) {
			R *= (Q/j);
			T += R;
		}
		P = 1.0 - exp(-Q)*T;
		kp->P_val = P;
	}

	karl[n_kars-1]->next = first_karlin;
	for (i = n_kars-2; i >= 0; --i)
		karl[i]->next = karl[i+1];
	first_karlin = karl[0];
}

print_karlins()
{
	int i;
	Karlin *kp1, *kp2;
	char *ckalloc();

	if (haskarlins < MAX_KARLIN)
		karl_list = karl;
	else
		karl_list = (Karlin_ptr *)ckalloc(haskarlins*sizeof(Karlin_ptr));
	for (i = 0; i < haskarlins; ++i) {
		karl_list[i] = first_karlin;
		for (kp1 = first_karlin; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		first_karlin = kp2;
	}

	/* for each sequence, compute the smallest P_value of its karlins */
	for (i = 0; i < haskarlins; ++i) {
		kp1 = karl_list[i];
		kp1->smallest_P = kp1->P_val;
		for (kp2 = kp1->next; kp2; kp2 = kp2->next)
			if (kp2->P_val < kp1->smallest_P)
				kp1->smallest_P = kp2->P_val;
	}

	sort_karlins();

	/* print the matches for each sequence */
	for (i = 0; i < haskarlins; ++i)
		print_seq(i);
}

sort_karlins()
{
	int i;
	Karlin_ptr kp;

	for (i = (haskarlins/2) - 1; i >= 0; --i)
		heapify(i, haskarlins-1);
	for (i = haskarlins-1; i > 0; --i) {
		kp = karl_list[0];
		karl_list[0] = karl_list[i];
		karl_list[i] = kp;
		if (i > 1)
			heapify(0, i-1);
	}
}

heapify(i, last)
int i, last;
{
	int lim = (last-1)/2, left_son, small_son;
	Karlin_ptr kp;

	while (i <= lim) {
		left_son = 2*i + 1;
		if (left_son == last)
			small_son = left_son;
		else
			small_son = smaller(left_son, left_son+1);
		if (smaller(i, small_son) == small_son) {
			kp = karl_list[i];
			karl_list[i] = karl_list[small_son];
			karl_list[small_son] = kp;
			i = small_son;
		} else
			break;
	}
}

int smaller(i, j)	/* if tie, return first arg */
int i, j;
{
	Karlin_ptr ki = karl_list[i], kj = karl_list[j];

	if (ki->smallest_P == 0.0 && kj->smallest_P == 0.0)
		return ((ki->score <= kj->score) ? i : j);
	else
		return ((ki->smallest_P >= kj->smallest_P) ? i : j);
}


print_seq(i)
int i;
{
	int current, len, hpos, qpos, spos, n;
	Karlin_ptr kp;

	current = karl_list[i]->seq_nbr;

	/* get header line */
	lseek(hfile, header_beg[current], 0);
	len = header_beg[current+1] - header_beg[current];
	read(hfile, header, len);
	if (len > 78)
		len = 78;
	header[len-1] = '\0';
	printf("\n\n%s\n", header);

	/* get sequence */
	lseek(sfile, seq_beg[current], 0);
	len = seq_beg[current+1] - seq_beg[current];
	read(sfile, s_seq, len);
	s_seq[len] = '\0';
	
	for (kp = karl_list[i]; kp; kp = kp->next) {
		hpos = kp->s_hit_pos;
		qpos = kp->q_pos;
		spos = kp->s_pos;
		len = kp->len;
		n = hpos - qpos - WORD_SIZE;
		add_bars(qq_seq, q_seq + qpos, n, len);
		add_bars(ss_seq, s_seq + spos, n, len);
		report(ss_seq, qq_seq, spos, qpos, kp->score, len, kp->P_val);
	}
}

add_bars(to, from, n, len)
char *to, *from;
int n, len;
{
	int remainder = len - n - WORD_SIZE;
		
	strncpy(to, from, n);
	to += n;
	from += n;
	*to++ = '|';
	strncpy(to, from, WORD_SIZE);
	to +=  WORD_SIZE;
	from += WORD_SIZE;
	*to++ = '|';
	strncpy(to, from, remainder);
	to[remainder] = '\0';
}

/* report - print a match */
report(s, q, spos, qpos, score, len, P_val)
char *s, *q;
int spos, qpos, score, len;
double P_val;
{

	printf("\n  Score  = %d, len = %d, P-val = %g\n",
		score, len, P_val);
  	printf("  Query start index = %d, Subject start index = %d\n",
		qpos+1, spos+1);
	while (len > 0) {
		printf("\nQuery: %.60s\n",q);
		printf("Sbjct: %.60s\n",s);
		q += 60;
		s += 60;
		len -= 60;
	}
}

/* ----------------------------   utility routines   -------------------------*/

/* fatal - print message and die */
fatal(msg)
char *msg;
{
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

/* fatalf - format message, print it, and die */
fatalf(msg, val)
char *msg, *val;
{
	fprintf(stderr, msg, val);
	putc('\n', stderr);
	exit(1);
}
	
/* ckopen - open file; check for success */
FILE *ckopen(name, mode)
char *name, *mode;
{
	FILE *fopen(), *fp;

	if ((fp = fopen(name, mode)) == NULL)
		fatalf("Cannot open %s.", name);
	return(fp);
}

/* ckalloc - allocate space; check for success */
char *ckalloc(amount)
int amount;
{
	char *malloc(), *p;

	if ((p = malloc( (unsigned) amount)) == NULL)
		fatal("Ran out of memory.");
	return(p);
}

