/*
* BLASTALL - Compare a protein database against itself
*
* 	BLASTALL database
*
* where:
* 	database names a file containing protein sequences, separated by
* 		header lines beginning with the character '>'.  A typical
* 		sequence entry is:
*
*		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
*		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
*		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
*		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
*		RADLIAYLKKATNE
*
* 	query_sequence names a file containing a protein sequence (where the
* 		header line is optional).
* 
* The BLASTALL command permits optional additional arguments, such as X=50,
* that reset certain internal parameters.  The available parameters are:
*
*	M or m is the name of a file containing substitution costs
*		(default is "PAM120")
*	S or s gives the HSP score cutoff value (default is 45)
*	T or t gives the threshold for a word match (default is 17)
*	X or x gives the value for terminating word extensions (default is 20)
* Thus a typical command is
*
*	BLASTALL AABANK SEQ S=50 M=PAM250
*
*	(Note:  set tabstops every 4 columns for a better look).
*/
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>

#if defined(USE_SHM) && !defined(SYSV_SHM_AVAIL)
#undef USE_SHM
#endif

#define BLASTALL
#define EXTERN
#include "shmutil.h"
#include "blast.h"
#include "aabet.h"	/* the alphabet */

#define DFA_USEUNIONS
#include <dfa.h>

#define HEADERLEN_MAX	200		/* maximum length of header line */
#define QUERYLEN_MAX	5000	/* maximum length of query sequence */
#define FILENAME_MAX	100		/* maximum length of file name */
#define MAXWORD_SIZE	7	/* maximum length of critical word */

#define DEFAULT_W	4			/* default word size */
#define DEFAULT_S	40			/* default cutoff for saving diagonals */
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T	13	/* default threshold for neighborhood word hits */
#define DEFAULT_X	30	/* default score drop-off during hit extensions */


static Int4		lows, highs;
double	lambda, lambda3, lambdaS, K, K3, KS;

/* One index_entry per sequence in the database, stored in an index file */
typedef struct {
			char	name[12];	/* visible string identifier of the sequence */
			off_t	offset;		/* offset into data file of this seq's diags */
		} index_entry;

index_entry	ie;

/* One report_header per sequence, stored in the data file */
typedef struct {
			char	name[12];
			u_short	seqlen;		/* length of this sequence */
			u_short	diagcnt;	/* No. of diag_entrys to follow */
			double	lambda;
			double	K;
			double	lambda3;	/* Lambda value for 3-way alignment */
			double	K3;			/* K value for 3-way alignment */
			double	lambdaS;
			double	KS;
		} report_header;

report_header	rh;

#define THREEWAY

/* One diag_entry per saved diagonal.  All diag_entrys follow a report_header */
typedef struct {
			Uint2	seqno;		/* Points to row in index_entry table */
			Int2	diag;		/* Identity of the aligned diagonal */
			Uint1	score;		/* Score for this diagonal */
#ifdef THREEWAY
			Uint2	diag3;		/* Points to another entry in this list */
			Uint2	score3;		/* 3-way alignment score */
#endif
		} diag_entry;

diag_entry	de;

u_long	maxscore;	/* max. score achieved against a single sequence */
u_char	*maxscores;	/* array of max. scores against all db sequences */

char
	*dbbuf,			/* Memory-resident database w/null prefix (sentinel) byte*/
	*q_seq,			/* current query sequence */
	*s_seq,			/* subject sequence */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

long	totdblen;	/* total length of database */
int	q_len;			/* length of query sequence */
int	s_len;			/* length of current database sequence */

int
	totHSP,	/* total number of HSPs against this query sequence */
	hasHSP;	/* number of sequences containing a HSP */
long
	maxlen;		/* max length of a database sequence */
int
	W = DEFAULT_W,		/* word size */
	S = DEFAULT_S,		/* cutoff for HSPs */
	T = DEFAULT_T,		/* threshold for word hits */
	X = DEFAULT_X,		/* constant for hit extensions */
	minusX = -DEFAULT_X;

/* data structure to hold locations of matches for current database sequence */
unsigned int
	*_dlev,		/* extent of discovered matches... */
	*diag_level;	/* on a given diagonal */
int
	*diag_stack,/* list of diagonals with matches */
	*diag_top;	/* next diagonal with a match */

HSPPtr	firstHSP = NULL;
HSPPtr	*HSPlist = NULL;

static int
	curquery0,		/* starting number for this program run */
	curquery,		/* number of current query sequence */
	current,		/* number of current subject sequence */
	sfile,			/* file descriptor for sequences */
	hfile;			/* file descriptor for headers */

static off_t
	nbr_seqs,		/* number of sequences in database */
	*seq_beg, 		/* starts of sequences in file or memory-resident dbase */
	*header_beg;	/* starts of headers */

long		resume, runlength = -1;

static void	sigterm();
static void	search_aa();
static void	extend_aa();
static void	get_pvals();
static void	initdb();
static void	lhspsave PROTO((struct hsp far *));
static int	get_seq();
static struct hsp far	*kalloc();
off_t	lseek();
char	*strcpy();
static	in_ie(), out_ie(), in_rh(), out_rh(), in_de(), out_de();


DFA	*dp;	/* the automaton */
char	*dpbuf[6*KBYTE*KBYTE/sizeof(char *)];

char	*module;
char	buf[100];
char	doquit = 0;	/* TRUE if TERMinate signal has been received */
FILE	*indexfp, *datafp, *scorefp;

main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o, *a;

	module = basename(argv[0]);

	signal(SIGTERM, sigterm);

	if (argc < 2)
		fatal("Usage:  %s database [options]\nValid %s options are S, M, T, W and X.", module, module);
	while (--argc > 1) {
		if (argv[argc][1] == '=')
			o = argv[argc][0];
		else
			if (argv[argc][0] == '-')
				o = argv[argc][1];
			else
				fatal("Argument %d (%s) has improper form", argc+1, argv[argc]);
		a = argv[argc] + 2;
		switch (o) {
			case 'M':
			case 'm':
				if (strlen(a) >= FILENAME_MAX)
					fatal("PAM file name is too long.");
				(void) strcpy(M, a);
				break;
			case 'W':
			case 'w':
				if (sscanf(a, "%d", &W) != 1 || W <= 0)
					fatal("Bad value for W.");
				if (W > MAXWORD_SIZE)
					fatal("W is too big--must be less than %d.", MAXWORD_SIZE+1);
				break;
			case 'R': /* Start comparison with this sequence */
			case 'r':
				if (sscanf(a, "%ld", &resume) != 1 || resume < 1)
					fatal("Bad value for R.");
				break;
			/* Exit after performing this many comparisons */
			case 'N':
			case 'n':
				runlength = atoi(a);
				if (runlength <= 0)
					fatal("Bad value for N.");
				break;
			case 'S':
			case 's':
				if (sscanf(a, "%d", &S) != 1 || S < 0)
					fatal("Bad value for S.");
				break;
			case 'T':
			case 't':
				if (sscanf(a, "%d", &T) != 1 || T < 1)
					fatal("Bad value for T.");
				break;
			case 'X':
			case 'x':
				if (sscanf(a, "%d", &X) != 1 || X < 1 || X > 1000000)
					fatal("Bad value for X.");
				minusX = -X;
				break;
			default:
				fatal("Valid options are S, M, T, W and X.");
		}
	}

	get_weights();		/* read file of substitution weights */

	initdb(argv[1]);	/* initialize database */

	hspsave = lhspsave;	/* local HSP save routine */

	while (get_query()) {
		bld_table();		/* build table of critical words */
		S = lambdak(lows, highs, q_seq, q_len);
		printf("S = %d\n", S);
		fflush(stdout);
		current = -1;
		while (get_seq()) {
			if (curquery == current)
				continue;
			if (doquit)
				break;
			search_aa();	/* search a database sequence */
		}
		if (doquit)
			break;
		writereport();		/* save results */
		dfa_destruct(dp);	/* remove DFA */
		if (runlength > 0 && --runlength <= 0)
			break;
	}
	fclose(indexfp);
	fclose(datafp);
	fclose(scorefp);
	if (doquit)
		fatal("TERMinate signal received.\n %d seqs processed in this run, ending with #%d\n",
				curquery - curquery0 - 1, curquery);

	fprintf(stderr, "\nNormal termination--full BLAST run completed.\n");
	exit(0);
	/*NOTREACHED*/
}

/* get_query - read the next query sequence */
get_query()
{
	char	*strchr();
	register char	*cp;
	int		len;

	/* seek to position in header table but skip the initial > symbol */
	if (++curquery >= nbr_seqs)
		return 0;
	lseek(hfile, header_beg[curquery]+1, 0);
	len = header_beg[curquery+1] - header_beg[curquery] - 1;
	if (len > sizeof(ie.name)-1)
		len = sizeof(ie.name)-1;
	read(hfile, ie.name, len);
	ie.name[sizeof(ie.name)-1] = NULLB;
	cp = strchr(ie.name, ' ');
	if (cp != NULL)
		while (cp < &ie.name[sizeof(ie.name)])
			*cp++ = NULLB;
	q_len = seq_beg[curquery+1] - seq_beg[curquery] - 1;
	q_seq = dbbuf + seq_beg[curquery];
	bzero(maxscores, nbr_seqs*sizeof(*maxscores));
	return 1;
}

/* UNIX-specific input routine; database must be processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	char	name[FILENAME_MAX];
	int		tfile, len;
	size_t	many;
	size_t	dbsize;

	(void) sprintf(name, "%s.idx", dbname);
	/* First find out how many entries already exist in the index file */
	curquery0 = -1;
	indexfp = fopen(name, "r");
	if (indexfp != NULL) {
		while (in_ie(indexfp, &ie))
			++curquery0;
		fclose(indexfp);
		if (resume == 0)
			printf("Program will resume at sequence #%d\n", curquery0+2);
		fflush(stdout);
	}
	if (resume != 0) {
			printf("Program will start at sequence #%d\n", resume);
			fflush(stdout);
			curquery0 = resume-2;
	}
	curquery = curquery0;
	indexfp = fopen(name, "a");
	if (indexfp == NULL)
		fatal("Unable to open/create index file \"%s\"", name);

	(void) sprintf(name, "%s.dat", dbname);
	datafp = fopen(name, "a");
	if (datafp == NULL)
		fatal("Unable to open/create data file \"%s\"", name);

	(void) sprintf(name, "%s.scr", dbname);
	scorefp = fopen(name, "a");
	if (scorefp == NULL)
		fatal("Unable to open/create score file \"%s\"", name);

	(void) sprintf(name, "%s.hed", dbname);
	hfile = open(name, 0);
	if (hfile == -1)
		fatal("Cannot open database file %s", name);
	if (sys_filesize(name) == 0)
		fatal("Database file has zero length:  %s", name);
	(void) sprintf(name, "%s.bsq", dbname);
	sfile = open(name, 0);
	if (sfile == -1)
		fatal("Cannot open database file %s", name);
	totdblen = dbsize = sys_filesize(name);
	if (dbsize == 0)
		fatal("Database file has zero length:  %s", name);
	(void) sprintf(name, "%s.tab", dbname);
	tfile = open(name, 0);
	if (tfile == -1)
		fatal("Cannot open database file %s", name);
	if (read(tfile, (char *)&nbr_seqs, sizeof(nbr_seqs)) != sizeof(nbr_seqs) ||
			read(tfile, (char *)&maxlen, sizeof(maxlen)) != sizeof(maxlen))
		fatal("corrupted database file:  %s", name);

	if (!shmexists(BLASTPKEY, dbsize+1)) {
		if ((dbbuf = shmalloc(BLASTPKEY, 0666, dbsize+1)) == NULL)
			dbbuf = ckalloc(dbsize+1);
		*dbbuf++ = NULLB; /* Initialize the null prefix (sentinel) byte */
		if (read(sfile, (char *)dbbuf, (int)dbsize) != dbsize)
			fatal("Database read error.");
	}
	else {
		if ((dbbuf = shmalloc(BLASTPKEY, 0666, dbsize+1)) == NULL)
			fatal("Confused about existence of db shared memory segment");
		++dbbuf;
	}

	many = sizeof(*seq_beg)*(nbr_seqs+1);
	seq_beg = (off_t *)ckalloc(many);
	if (read(tfile, (char *)seq_beg, (off_t)many) != many)
		fatal("improper sequence offsets file");

	many = sizeof(*header_beg)*(nbr_seqs+1);
	header_beg = (off_t *)ckalloc(many);
	if (read(tfile, (char *)header_beg, (off_t)many) != many)
		fatal("improper header offsets file");

	maxscores = (u_char *)ckalloc(nbr_seqs*sizeof(*maxscores));

	len = maxlen + maxlen + 1;
	_dlev = (int *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + maxlen;
	diag_top = diag_stack = (int *)ckalloc(sizeof(*diag_top)*len);
}

/* get_seq - read a database sequence and header line, return 0 on EOF */
static int
get_seq()
{
	if (++current >= nbr_seqs)
		return 0;
	s_len = seq_beg[current+1] - seq_beg[current] - 1;
	s_seq = dbbuf + seq_beg[current];
	if (s_len > maxlen)
		fatal("Database sequence longer than expected maxlen!");

	maxscore = 0;
	return 1;
}

/* -----------------   read file of substitution weights   ----------------*/


static signed char	achars[27],		/* array of amino acid names, plus NULLB */
	order[AAID_MAX+1][32];		/* order[i][j] = j-th best char to sub. for i */

int		alpha;			/* alpabet size */
int		minchar=INT_MAX, maxchar=INT_MIN; /* Min. and Max. letter values */

SCORE
	maxcost[AAID_MAX+1], /* max cost of substituting for a character */
	matrx[AAID_MAX+1][32]; /* matrx[i][j] = cost of replacing i by j */


get_weights()
{
	FILE *mfile, *ckopen();
	Int4	score;
	int		i, j;
	SCORE	*m;
	signed char *o, *p, *q, t;

	mfile = ckopen(M, "r");

	/* read residue names */
	p = achars;
	while ((i = fgetc(mfile)) != '\n' && i != 0) {
		if (isalpha(i) && isupper(i)) {
			if (p - achars >= sizeof(achars)-1)
				fatal("Too many amino acid names in %s.", M);
			j = aa_atob[i]; /* Binary-encode the residue name */
			if (j > AAID_MAX)
				fatal("%s:  Bad residue name: ``%c''.", M, i);
			*p++ = j;
			if (j > maxchar)
				maxchar = j;
			else
				if (j < minchar)
					minchar = j;
		}
		else {
			if (isspace(i))
				continue;
			if (isprint(i))
				fatal("%s:  Bad residue name: ``%c.''", M, i);
			else
				fatal("%s:  Bad residue name: #%d.", M, i);
		}
	}

	alpha = p - achars;

	/* for illegal characters, set weight to lowest possible value */
	mem_setl(maxcost, SCORE_MIN/2, DIM1(maxcost));
	mem_setl(&matrx[0][0], SCORE_MIN/2, sizeof(matrx)/sizeof(matrx[0][0]));

	/* read substitution weights for each residue */
	for (i = 0; i < alpha; i++) {
		m = matrx[achars[i]];
		for (j = 0; j < alpha; ++j) {
			if (fscanf(mfile, " %d", &score) != 1)
				fatal("Conversion error reading PAM file ``%s'', on line %d.",
							M, i+2);
			m[achars[j]] = score;
			if (lows > score)
				lows = score;
			if (highs < score)
				highs = score;
		}
		o = order[achars[i]];
		for (j = 0; j < alpha; ++j)
			o[j] = achars[j];
		/* Bubble-sort residue names by cost of substitution for i-th residue */
		for (q = o + (alpha-1); q > o; --q)
			for (p = o; p < q; ++p)
				if (m[*p] < m[*q]) {
					t = *p;
					*p = *q;
					*q = t;
				}
		maxcost[achars[i]] = m[*o];
	}
	(void) fclose(mfile);
}

/* -----------------------   build table of critical words   -----------------*/


signed char
	word[MAXWORD_SIZE];	/* candidate for critical word */

SCORE
	best[MAXWORD_SIZE];	/* best[i] = smallest score of word[0]..word[i]
				   that can be extended to a critical word */
char	*query;			/* pointer into query sequence */

static
bld_table()
{
	register int i;
	register int	wm1 = W-1;
	register int	score;

	dp = dfa_open((char *)dpbuf, sizeof(dpbuf),
				(LETTER *)AAID_MIN, AAID_MAX-AAID_MIN+1, MAXWORD_SIZE,
				ckalloc, 256*KBYTE, NULL, (DFA_HALTNONALPHAS|DFA_NOXLTAB));
	if (dp == NULL) {
		dfa_perror(dp, module);
		fatal("bld_table");
	}

	for (score=i=0; i<wm1; ++i)
		score += matrx[q_seq[i]][q_seq[i]];

	for (query = q_seq; query[wm1] != NULLB; query++) {
		best[wm1] = T;
		for (i = wm1; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);

		/* Check score for perfect match */
		score += matrx[query[wm1]][query[wm1]];
		if (score < T && score > 0) {
			/* Add perfect match even if its score is less than T */
			if (dfa_add(dp, (LETTER *)query, W, (memindex)(query+W)) == NULL)
				fatal("error in dfa_add");
		}
		score -= matrx[query[0]][query[0]];
	}
	if (dfa_close(dp) != 0)
		fatal("error in dfa_close");
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for word[0]..word[i-1] */
enumerate(i, score)
	register int		i;
	register int	score;
{
	/* possible letters by decreasing weight */
	register signed char	*o = order[query[i]];
	register int	c;
	register int	new_score;

	for (c = 0; c < alpha; ++c) {
		new_score = score + matrx[query[i]][word[i] = o[c]];
		if (new_score < best[i])
			break;
		if (i < W-1)
			enumerate(i+1, new_score);
		else {
			if (dfa_add(dp, (LETTER *)word, W, (memindex)(query + W)) == NULL)
				fatal("error in dfa_add");
		}
	}
}

/* -----------------------   search a database sequence   --------------------*/
static void
search_aa()
{
	register STATE	*S;
	register dfa_accept	*A;
	register char *s;

	S = dfa_state0(dp);
	s = s_seq;

	for (;;) {
		/*
			The machine "accepts" when a word hit is encountered,
			and "halts" on null terminator or other non-alphabetic character
			(sentinel bytes)
		*/
		while (!IsAccepting(S = S->next[*s++]))
			;

		/* Has machine halted (e.g., end of sequence)? */
		if ((A = (dfa_accept *)Unmunged(S)) == NULL)
			break; /* yes */

		/* Found at least one hit... */

		/* Does list of hit positions contain more than one item? */
		if (!IsMunged(S = A->ptr.retState))
			A = A->val.next; /* yes, and A->val.next points to head of list */
		S = (STATE *)Unmunged(S); /* Always Unmunge S, to avoid a jump */

		/* Extend each hit in the linked list */
		do
			extend_aa(s, (char *)A->val.patID);
		while (!IsMunged(A = A->ptr.chain));
	}

	if (curHSP > 0)
		get_pvals();

	while (diag_top > diag_stack)
		diag_level[*--diag_top] = 0;
}

/* extend_aa - extend a word-sized hit to a longer match */
static void
extend_aa(s_pos, q_pos)
	register char	*s_pos;
	register char	*q_pos;
{
	register char	*q, *s;
	register long	sum, score;
	register int	x;
	register char	*q_beg, *q_end;
	register int	diag;

	x = q_pos - q_seq;
	if (diag_level[diag = (s_pos - s_seq) - x] > x)
		return;
	if (diag_level[diag] == 0)
		*diag_top++ = diag;

	/* Find where positive scoring starts & ends within the word hit */
	q_beg = q = q_pos - W;
	s = s_pos - W;
	score = sum = 0;
	do
		if ((sum += matrx[*q++][*s++]) > score) {
			q_end = q;
			score = sum;
		}
		else
			if (sum <= 0) {
				q_beg = q;
				score = sum = 0;
			}
	while (q < q_pos);

	x = minusX;

	/* extend to the right */
	s = s_pos + ((q = q_end) - q_pos);
	sum = 0;
	do
		if ((sum += matrx[*q++][*s++]) > 0) {
			score += sum;
			sum = 0;
			q_end = q;
		}
	while (sum >= x); /* fails on null terminator of either sequence */

	/* extend to the left */
	s = s_pos - (q_pos - (q = q_beg));
	sum = 0;
	do
		if ((sum += matrx[*--q][*--s]) > 0) {
			score += sum;
			sum = 0;
			q_beg = q;
		}
	while (sum >= x); /* fails on null prefix of either sequence */

	if (score < S) {	/* Score is not reportable */
		/* Record how far this diagonal has been traversed */
		diag_level[diag] = (q_end - q_seq) + W;
		if (score > maxscore)
			maxscore = score;
		return;
	}

	/* Score is reportable.  Now examine the diagonal rigorously... */
	/* (This was not done by the program described in the JMB paper) */

	/* Mark this diagonal (in advance) as having been completely examined */
	diag_level[diag] = UINT_MAX;

	/* Start at the very beginning of the diagonal */
	s = s_seq;
	q = q_seq;
	if (diag > 0)
		s += diag;
	else
		q -= diag;

	q_beg = q;
	score = sum = 0;
	for (;;) {
		if ((sum += matrx[*q++][*s++]) <= 0) {
			if (score > 0) {
				q = q_end;
				s = s_seq + (q_end - q_seq) + diag;
				if (score >= S) {
					struct hsp	h;

					if (score > maxscore)
						maxscore = score;
					/* Record the HSP */
					h.score = score;
					h.seq_nbr = current;
					h.len = q_end - q_beg;
					h.s_pos = (h.q_pos = q_beg - q_seq) + diag;
					ckoverlap(&h);
				}
			}
			else
				if (sum < SCORE_MIN/4)
					break; /* Hit a sentinel byte */
			q_beg = q_end = q;
			score = sum = 0;
		}
		else
			if (sum > score) {
				score = sum;
				q_end = q;
			}
	}
}

static void
lhspsave(hp)
	struct hsp far	*hp;
{
	register struct hsp far	*newhp;

	if (curHSP >= maxHSP)
		HSP = (struct hsp far **)ckrealloc(HSP, (maxHSP+=100)*sizeof(HSPPtr));
	HSP[curHSP++] = newhp = kalloc();
	mem_cpy((char far *)newhp, (char far *)hp, sizeof(*newhp));
}

int
compar(hp1, hp2)
	HSPPtr	*hp1, *hp2;
{
	register HSPPtr	hi = *hp1, hj = *hp2;

	return hj->score - hi->score;
}

static void
get_pvals()
{
	register int i, olddiag, newdiag, highs;
	register HSPPtr	*kpp;
	register HSPPtr kp, kp2;

	if (maxscore >= UINT1_MAX)
		maxscore = UINT1_MAX;
	maxscores[current] = maxscore;

	if (curHSP == 0)
		return;

	/* Make sure a given diagonal is represented only once in HSP[*] */
	kpp = &HSP[0];
	olddiag = INT_MIN;
	for (i=0; i<curHSP; ++i) {
		kp = HSP[i];
		if ((newdiag = kp->s_pos - kp->q_pos) != olddiag) {
			olddiag = newdiag;
			highs = (*kpp++ = kp2 = kp)->score;
		}
		else
			if (highs < kp->score)
				kp2->score = highs = kp->score; /* only the score is copied! */
	}
	if (curHSP != kpp - &HSP[0]) {
		printf("%d going in; %d coming out.\n", curHSP, kpp-&HSP[0]);
		fflush(stdout);
	}
	curHSP = kpp - &HSP[0];

	++hasHSP;
	totHSP += curHSP;
	if (curHSP > 1)
		hsort((char *)HSP, curHSP, sizeof(HSP[0]), compar);

	/* add HSPs for this sequence to the cumulative list */
	HSP[curHSP-1]->next = firstHSP;
	for (i = curHSP-2; i >= 0; --i)
		HSP[i]->next = HSP[i+1];

	firstHSP = HSP[0];
	curHSP = 0;
}

writereport()
{
	register HSPPtr kp1, kp2;
	register int i;
	static int	maxHSPs;

	if (hasHSP > maxHSPs) {
		maxHSPs = hasHSP+1000;
		HSPlist = (HSPPtr *)ckrealloc(HSPlist, maxHSPs*sizeof(*HSPlist));
	}

	rh.seqlen = q_len;
	if (totHSP > UINT2_MAX) {
		printf("ERROR:  all %d diagonals could not be saved\n", totHSP);
		rh.diagcnt = UINT2_MAX;
	}
	else
		rh.diagcnt = totHSP;
	rh.lambda = lambda;
	rh.K = K;
	rh.lambda3 = lambda3;
	rh.K3 = K3;
	rh.lambdaS = lambdaS;
	rh.KS = KS;

	ie.offset = ftell(datafp);
	out_ie(indexfp, &ie);
	(void) fflush(indexfp);

	printf(">%s:  No. of saved diags = %d (%d overlaps removed)\n",
			ie.name, rh.diagcnt, overcnt);
	bcopy(ie.name, rh.name, sizeof(rh.name));
	out_rh(datafp, &rh);

	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    	kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		firstHSP = kp2;
	}

	/* print the matches for each sequence */
	for (i = hasHSP-1; i >= 0; --i)
		print_seq(i);
	(void) fflush(datafp);

	/* Store the maximum diagonal scores */
	fwrite(maxscores, curquery, sizeof(*maxscores), scorefp);
	fflush(scorefp);

	if (ferror(datafp))
		fatal("ERROR ON DATA FILE");

	sync();

	firstHSP = NULL;
	hasHSP = totHSP = overcnt = 0;
	printf("%d seqs done\n", curquery+1);
	fflush(stdout);
}

print_seq(i)
	int	i;
{
	register HSPPtr	kp;
	
	for (kp = HSPlist[i]; kp != NULL; kp = kp->next) {
		de.seqno = kp->seq_nbr;
		if (kp->score > UINT1_MAX)
			de.score = UINT1_MAX;
		else
			de.score = kp->score;
		de.diag = kp->s_pos - kp->q_pos;
		out_de(datafp, &de);
	}
	/* Free up the stored HSPs */
	kfree();
}

double fq[21] = {0, .081, .057, .045, .054, .015, .039, .061, .068, .022, .057,
 		  .093, .056, .025, .040, .049, .068, .058, .013, .032, .067};
double	prob[200];
double	fq1[AAID_MAX+1];
Uint4	iq1[DIM1(fq1)];

int
lambdak(lows, highs, q_seq, q_len)
	Int4	lows, highs;
	Uint4	q_len;
	register char	*q_seq;
{
	register int	i, j, k;
	register Uint4	range, ui;
	int		S;
	double	H;

	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		iq1[i]=0;
	for (ui=0; ui<q_len; ++ui)
		++iq1[q_seq[ui]];
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		fq1[i] = (double)iq1[i] / q_len;

	range = highs - lows;
	for (i=0; i<=range; ++i)
		prob[i]=0.;
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		for (j=AAID_MIN; j<DIM1(fq); ++j)
			prob[matrx[i][j]-lows] += fq1[i]*fq[j];
	if (karlin(lows, highs, prob, &lambda, &K, &H) == 0) {
		S = ceil(log(K*(double)totdblen*(double)q_len/300.)/lambda);
		if (S>42) S = 42;
	}
	else {
		lambda = -1.;
		K = -1.;
		S = DEFAULT_S;
		printf("karlin1 failed\n");
		fflush(stdout);
	}

	lows *= 3;
	highs *= 3;
	range = highs - lows;
	for (i=0; i<=range; ++i)
		prob[i]=0.;
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		for (j=AAID_MIN; j<DIM1(fq); ++j)
			for (k=AAID_MIN; k<DIM1(fq); ++k)
		prob[matrx[i][j]+matrx[i][k]+matrx[j][k]-lows] += fq1[i]*fq[j]*fq[k];
	if (karlin(lows, highs, prob, &lambda3, &K3, &H) != 0) {
		lambda3= -1.;
		K3 = -1.;
		printf(" karlin3 failed\n");
		fflush(stdout);
	}
	return S;
}

/* ----------------------------   utility routines   -------------------------*/
static void
sigterm()
{
	doquit = 1;
}

typedef struct kbuf {
		struct kbuf far	*chain;
		struct hsp	buf[1024];
	} KBUF;


KBUF	kb0, *kbp = &kb0;
HSPPtr	curbp = &kb0.buf[0], /* Ptr to beginning of current area */
		bp = &kb0.buf[0]; /* Ptr to last-allocated HSP in current area */

static struct hsp far *
kalloc()
{
	if (bp - curbp >= DIM1(kb0.buf)) {
		register KBUF	*nkp;
		if (kbp->chain == NULL) {
			nkp = (KBUF *)ckalloc(sizeof(KBUF));
			nkp->chain = NULL;
			kbp->chain = nkp;
			kbp = nkp;
		}
		else
			kbp = kbp->chain;
		curbp = bp = &kbp->buf[0];
	}
	return bp++;
}

kfree()
{
	kbp = &kb0;
	curbp = bp = &kb0.buf[0];
}

#define FREAD(a,b,c,d)	if (fread(a,b,c,d) != c) return 0
#define FWRITE(a,b,c,d)	if (fwrite(a,b,c,d) != c) return 0
	

static
out_rh(datafp, rh)
	register FILE	*datafp;
	register report_header	*rh;
{
	Int4	i;
	Uint2	ui;

	FWRITE(rh->name, sizeof(rh->name), 1, datafp);

	ui = htons(rh->seqlen);
	FWRITE((char *)&ui, sizeof(ui), 1, datafp);

	ui = htons(rh->diagcnt);
	FWRITE((char *)&ui, sizeof(ui), 1, datafp);

	i = rh->lambda*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);

	i = rh->K*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);

	i = rh->lambda3*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);

	i = rh->K3*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);

#ifdef foo
	i = rh->lambdaS*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);

	i = rh->KS*1000000000.;
	i = htonl(i);
	FWRITE((char *)&i, sizeof(i), 1, datafp);
#endif /* foo */

	if (!ferror(datafp))
		return 1;
	return 0;
}

static
in_rh(datafp, rh)
	register FILE	*datafp;
	register report_header	*rh;
{
	Int4	i;
	Uint2	ui;

	FREAD(rh->name, sizeof(rh->name), 1, datafp);

	FREAD((char *)&ui, sizeof(ui), 1, datafp);
	rh->seqlen = ntohs(ui);

	FREAD((char *)&ui, sizeof(ui), 1, datafp);
	rh->diagcnt = ntohs(ui);

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->lambda = ((double)i)/1000000000.;

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->K = ((double)i)/1000000000.;

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->lambda3 = ((double)i)/1000000000.;

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->K3 = ((double)i)/1000000000.;

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->lambdaS = ((double)i)/1000000000.;

	FREAD((char *)&i, sizeof(i), 1, datafp);
	rh->KS = ((double)i)/1000000000.;

	if (!ferror(datafp))
		return 1;
	return 0;
}


static
out_de(datafp, de)
	register FILE	*datafp;
	register diag_entry	*de;
{
	Uint2	us;
	Int2	is;

	us = htons(de->seqno);
	FWRITE((char *)&us, sizeof(us), 1, datafp);
	is = htons(de->diag);
	FWRITE((char *)&is, sizeof(is), 1, datafp);
	FWRITE((char *)&de->score, sizeof(de->score), 1, datafp);
#ifdef THREEWAY
	us = htons(de->diag3);
	FWRITE((char *)&us, sizeof(us), 1, datafp);
	FWRITE((char *)&de->score3, sizeof(de->score3), 1, datafp);
#endif
	if (!ferror(datafp))
		return 1;
	return 0;
}

static
in_de(datafp, de)
	register FILE	*datafp;
	register diag_entry	*de;
{
	Uint2	us;
	Int2	is;

	FREAD((char *)&us, sizeof(us), 1, datafp);
	de->seqno = ntohs(us);
	FREAD((char *)&is, sizeof(is), 1, datafp);
	de->diag = ntohs(is);
	FREAD((char *)&de->score, sizeof(de->score), 1, datafp);
#ifdef THREEWAY
	FREAD((char *)&us, sizeof(us), 1, datafp);
	de->diag3 = ntohs(us);
	FREAD((char *)&us, sizeof(us), 1, datafp);
	de->score3 = ntohs(us);
#endif
	if (!ferror(datafp))
		return 1;
	return 0;
}

static
out_ie(indexfp, ie)
	register FILE	*indexfp;
	register index_entry	*ie;
{
	Int4	i;

	FWRITE(ie->name, sizeof(ie->name), 1, indexfp);
	i = htonl(ie->offset);
	FWRITE((char *)&i, sizeof(i), 1, indexfp);
	fflush(indexfp);
	if (!ferror(indexfp))
		return 1;
	return 0;
}

static
in_ie(indexfp, ie)
	register FILE	*indexfp;
	register index_entry	*ie;
{
	Int4	i;

	if (feof(indexfp))
		return 0;
	FREAD(ie->name, sizeof(ie->name), 1, indexfp);
	FREAD((char *)&i, sizeof(i), 1, indexfp);
	ie->offset = ntohl(i);
	if (!ferror(indexfp))
		return 1;
	return 0;
}
