/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
BLASTP - Search a protein database for regions that approximately match
regions of a query sequence.  The basic command syntax is

 	BLASTP database query_sequence

where:
 	"database" names a file containing protein sequences, separated by
	header lines beginning with the character '>'.  A typical
	sequence entry is:

		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
		RADLIAYLKKATNE

	The database must be pre-processed with the SETDB program.

	"query_sequence" names a file containing a protein sequence (where the
	header line is optional).

The BLASTP command permits optional additional arguments, such as X=50,
that reset certain internal parameters.  The available parameters are:

	E or e is the _expected_ no. of HSPs to produce (which influences S)
		(default is 25).
	S or s gives the High-scoring Segment Pair (HSP) cutoff score
		(default is calculated from E).
	T or t gives the threshold for a word match (default is calculated,
		but is usually 17 or 18 with the PAM120 matrix).
	X or x gives the value for terminating word extensions (default is 20)
	M or m is the name of a file containing substitution costs matrix
		(default is "PAM120")
	Y or y is the effective length of the querY sequence for statistical
		significance calculations (default uses the actual query length)
	Z or z is the effective databaZe length for statistical significance
		calculations (default uses the actual database length)
	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer (i.e., SGI).

Thus a typical command is

	blastp aabank myseq e=0.5 x=25

Literature:

    Karlin & Altschul, "Methods for assessing the statistical significance
        of molecular sequence features by using general scoring schemes,"
        Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
    Altschul, Gish, Miller, Myers and Lipman, "Basic local alignment search
        tool," J. Mol. Biol. 214 (1990) in press.

 (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>

#define CKTIME	40	/* Check cpu time expended at this interval (in seconds) */
#define MAXTIME	1200	/* Maximum cpu time (in seconds) that can be expended */

#define EXTERN
#define BLASTP
#include "blast.h"
#include "aabet.h"	/* the alphabet */

/* default directory containing blastp databases */
#ifndef BLASTP_DIR
#define BLASTP_DIR	"/usr2/db"
#endif

#define HEADERLEN_MAX	200	/* maximum length of header line */
#define WORDSIZE_MAX	7	/* maximum length of critical word */

#define DEFAULT_E	25.	/* default Expected no. of HSPs */
#define DEFAULT_S	1	/* default cutoff for reporting HSPs */
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T	17	/* default threshold for neighborhood word hits */
#define DEFAULT_E2	0.1	/* secondary Expect, when at least 1 HSP is found */
#define DEFAULT_S2	0	/* secondary cutoff, when at least 1 HSP is found */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_W	4	/* default word size */

		/* Parameters for statistics */
#define DEFAULT_L	0.336
#define DEFAULT_K	0.179
static double	Lambda = DEFAULT_L,
				K = DEFAULT_K;

char
	header[HEADERLEN_MAX],
	*dbtitle,
	*qname,
#ifdef USE_SHM
	*dbbuf,			/* Memory-resident database */
#endif
	*s_seq0, *s_seq, *ss_seq,
	_q_seq[QUERYLEN_MAX+2],	/* null prefix + query sequence + null terminator */
	*q_seq = _q_seq+1,	/* query sequence + null terminator */
	qq_seq[QUERYLEN_MAX+1],	/* binary-encoded query sequence + null terminator */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

unsigned long
		q_len,		/* length of query sequence */
		totdblen,	/* total no. of residues in the database */
		Meff,		/* effective length of query sequence */
		Neff;		/* effective length of database */

int		s_len;		/* length of current database sequence */

Uint4
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

long	*score_hist;

SCORE
	score_hist_dim,
	highs_dummy,
	high_score,
	highs_exp,	/* expected high score */
	highs_obs;	/* observed high score */


unsigned long
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	S2 = DEFAULT_S2;	/* secondary cutoff score after at least 1 HSP found */

int
	S_set,
	S2_set,
	T = DEFAULT_T,		/* threshold for word hits */
	T_set,
	X = DEFAULT_X,		/* constant for word-hit extensions */
	minusX = -DEFAULT_X, /* -X */
	X_set,
	W = DEFAULT_W,		/* word size */
	W_set,
	W2 = DEFAULT_W,
	W2_set,
	T2 = DEFAULT_T,
	T2_set,
	X2 = DEFAULT_X,
	X2_set,
	E_set,
	E2_set,
	M_set;

double
	H,				/* Relative entropy of the p's and q's */
	E = DEFAULT_E,	/* Expected number of HSPs achieving cutoff score */
	E2 = DEFAULT_E2;

/* data structure to hold locations of matches for current database sequence */
DIAG_TYPE
	*_dlev,			/* flags for which diagonals... */
	*diag_level;	/* ...have been rigorously examined */

DIAG_PTR
	diag_stack,	/* list of diagonals with matches */
	diag_top;	/* next diagonal with a match */

HSPPtr	firstHSP = NULL;
HSPPtr	*HSPlist;

static int
	maxscore,		/* maximum achievable score */
	current;		/* number of current sequence */

static FILE
	*sfile;			/* file pointer for sequences */

static MFILE
	*hfile;			/* file pointer for headers */

static int
	highs,		/* highest score in the PAM matrix */
	lows;		/* lowest score in the PAM matrix */

static off_t
	nbr_seqs,		/* number of sequences in database */
	*seq_beg, 		/* starts of sequences in file or memory-resident dbase */
	*header_beg;	/* starts of headers */

DFA		*dp;	/* the automaton */

int		alphasize;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static signed char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

SCORE
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

static void	initdb PROTO((char far *dbname));
static void	dosearch PROTO(());
static int	get_seq PROTO((MPPROTO));
static void	search_aa PROTO((MPPROTO));
static void	get_pvals PROTO((MPPROTO));
static void	print_seq PROTO((HSPPtr hp));
static void	report PROTO((char far *s, char far *q,
							int spos, int qpos, int score, int len,
							double DBP_val));

#ifdef USE_MPROC
MPContext	mpc[PROC_MAX];
int		maxprocs, maxpprocs, numprocs = PROC_DEFAULT;
#else
int		numprocs = 1;
#endif

Uint4	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

#ifndef USE_MPROC
char	*version = "1.1.8 [6-Dec-90]";
#else
char	*version = "1.1.8MP [6-Dec-90]";
#endif
/*
Version	Date		Modification
-------	--------	----------------------------------------------------
1.1.0				Added REFINED_STATS code to modify Expect.
1.1.1				Added REFINED_STATS code to modify Poisson P-values.
1.1.2				Modified the *.tab file to include the dbtype.
1.1.3				Renamed *.tab file to *.atb, and *.hed to *.ahd.
1.1.4	11-21-90	Corrected the calculation of Poisson P-values.
1.1.5	11-24-90	When at least one HSP is found involving a given database
					sequence, it is scanned again with a lower cutoff score.
1.1.6	12-4-90		Fixed hit extension algorithm to halt when score <= 0
1.1.7	12-5-90		Added histogram of highest-scoring hit extensions
1.1.8	12-6-90		Moved many statistical calculations into library routines
*/

main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o, *a;
	int		i;
	Boolean	secondary;

	sys_times(&t0);

	banner(argv[0], version);

	if (argc > 3 && str_ncasecmp("-over", argv[argc-1], 5) == 0) {
		ovlap = ovlap_0;
		--argc;
	}

	if (argc < 3)
		fatal("Usage:  %s database query_sequence [options]\nValid %s options are E, S, T, E2, S2, X, M, W, Y, and Z.", module, module);
	while (--argc > 2) {
		secondary = FALSE;
		if (argv[argc][1] == '=')
			o = argv[argc][0];
		else
			if (argv[argc][0] == '-')
				o = argv[argc][1];
			else
				if (argv[argc][1] == '2' && argv[argc][2] == '=') {
					o = argv[argc][0];
					secondary = TRUE;
				}
				else
					fatal("Argument %d (%s) has improper form", argc+1, argv[argc]);
		a = argv[argc] + 2;
		if (secondary)
			++a;
		o = TOUPPER(o);
		if (!secondary)
		switch (o) {
			case 'P':
#ifdef USE_MPROC
				numprocs = atoi(a);
				if (numprocs <= 0)
					fatal("Bad value for P (the number of processors)");
				if (numprocs > PROC_MAX)
					fprintf(stderr, "\n%s:  Pmax = %d\n", module, PROC_MAX);
#else
				printf("NOTE:  Parameter P is invalid--\n       this program has not been compiled to use multiprocessing\n");
#endif
				break;
			case 'M':
				if (str_len(a) >= FILENAME_MAX)
					fatal("PAM file name is too long.");
				str_cpy(M, a);
				if (str_cmp(M, DEFAULT_M) != 0)
					M_set = 1;
				break;
			case 'W':
				if ((W = atoi(a)) <= 0)
					fatal("Bad value for W.");
				if (W > WORDSIZE_MAX)
					fatal("W is too big--must be less than %d.", WORDSIZE_MAX+1);
				W_set = 1;
				break;
			case 'S':
				if ((S = atol(a)) == 0 || S > 1000000)
					fatal("Bad value for S.");
				S_set = 1;
				break;
			case 'E':
				if (sscanf(a, "%lg", &E) != 1 || E <= 0.0 || E > 1000.0)
					fatal("Bad value for E.");
				E_set = 1;
				break;
			case 'T':
				if ((T = atoi(a)) < 1)
					fatal("Bad value for T.");
				T_set = 1;
				break;
			case 'X':
				if ((X = atoi(a)) < 1 || X > 100000)
					fatal("Bad value for X.");
				X_set = 1;
				minusX = -X;
				break;
			case 'Y':
				if ((Meff = atoi(a)) == 0)
					fatal("Bad value for Y.");
				break;
			case 'Z':
				if ((Neff = atoi(a)) == 0)
					fatal("Bad value for Z.");
				break;
			default:
				fatal("Valid options are E, S, T, E2, S2, X, M, W, Y, Z, and P.");
		}
		else
		/* Secondary search options */
		switch (o) {
			case 'E':
				if (sscanf(a, "%lg", &E2) != 1 || E2 > 1000.0)
					fatal("Bad value for E2.");
				E2_set = 1;
				break;
			case 'S':
				if ((S2 = atol(a)) == 0)
					fatal("Bad value for S2.");
				S2_set = 1;
				break;
			case 'T':
			case 'W':
			case 'X':
			default:
				fatal("Valid options are E, S, T, W, X, E2, S2, T2, W2, M, Y, Z, and P.");
		}
	}

	HSP = (struct hsp **)ckalloc(sizeof(struct hsp)*maxHSP);

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alphasize, &minchar, &maxchar, achars,
					order, &lows, &highs, maxcost, matrx);

	get_query(argv[2]);
	if (qname != NULL)
		(void) printf("Query:  %s\n", qname);

#ifdef CKTIME
	/* Poll the cpu time that has been consumed every CKTIME seconds */
	maxtime /= numprocs;
	cktime /= numprocs;
	(void) signal(SIGALRM, sigalrm);
	(void) alarm(cktime);
#endif

	initdb(argv[1]);

	(void) fflush(stdout);

#ifdef USE_MPROC
	(void) prctl(PR_SETEXITSIG, SIGTERM);

	maxprocs = prctl(PR_MAXPROCS);/* max. processes per user */
	maxpprocs = prctl(PR_MAXPPROCS); /* max. parallel processes */
	numprocs = MIN(numprocs, PROC_MAX);
	numprocs = MAX(1, numprocs);
	numprocs = MIN(numprocs, maxprocs);
	numprocs = MIN(numprocs, maxpprocs - PROC_RESERVE);
	numprocs = MAX(1, numprocs);
#endif
	
	/*
	It would be more accurate to call lambdak() for every sequence in
	the database, but then the program would run much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP containing part of a specific database
	sequence.  For the sake of speed, though, lambdak() is called only once.
	*/
	if (lambdak(lows, highs, q_seq, q_len, &Lambda, &K, &H) != 0)
		fatal("Could not calculate Lambda, K, and H");

	if (!T_set) {
		/* Ad hoc relationship for the PAM120 substitution matrix only */
		T = ceil((3.4 + 2.6*H)/Lambda - 0.5);
#ifdef USE_MPROC
		T -= 4; /* extra sensitivity when the processing power is available */
#endif
	}

	/* Determine the cutoff score, S */
	if (S_set && !E_set)
		E = 0.;
	cutoffs(&S, &E, &highs_exp, Lambda, K, H, Neff, Meff);

	/* Determine the secondary cutoff score, S2, to use */
	if ((E2_set && !S2_set && E2 == 0.) || (S2_set && S2 > S))
		E2 = 0., S2 = S;
	else {
		if (S2_set && !E2_set)
			E2 = 0.;
		/* Assume an "average" protein's length is 300 a.a. */
		cutoffs(&S2, &E2, &highs_dummy, Lambda, K, H, 300, 300);
		/* Adjust S2 to be in line with S, as necessary */
		S2 = MAX(S2, 1);
		if (S2 > S)
			S2 = S;
		E2 = stoe(S2, Lambda, K, H, 300, 300);
	}

	score_hist_dim = etos(0.01, Lambda, K, H, Neff, Meff) + 1;
	score_hist = (long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));

	/* The following relation is safe to use, because should the score
	ever reach S during hit extension, the entire diagonal will be
	searched rigorously ignoring X anyway.
	*/
	X = MIN(X, S);
	minusX = -X;

	maxscore = fmaxscore(q_seq, q_len);
	if (maxscore < S) {
		(void) printf("The maximum achievable score with the query sequence is only %d,\n",
				maxscore);
		(void) printf("whereas the cutoff score for reporting HSPs is %d.\n", S);
		(void) printf("You might lower the cutoff, use a different scoring scheme,\n");
		(void) printf("or increase the length of the query sequence.\n");
		exit(1);
	}

	sys_times(&t1);
	/* build table of critical (neighborhood) words */
	dp = bld_table_aa(q_seq, W, T, alphasize, matrx, maxcost, order,
						&numwords, &numexact);
	sys_times(&t2);

#ifdef USE_MPROC
    for (i=0; i<numprocs; ++i) {
        hspinit(&mpc[i]);
        initmpc(i, &mpc[i]);
    }
    m_set_procs(numprocs);
    m_fork(dosearch);
    m_kill_procs();
#else
    hspinit();
    dosearch();
#endif
 
#ifdef CKTIME
	(void) alarm(0);
#endif

	dfanstates = dfa_nstates(dp);
	dfastatesize = dfa_statesize(dp);
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

#ifdef USE_MPROC
	/* Collate the results from all processors */
    {
        register int    i, j;
        register struct hsp *hp;
 
        for (i=0; i<numprocs; ++i) {
            if (mpc[i].firstHSP != NULL) {
                if (firstHSP == NULL)
                    firstHSP = mpc[i].firstHSP;
                else {
                    /* Find the tail of the linked list */
                    hp = mpc[i].firstHSP;
                    while (hp->next != NULL)
                        hp = hp->next;
                    hp->next = firstHSP;
                    firstHSP = mpc[i].firstHSP;
                }
            }
            hasHSP += mpc[i].hasHSP;
            totHSP += mpc[i].totHSP;
            overcnt += mpc[i].overcnt;
            numhits += mpc[i].numhits;
            numfailed += mpc[i].numfailed;
			if (i > 0)
				for (j=0; j <= score_hist_dim; ++j)
					score_hist[j] += mpc[i].score_hist[j];
        }
    }
#endif

	sort_HSPs();

	sys_times(&t3);
 
	print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
					Meff, Neff, nbr_seqs);

	print_HSPs();

	print_parms();

	print_stats();

	exit(0);
	/*NOTREACHED*/
}

/* print_parms() -- print the parameters */
print_parms()
{
	(void) printf("\n\nParameters:\n");
	(void) printf("  E = %#0.2lg, S = %s, T = %d, X = %d, W = %d, M = %s\n",
		E, lu2s(S), T, X, W, M);
	(void) printf("  E2 = %#0.2lg, S2 = %s\n", E2, (S2 != S ? lu2s(S2) : "not used"));
}


/* print_stats() -- print the statistics of the search */
print_stats()
{
	(void) printf("\nStatistics:\n");
	(void) printf("  Lambda = %#0.3lg, Lambda/ln2 = %#0.3lg, K = %#0.3lg, H = %#0.3lg\n",
				Lambda, Lambda/log((double)2.), K, H);
	(void) printf("  Expected/Observed high score = %s/%s\n",
			ld2s(highs_exp), ld2s(highs_obs));
	if (S > highs_exp && highs_obs == 0)
		(void) printf("  NOTE:  the cutoff score is greater than the expected high score\n");
	if (highs_exp > maxscore)
		(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
				maxscore);
	(void) printf("  # of residues in query:  %s", lu2s(q_len));
	if (Meff != q_len)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		(void) putchar('\n');
	(void) printf("  # of neighborhood words in query:  %s\n", lu2s(numwords));
	(void) printf("  # of exact words scoring below T:  %s\n", lu2s(numexact));
	(void) printf("  Database:  %s\n", dbtitle);
	(void) printf("  # of residues in database:  %s", lu2s(totdblen));
	if (Neff != totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff));
	else
		(void) putchar('\n');
	(void) printf("  # of word hits against database:  %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions:  %s\n", lu2s((long)numfailed));
	(void) printf("  # of excluded hits:  %s\n",
				lu2s((long)numhits - (long)numfailed - totHSP - overcnt));
	(void) printf("  # of successful extensions:  %s\n", lu2s((long)totHSP+overcnt));
	(void) printf("  # of HSPs reported:  %s\n", lu2s((long)totHSP));
	(void) printf("  # of overlapping HSPs discarded:  %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of sequences in database:  %s\n", lu2s((long)nbr_seqs));
	(void) printf("  # of database sequences with at least one HSP:  %s\n",
				lu2s((long)hasHSP));

	(void) printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	(void) printf("Time to generate neighborhood:  ");
	sys_showtime(&t1, &t2);
#ifdef USE_MPROC
	(void) printf("No. of processors used:  %d\n", numprocs);
#endif
	(void) printf("Time to search database:  ");
	sys_showtime(&t2, &t3);
	(void) printf("Total cpu time:  ");
	sys_times(&t3);
	sys_showtime((Cputime_t *)NULL, &t3);
}

static void
dosearch()
{
	int		id = 0;
#ifdef USE_MPROC
	MPDEF

	id = m_get_myid();
	mp = &mpc[id];
	if (id == 0)
		signal(SIGTERM, sigterm);
#endif


	/* For each sequence in the database */
	while (get_seq(MP) == 0) {

		MPPTR high_score = 0;

		/* search the database sequence */
		search_aa(MP);

		/* keep track of the highest score */
		if (MPPTR high_score < score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];
		if (MPPTR high_score > highs_obs) {
			M_LOCK;
			if (MPPTR high_score > highs_obs)
				highs_obs = MPPTR high_score;
			M_UNLOCK;
		}

		/* Save any HSPs that were found */
		if (MPPTR curHSP > 0)
			get_pvals(MP);
	}
}

#ifdef USE_MPROC
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
    int     id;
    MPContextPtr    mp;
{    
    int     len;
    register int    i;
 
    mp->id = id;
 
    mp->current = mp->nbr_seqs = -1;
 
    len = maxlen + q_len + 1;
    mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
    mp->diag_level = mp->_dlev + q_len;
    mp->diag_top = mp->diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);
    mp->s_seq0 = ckalloc(maxlen+2);
    *mp->s_seq0++ = NULLB;
    mp->s_seq = mp->s_seq0;
	if (id == 0)
		mp->score_hist = score_hist;
	else
		mp->score_hist
				= (long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
}
#endif

/* get_query - read the query sequence */
get_query(filename)
char *filename;
{
	char	*fgets();
	register FILE	*qp;
	register char *p, *q, ch;
		
	if (str_cmp(filename, "-") == 0)
		qp = fdopen(0, "r"); /* read query sequence from stdin */
	else
		qp = ckopen(filename, "r");
	for (q = q_seq; ; ) {
		p = q;
		if (fgets(p, QUERYLEN_MAX + 1 - (p - q_seq), qp) == NULL) {
			(void) fclose(qp);
			break;
		}
		if (*q == '>')	/* Optional header/title line */
			if (q == q_seq) {
				qname = str_dup(q);
				continue;
			}
			else
				break;
		while ((ch = *p++) != NULLB) {
			if (aa_atob[ch] < AAID_MAX+1)
				*q++ = ch;
		}
		if (q - q_seq > QUERYLEN_MAX)
			fatal("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
	*q = NULLB;
	q_len = q - q_seq;
	if (q_len < W)
		fatal("query sequence is shorter than %d, the Word length.", W);
	if (Meff == 0)
		Meff = q_len;

	if (str_cnt(q_seq, "ACGTacgt")/(double)q_len > 0.85) {
		printf("\nNote: the query sequence appears to be nucleic acid when protein is expected.\n\n");
		fflush(stdout);
	}
	q = q_seq;
	while ((ch = *q) != NULLB)
		*q++ = aa_atob[ch];
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	MFILE far	*tfile;
	Int4	dbtype, dbformat;
	char	*dirname = CWDSTR;
	char	database[FILENAME_MAX];
	char	name[FILENAME_MAX];
	Int4	i;
	long	many, len, dbsize;
	int		once=0;

	current = -1;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(database, "%s%c%s", dirname, DIRDELIMCHR, dbname);
	else
		STRNCPY(database, dbname, sizeof(database));

	(void) sprintf(name, "%s.atb", database);
	tfile = mfil_open(name, "r");
	if (tfile == NULL) {
		if ((dirname = getenv("BLASTP")) == NULL)
			dirname = BLASTP_DIR;
		if (++once == 1)
			goto RETRY;
		fatal("couldn't open database file ``%s''", name);
	}
	if (mfil_dup(tfile, (char far *)&dbtype, sizeof(dbtype)) != sizeof(dbtype))
		fatal("Read error on file %s", name);
	if (dbtype != DB_TYPE_PRO)
		fatal("``%s'' is not a protein sequence database", dbname);
	if (mfil_dup(tfile, (char far *)&dbformat, sizeof(dbformat)) != sizeof(dbformat)
		|| dbformat != AAFORMAT)
			fatal("Database format is an incompatible version; rerun ``setdb''");

	if (mfil_dup(tfile, (char far *)&i, sizeof(i)) != sizeof(i)
		|| (dbtitle = mfil_get(tfile, i)) == NULL)
			fatal("Corrupted database file:  %s", name);
	if (i%4 != 0) {
		i = 4 - (i%4);
		mfil_seek(tfile, i, 1);
	}

	if (mfil_dup(tfile, (char far *)&nbr_seqs, sizeof(nbr_seqs)) != sizeof(nbr_seqs)
		|| mfil_dup(tfile, (char far *)&totdblen, sizeof(totdblen)) != sizeof(totdblen)
		|| mfil_dup(tfile, (char far *)&maxlen, sizeof(maxlen)) != sizeof(maxlen))
			fatal("corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) printf("Database:  %s\n", dbtitle);
	(void) printf("           %s sequences; %s total residues.\n",
			ld2s(nbr_seqs), ld2s(totdblen));

	ss_seq = ckalloc(sizeof(*ss_seq)*(maxlen+1));

	many = nbr_seqs + 1;
	seq_beg = (off_t far *)mfil_get(tfile, sizeof(*seq_beg)*many);
	if (seq_beg == NULL)
		fatal("improper sequence offsets in database");
	header_beg = (off_t far *)mfil_get(tfile, sizeof(*header_beg)*many);
	if (header_beg == NULL)
		fatal("improper header offsets in database");

	(void) sprintf(name, "%s.ahd", database);
	hfile = mfil_open(name, "r");
	if (hfile == NULL)
		fatal("Can't find database file %s", name);

	(void) sprintf(name, "%s.bsq", database);
#ifdef USE_SHM
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	if ((dbbuf = shm_attfile(name, "r")) == NULL)
#endif
	{
		sfile = fopen(name, "r");
		if (sfile == NULL)
			fatal("Couldn't open database file %s", name);
#ifndef USE_MPROC
		(void) fseek(sfile, (long)seq_beg[0], 0);
		s_seq = ckalloc(sizeof(*s_seq)*(maxlen+2));
		*s_seq++ = NULLB;
#else
		dbsize = sys_filesize(name);
		s_seq0 = ckalloc(sizeof(*s_seq0)*dbsize);
		if (fread((char *)s_seq0, (int)dbsize, 1, sfile) != 1)
			fatal("Database read error");
#endif /* USE_MPROC */
	}

	len = maxlen + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;

	diag_top = diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

#ifdef USE_MPROC
    if (numprocs > 1) {
        chunksize = nbr_seqs / (numprocs*50);
        if (numprocs < 4)
            chunksize /= 2;
        chunksize = MAX(chunksize, 1);
    }
    else
        chunksize = nbr_seqs;
#endif
}


/* get_seq - read a database sequence and header line, return nonzero on EOF */
static int
get_seq(MP)
	MPDEF
{
	if (++MPPTR current >= MPPTR nbr_seqs)
#ifndef USE_MPROC
		return 1;
#else
	{   /* Allocate a chunk of sequences for this processor to scan */
		MPPTR current = m_next() * chunksize;
		if (MPPTR current >= nbr_seqs)
			return 1;
		MPPTR nbr_seqs = MPPTR current + chunksize;
		MPPTR nbr_seqs = MIN(MPPTR nbr_seqs, nbr_seqs);
	}
#endif
#ifdef USE_SHM
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[MPPTR current+1] - seq_beg[MPPTR current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[MPPTR current];
	}
	else
#endif
	{
#ifndef USE_MPROC
		s_len = seq_beg[current+1] - seq_beg[current];
		if (fread(s_seq, s_len, 1, sfile) != 1)
			fatal("error reading database sequence");
		--s_len;
#else
		MPPTR s_len = seq_beg[MPPTR current+1] - seq_beg[MPPTR current] - 1;
		MPPTR s_seq = s_seq0 + seq_beg[MPPTR current];
#endif
	}
	if (MPPTR s_len > maxlen)
		fatal("Database sequence longer than maxlen! %d", MPPTR current);
	return 0;
}

/*****************************/
#include "searcha.inc"
/*****************************/

static int
hsp_score_cmp(h1, h2)
	HSPPtr	*h1, *h2;
{
	return (*h2)->score - (*h1)->score;
}

static void
get_pvals(MP)
	MPDEF
{
	register int i, j;
	register HSPPtr	hp;
	register double	Q, R, T;

	MPPTR totHSP += MPPTR curHSP;
	++MPPTR hasHSP;

	/* Sort HSPs by score, highest to lowest */
	if (MPPTR curHSP > 1)
		hsort((char far *)MPPTR HSP, MPPTR curHSP, sizeof(MPPTR HSP[0]), hsp_score_cmp);

	for (i = 0; i < MPPTR curHSP; ++i) {
		hp = MPPTR HSP[i];
		Q = stoe(hp->score, Lambda, K, H, MPPTR s_len, Meff);
		T = R = 1.0;
		for (j = 1; j <= i; ++j) {
			R *= (Q/j);
			T += R;
		}
		if (Q > 0.01)	/* Is limit approximation applicable? */
			hp->rawp.P_val = 1.0 - exp(-Q)*T;
		else
			hp->rawp.P_val = R * (Q/j);
	}

	/* add HSPs for this sequence to the cumulative list */
	MPPTR HSP[MPPTR curHSP-1]->next = MPPTR firstHSP;
	for (i = MPPTR curHSP-2; i >= 0; --i)
		MPPTR HSP[i]->next = MPPTR HSP[i+1];
	MPPTR firstHSP = MPPTR HSP[0];

	MPPTR curHSP = 0;
}

sort_HSPs()
{
	register HSPPtr kp1, kp2;
	register int i;
	register double	R, minp, minp2;
	double	xslen;

	if (hasHSP < maxHSP)
		HSPlist = HSP;
	else
		HSPlist = (HSPPtr *)ckalloc(sizeof(HSPPtr)*hasHSP);

	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		firstHSP = kp2;
	}

	/*
		For each sequence, calculate the P_val adjusted for the total
		length of the database (DBP_val); then find the smallest DBP_val
		of its HSPs
	 */
	for (i = 0; i < hasHSP; ++i) {
		kp1 = HSPlist[i];
		xslen = seq_beg[kp1->seq_nbr+1] - seq_beg[kp1->seq_nbr];
		R = Neff/xslen;
		minp = R * kp1->rawp.P_val;
		if (minp > 0.01) /* Is limit approximation applicable? */
			minp = (double)1.0 - exp(-minp);
		kp1->DBP_val = minp;
		for (kp2 = kp1->next; kp2 != NULL; kp2 = kp2->next) {
			minp2 = R * kp2->rawp.P_val;
			if (minp2 > 0.01)	/* Is limit approximation applicable? */
				minp2 = (double)1.0 - exp(-minp2);
			kp2->DBP_val = minp2;
			if (minp2 < minp)
				minp = minp2;
		}
		kp1->rawp.smallest_P = minp;
	}

	(void) hsort((char *)HSPlist, hasHSP, sizeof(HSPlist[0]), smaller);
}

print_HSPs()
{
	register int	i;

	/* print the matches for each sequence */
	for (i = 0; i < hasHSP; ++i)
		print_seq(HSPlist[i]);
}

static void
print_seq(hp)
	HSPPtr	hp;
{
	int		current, len, qpos, spos;
	HSPPtr	kp;
	register char	*cp;

	current = hp->seq_nbr;

	/* get header line */
	if (mfil_seek(hfile, header_beg[current], 0) == -1)
		fatal("fseek error on header file");
	len = header_beg[current+1] - header_beg[current];
	if (len > sizeof(header)-1)
		len = sizeof(header)-1;
	if (mfil_dup(hfile, header, len) != len)
		fatal("error reading header");
	header[len] = NULLB;
	(void) printf("\n\n%s\n", header);

	/* get sequence */
#ifdef USE_SHM
	if (dbbuf != NULL) {
		len = seq_beg[current+1] - seq_beg[current] - 1;
		s_seq = dbbuf + seq_beg[current];
	}
	else
#endif
	{
#ifndef USE_MPROC
		len = seq_beg[current+1] - seq_beg[current];
		if (fseek(sfile, seq_beg[current], 0) == -1)
			fatal("fseek error on sequence file");
		if (fread(s_seq, 1, len, sfile) != len)
			fatal("error reading sequence");
		--len;
#else
		s_seq = s_seq0 + seq_beg[current];
		len = seq_beg[current+1] - seq_beg[current] - 1;
#endif
	}
	(void) printf("\tLength = %s\n", ld2s((long)len));
	
	for (kp = hp; kp != NULL; kp = kp->next) {
		qpos = kp->q_pos;
		spos = kp->s_pos;
		len = kp->len;

		/* Extract the aligned subsequences from q_seq & s_seq */

		mem_cpy(qq_seq, q_seq + qpos, len);
		qq_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = qq_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		mem_cpy(ss_seq, s_seq + spos, len);
		ss_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = ss_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		report(ss_seq, qq_seq, spos, qpos, kp->score, len, kp->DBP_val);
		(void) fflush(stdout);
	}
}

/* report - print a match */
static void
report(s, q, spos, qpos, score, len, DBP_val)
	char	*s, *q;
	int		spos, qpos, score, len;
	double	DBP_val;
{
	double	Expect;
	int		numdone = 0;
	int		mlen;

	Expect = stoe(score, Lambda, K, H, Neff, Meff);

	(void) printf("\n  Score = %d, Expect = %#0.2lg, Poisson P = %#0.2lg, length = %s\n",
			score, Expect, DBP_val, ld2s(len));

	while (len > 0) {
		mlen = MIN(len, 60);
		(void) printf("\nQuery: %5d %.60s %d\n",
				qpos + numdone + 1, q, qpos + numdone + mlen);
		printmid(q, s, len);
		(void) printf("Sbjct: %5d %.60s %d\n",
				spos + numdone + 1, s, spos + numdone + mlen);
		q += 60;
		s += 60;
		len -= 60;
		numdone += 60;
	}
}

printmid(s1, s2, len)
	register char	*s1, *s2;
	register int	len;
{
	register char	c1, c2;

	(void) printf("%13s", "");
	if (len > 60)
		len = 60;
	while (len-- > 0) {
		if ((c1 = *s1++) == (c2 = *s2++) && c1 != 'X') {
			(void) putchar(c1);
			continue;
		}
		if (matrx[aa_atob[c1]][aa_atob[c2]] > 0)
			(void) putchar('+');
		else
			(void) putchar(' ');
	}
	(void) putchar('\n');
}

/* ----------------------------  parameter calculations  ---------------------*/
/*
	Amino acid residue frequencies used by S. Altschul.
*/
double fq[21] = {0.,
		.081, .057, .045, .054, .015, .039, .061, .068, .022, .057,
		.093, .056, .025, .040, .049, .068, .058, .013, .032, .067 };

/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, Lambda, K, H)
	Int4	lows, highs;
	Uint4	q_len;
	register char	*q_seq;
	double	*Lambda, *K, *H;
{
	double	fq1[AAID_MAX+1];
	Uint4	iq1[AAID_MAX+1];
	double	*prob;
	Uint4	range;
	register Uint4	ui;
	register int	i, j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM1(iq1); ++i)
		iq1[i]=0;
	for (ui=0; ui<q_len; ++ui)
		++iq1[q_seq[ui]];

	/* Convert frequencies to probabilities */
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		fq1[i] = (double)iq1[i] / q_len;

	/* Calculate probability of each score */
	range = highs - lows;
	prob = (double *)ckalloc(sizeof(*prob)*(range+1));
	for (ui=0; ui<=range; ++ui)
		prob[ui] = 0.;

	for (i=minchar; i<=maxchar; ++i)
		for (j=AAID_MIN; j<DIM1(fq); ++j)
			prob[matrx[i][j]-lows] += fq1[i]*fq[j];

	i = karlin(lows, highs, prob, Lambda, K, H);
	free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}


/* fmaxscore(seq, len) -- returns the maximum achievable score */
/* Note:  effect of parameter X is not taken into consideration here */
int
fmaxscore(seq, len)
	register char	*seq;
	register Uint4	len;
{
	register char	*maxseq;
	register int	sum, maxsum;

	maxseq = seq+len;
	sum = maxsum = 0;
	while (seq < maxseq) {
		sum += maxcost[*seq++];
		if (sum < 0)
			sum = 0;
		else
			if (sum > maxsum)
				maxsum = sum;
	}
	return maxsum;
}
