/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
* BLASTX - Search a PROTEIN database for regions that approximately match
* regions of a TRANSLATED NUCLEOTIDE query sequence.  The basic command
* syntax is:
*
* 	BLASTX database query_sequence [parameters] [[top][bottom]]
*
* where:
* 	database names a file containing protein sequences, separated by
* 		header lines beginning with the character '>'.  A typical
* 		sequence entry is:
*
*		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
*		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
*		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
*		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
*		RADLIAYLKKATNE
*
* 	query_sequence names a file containing a nucleotide sequence (where the
* 		header line is optional).
* 
* The BLASTX command permits optional additional arguments, such as X=50,
* that reset certain internal parameters.  The available parameters are:
*
*	E or e is the expected number of HSPs
*	S or s gives the High-scoring Segment Pair (HSP) cutoff score (default 45)
*	M or m is the name of a file containing substitution costs
*		(default is "PAM120")
*	T or t gives the threshold for a word match (default is 17)
*	X or x gives the value for terminating word extensions (default is 20)
* Thus a typical command is
*
*	BLASTX AABANK SEQ E=50 M=PAM250
*
*	The "top" and "bottom" options must be specified last on the
*	command line, and direct BLASTX to translate only the top or
*	only the bottom strand, respectively.
*
*   (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>

#define BLASTX
#define CKTIME	20	/* Check cpu time expended at this interval (in seconds) */
#define MAXTIME	1800	/* Maximum cpu time (in seconds) that can be expended */

#define EXTERN
#include "blast.h"
#include "aabet.h"	/* the alphabet */
#include "gcodes.h"

/* default directory containing blastp databases */
#ifndef BLASTP_DIR
#define BLASTP_DIR	"/usr2/db"
#endif

#define HEADERLEN_MAX	200	/* maximum length of header line */
#define WORDSIZE_MAX	7	/* maximum length of critical word */

#define DEFAULT_C	0	/* default genetic code is Standard */
#define DEFAULT_E	25.	/* default Expected no. of HSPs */
#define DEFAULT_S	1	/* default cutoff for reporting HSPs */
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T	15	/* default threshold for neighborhood word hits */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_W	4	/* default word size */

		/* Parameters for statistics */
#define DEFAULT_L	0.336
#define DEFAULT_K	0.179
static double	Lambda = DEFAULT_L,
				K = DEFAULT_K;

char
	header[HEADERLEN_MAX],
	*qname,
	*dbtitle,
#ifdef USE_SHM
	*dbbuf,			/* Memory-resident database */
#endif
	*s_seq, *ss_seq,
	*qq_seq,
	*t_seq,				/* temporary sequence storage */
	n_seq[QUERYLEN_MAX+1],	/* nucleotide query sequence + null terminator */
	*q_seq,				/* query sequence + null terminator */
	*n_rev,				/* reverse complement of query sequence */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */


int		n_len;		/* length of nucleotide query sequence */
int		s_len;		/* length of current database sequence */

unsigned long
		q_len,		/* length of translated sequence */
		Meff,		/* effective length of the query sequence */
		Neff;		/* effective length of the database */

unsigned long	totdblen;	/* total no. of residues in the database */

unsigned long
	maxlen,		/* maximum length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

static int
	top=1,				/* TRUE ==> search top strand */
	bottom=1,			/* TRUE ==> search bottom strand */
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	S_set = FALSE,
	T = DEFAULT_T,		/* threshold for word hits */
	T_set = FALSE,
	X = DEFAULT_X,		/* constant for word-hit extensions */
	X_set = FALSE,
	W = DEFAULT_W,		/* word size */
	W_set = FALSE,
	C = DEFAULT_C,		/* genetic code */
	C_set = FALSE,
	E_set = FALSE,
	M_set = FALSE;

double
	H,					/* Relative entropy of the p's and q's */
	E = DEFAULT_E;		/* Expected number of HSPs achieving cutoff score */

seq_data	sd[6] = {
	{0,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.},
	{1,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.},
	{2,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.},
	{3,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.},
	{4,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.},
	{5,(char *)0, (DIAG_TYPE *)0, (DIAG_TYPE *)0, (DIAG_PTR)0, (DIAG_PTR)0,
			0, 0, 0., 0,0,0, 0,0, 0., 0.,0.,0.}
		},	*sdp;


HSPPtr	firstHSP = NULL;
HSPPtr	*HSPlist;

static int
	minframe=0,
	maxframe=6,
	current;		/* number of current sequence */

static FILE
	*sfile;			/* file pointer for sequences */

static MFILE
	*tfile,			/* file pointer to table */
	*hfile;			/* file pointer for headers */

static int
	frame,		/* current reading frame */
	highs,		/* highest score in the PAM matrx */
	lows;		/* lowest score in the PAM matrix */

static off_t
	nbr_seqs,		/* number of sequences in database */
	*seq_beg, 		/* starts of sequences in file or memory-resident dbase */
	*header_beg;	/* starts of headers */

DFA		*dp;	/* the automaton */

int		alpha;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static signed char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

SCORE
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

static void	bld_table PROTO(());
static void	print_seq PROTO((HSPPtr hp));
static void	search_x PROTO((MPPROTO));
static void	get_pvals PROTO((MPPROTO));
static void	report PROTO((char far *s, char far *q,
							int spos, int qpos, int score, int len,
							double DBP_val));
static void	initdb PROTO((char far *dbname));
static int	get_seq PROTO((MPPROTO));
static void	dosearch PROTO((MPPROTO));

unsigned long	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

#ifndef USE_MPROC
char	*version = "1.1.5 [24-Nov-90]";
#else
char	*version = "1.1.5MP [24-Nov-90]";
#endif

#ifdef USE_MPROC
MPContext	mpc[PROC_MAX];
int		maxprocs, maxpprocs, numprocs = PROC_DEFAULT;
#else
int		numprocs = 1;
#endif

main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o, *a;
	int		i;

	banner(argv[0], version);

	if (argc < 3)
		fatal("Usage:  %s database query_sequence [options]\nValid %s options are E, S, T, X, M, W, Y and Z.", module, module);

	if (str_casecmp(argv[argc-1], "top") == 0) {
		bottom=0;
		top=1;
		--argc;
	}
	if (str_casecmp(argv[argc-1], "bottom") == 0) {
		top=0;
		bottom=1;
		--argc;
	}
	minframe = top ? 0 : 3;
	maxframe = bottom ? 6 : 3;

	while (--argc > 2) {
		if (argv[argc][1] == '=')
			o = argv[argc][0];
		else
			if (argv[argc][0] == '-')
				o = argv[argc][1];
			else
				fatal("Argument %d (%s) has improper form", argc+1, argv[argc]);
		a = argv[argc] + 2;
		o = TOUPPER(o);
		switch (o) {
			case 'P':
#ifdef USE_MPROC
				numprocs = atoi(a);
				if (numprocs <= 0)
					fatal("Bad value for P (the number of processors)");
				if (numprocs > PROC_MAX)
					fprintf(stderr, "\n%s:  Pmax = %d\n", module, PROC_MAX);
#else
				fprintf(stderr,
					"\n%s:  Parameter P is ignored--this program\n       has not been compiled to use multiprocessing\n", module);
#endif
				break;
			case 'M':
				if (str_len(a) >= FILENAME_MAX)
					fatal("PAM file name is too long.");
				str_cpy(M, a);
				if (str_cmp(M, DEFAULT_M) != 0)
					M_set = TRUE;
				break;
			case 'W':
				if (sscanf(a, "%d", &W) != 1 || W <= 0)
					fatal("Bad value for W.");
				if (W > WORDSIZE_MAX)
					fatal("W is too big--must be less than %d.", WORDSIZE_MAX+1);
				W_set = TRUE;
				break;
			case 'S':
				if (sscanf(a, "%d", &S) != 1 || S < 0)
					fatal("Bad value for S.");
				S_set = TRUE;
				break;
			case 'E':
			case 'e':
				if (sscanf(a, "%lg", &E) != 1 || E <= 0. || E > 1000.)
					fatal("Bad value for E.");
				E_set = TRUE;
				break;
			case 'T':
				if (sscanf(a, "%d", &T) != 1 || T < 1)
					fatal("Bad value for T.");
				T_set = TRUE;
				break;
			case 'X':
				if (sscanf(a, "%d", &X) != 1 || X < 1 || X > 1000000)
					fatal("Bad value for X.");
				X_set = TRUE;
				break;
			case 'Y':
				if (sscanf(a, "%d", &Meff) != 1 || Meff == 0)
					fatal("Bad value for Y.");
				break;
			case 'Z':
				if (sscanf(a, "%d", &Neff) != 1 || Neff == 0)
					fatal("Bad value for Z.");
				break;
			case 'C':
				if (isalpha(*a)) {
					display_gcodes();
					exit(0);
				}
				C = atoi(a);
				if (C < 0 || C >= DIM1(gcodes)) {
					printf("Bad value for C\n");
					display_gcodes();
					exit(0);
				}
				C_set = TRUE;
				break;
			default:
				fatal("Valid options are E, S, T, X, M, C, W, Y and Z.");
		}
	}

	HSP = (struct hsp **)ckalloc(sizeof(struct hsp)*maxHSP);

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alpha, &minchar, &maxchar, achars,
					order, &lows, &highs, maxcost, matrx);

	get_query(argv[2]);
	if (qname != NULL)
		(void) printf("Query:  %s\n", qname);
	(void) printf("  Translating ");
	if (top && bottom)
		(void) printf("both strands in all 6 reading frames\n");
	if (top && !bottom)
		(void) printf("top strand in 3 reading frames\n");
	if (!top && bottom)
		(void) printf("bottom strand in 3 reading frames\n");

#ifdef USE_MPROC
	(void) prctl(PR_SETEXITSIG, SIGTERM);

	maxprocs = prctl(PR_MAXPROCS);
	maxpprocs = prctl(PR_MAXPPROCS);
	numprocs = MIN(numprocs, PROC_MAX);
	numprocs = MAX(1, numprocs);
	numprocs = MIN(numprocs, maxprocs);
	numprocs = MIN(numprocs, maxpprocs - PROC_RESERVE);
	numprocs = MAX(1, numprocs);
#endif

#ifdef CKTIME
	/* Poll the cpu time that has been consumed every CKTIME seconds */
	cktime /= numprocs;
	maxtime /= numprocs;
	(void) signal(SIGALRM, sigalrm);
	(void) alarm(cktime);
#endif

	sys_times(&t0);

	initdb(argv[1]);
	(void) fflush(stdout);

	xlat_query(); /* must come after the call to initdb() so dbsize is known */

	/*
	It might be better to call lambdak() for every sequence in
	the database, but then the program would be much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP.  But for speed, lambdak() is
	called only once.
	*/
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		if (lambdak(lows, highs, sdp->seq, sdp->len,
				&sdp->Lambda, &sdp->K, &sdp->H) != 0)
			fatal("Could not calculate Lambda, K, and H in frame %d",
						(frame < 3 ? frame+1 : 2-frame));
		if (!T_set) {
			/* Ad hoc relationship for the PAM120 substitition matrix only */
			sdp->T = ceil((3.4 + 2.6*sdp->H)/sdp->Lambda - 0.5);
#ifdef USE_MPROC
			sdp->T -= 4; /* extra sensitive when processing power available */
#endif
		}
		else
			sdp->T = T;
		/* Determine the cutoff score, S */
		if (S_set && !E_set)
			E = 0.;
		sdp->E = E / (top && bottom ? 6. : 3.);
		sdp->S = S;
		cutoffs(&sdp->S, &sdp->E, &sdp->highs_exp, sdp->Lambda, sdp->K, sdp->H,
						Neff, (Meff != 0 ? Meff : sdp->len));

		/* The following relation is safe to use, because should the score
		ever reach S during hit extension, the entire diagonal will be
		rigorously searched ignoring X.
		*/
		sdp->X = MIN(X, sdp->S);
		sdp->minusX = -sdp->X;
	}

	E = 0.;
	for (frame=0; frame<6; ++frame)
		E += sd[frame].E;

	i=0;
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		sdp->maxscore = fmaxscore(sdp->seq, sdp->len);
		if (sdp->maxscore >= sdp->S)
			i = 1;
	}
	if (i == 0) {
		(void) printf("\nThe maximum achievable scores are insufficient to satisfy the cutoff scores\n");
		(void) printf("\nThe max. achievable scores are:\n");
		(void) printf("\tFrame\tMax. Score\n");
		for (frame=minframe; frame<maxframe; ++frame)
			(void) printf("\t  %c%d\t   %d\n",
				(frame < 3 ? '+' : '-'),
				(frame < 3 ? frame+1 : frame-2), sd[frame].maxscore);
		(void) printf("\nYou might lower the cutoff, use a different scoring scheme,\n");
		(void) printf("or use a longer query sequence.\n");
		exit(1);
	}

#ifdef USE_MPROC
	{
		int		i, frame, len;
		seq_data	*sdp;

		for (i=0; i<numprocs; ++i)
			mem_cpy((char *)&mpc[i].sd[0], (char *)&sd[0], sizeof(sd));
		for (i=1; i<numprocs; ++i) {
			for (frame=minframe; frame<maxframe; ++frame) {
				sdp = &mpc[i].sd[frame];
				len = maxlen + sdp->len + 1;
				sdp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(sdp->_dlev[0])*len);
				sdp->diag_top = sdp->diag_stack =
					(DIAG_PTR)ckalloc(sizeof(sdp->diag_top[0])*len);
				sdp->diag_level = sdp->_dlev + sdp->len;
			}
		}
	}
#endif

	sys_times(&t1);

	dp = dfa_open(NULL, 0,
				(LETTER *)AAID_MIN, AAID_MAX-AAID_MIN+1, WORDSIZE_MAX,
				ckalloc, 50*KBYTE, NULL, (DFA_HALTNONALPHAS|DFA_NOXLTAB));
	if (dp == NULL) {
		dfa_perror(dp, module);
		fatal("dfa_open was gagged with a spoon");
	}

	/* build table of critical (neighborhood) words */
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		q_seq = sdp->seq;
		q_len = sdp->len;
		T = sdp->T;
		bld_table();
	}

	if (dfa_close(dp) != 0) {
		dfa_perror(dp, module);
		fatal("dfa_close phoned home");
	}

	sys_times(&t2);

#ifdef USE_MPROC
	m_set_procs(numprocs);
	m_fork(dosearch);
	m_kill_procs();
#else
	dosearch();
#endif

#ifdef CKTIME
	(void) alarm(0);
#endif

	dfanstates = dfa_nstates(dp);
	dfastatesize = dfa_statesize(dp);
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

#ifdef USE_MPROC
	/* Collate the results from all processors */
	{
		register int	i;
		register struct hsp	*hp;

		for (i=0; i<numprocs; ++i) {
			if (mpc[i].firstHSP != NULL) {
				if (firstHSP == NULL)
					firstHSP = mpc[i].firstHSP;
				else {
					/* Find the tail of the linked list */
					hp = mpc[i].firstHSP;
					while (hp->next != NULL)
						hp = hp->next;
					hp->next = firstHSP;
					firstHSP = mpc[i].firstHSP;
				}
			}
			hasHSP += mpc[i].hasHSP;
			totHSP += mpc[i].totHSP;
			overcnt += mpc[i].overcnt;
			numhits += mpc[i].numhits;
			numfailed += mpc[i].numfailed;
		}
	}
#endif

	sys_times(&t3);

	print_HSPs();

	print_parms();

	print_stats();

	exit(0);
	/*NOTREACHED*/
}

/* print_parms() -- print the parameters */
print_parms()
{
	int		f, frame;

	(void) printf("\n\nParameters:\n");
	(void) printf("  E = %#0.3lg, X = %d, W = %d, M = %s\n",
		E, X, W, M);
	(void) printf("  C = %d (%s genetic code)\n\n", C, gcodes[C].name);
	(void) printf("  Frame\tLength\t E\t S\t T\n");
	for (f=minframe; f<maxframe; ++f) {
		frame = (f<3 ? 2-f : f);
		sdp = &sd[frame];
		(void) printf("   %s%d",
				(frame<3? "+":""), (frame < 3 ? frame+1 : 2-frame));
		(void) printf("\t %s\t%#0.2lg\t%d\t%d\n",
					ld2s(sdp->len), sdp->E, sdp->S, sdp->T);
	}
}

/* print_stats() -- print the statistics of the search */
print_stats()
{
	seq_data	*sdp;
	int		f, frame;

	(void) printf("\nStatistics:\n");
	(void) printf("  Frame\tLambda\tLambda/ln2\tK\tH\tExpHigh\tObsHigh\n");
	for (f=minframe; f<maxframe; ++f) {
		frame = (f<3 ? 2-f : f);
		sdp = &sd[frame];
		(void) printf("   %s%d\t%#0.3lg\t%#0.3lg\t\t%#0.3lg\t%#0.3lg\t%6s\t%6s\n",
				(frame<3? "+":""), (frame < 3 ? frame+1 : 2-frame),
				sdp->Lambda, sdp->Lambda/log(2.), sdp->K, sdp->H,
			ld2s((long)sdp->highs_exp), ld2s((long)sdp->highs_obs));
		if (sdp->S > sdp->highs_exp && sdp->highs_obs <= 0)
			(void) printf("  NOTE:  the cutoff score is greater than the expected high score\n");
		if (sdp->highs_exp > sdp->maxscore)
			(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
					sdp->maxscore);
	}

	(void) printf("  # of nucleotides in query = %s", ld2s((long)n_len));
	if (Meff != 0)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		putchar('\n');
	(void) printf("  # of neighborhood words in query = %s\n", lu2s((long)numwords));
	(void) printf("  # of exact words scoring below T = %s\n", lu2s((long)numexact));
	(void) printf("  Database:  %s\n", dbtitle);
	(void) printf("  # of residues in database = %s aa", lu2s((long)totdblen));
	if (Neff != totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff));
	else
		putchar('\n');
	(void) printf("  # of word hits against database = %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions = %s\n", lu2s((long)numfailed));
	(void) printf("  # of excluded hits = %s\n",
				lu2s((long)numhits - (long)numfailed - totHSP - overcnt));
	(void) printf("  # of successful extensions = %s\n", lu2s((long)totHSP+overcnt));
	(void) printf("  # of HSPs reported = %s\n", lu2s((long)totHSP));
	(void) printf("  # of overlapping HSPs discarded = %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of sequences in database = %s\n", lu2s((long)nbr_seqs));
	(void) printf("  # of database sequences with at least one HSP = %s\n",
				lu2s((long)hasHSP));

	(void) printf("No. of states in DFA:  %s (%s KB)\n", lu2s(dfanstates),
				lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
				lu2s(HOWMANY(dfasize, KBYTE)),
				lu2s(HOWMANY(dfaextent, KBYTE)));
	(void) printf("Time to generate neighborhoods:  ");
	sys_showtime(&t1, &t2);
#ifdef USE_MPROC
	(void) printf("No. of processors used:  %d\n", numprocs);
#endif
	(void) printf("Time to search database:  ");
	sys_showtime(&t2, &t3);
	(void) printf("Total cpu time:  ");
	sys_times(&t3);
	sys_showtime((Cputime_t *)NULL, &t3);
	exit(0);
}

static void
dosearch()
{
	int		id = 0;

#ifdef USE_MPROC
	MPDEF

	id = m_get_myid();
	mp = &mpc[id];
	mp->id = id;
	mp->current = -1 - id;
	mp->nbr_seqs = INT_MIN;
	mp = &mpc[id];
	if (id == 0)
		signal(SIGTERM, sigterm);
	hspinit(MP);
#endif

	while (get_seq(MP) == 0) {
		search_x(MP);
		if (MPPTR curHSP > 0)
			get_pvals(MP);
	}
}


display_gcodes()
{                
    register int    i;
                       
    printf("The available genetic codes are:\n");
    for (i=0; i<DIM1(gcodes); ++i)                
        printf("  %d. %s\n", i, gcodes[i].name);
    printf("Specify a particular code by its number (0 is the default)\n");
}

/* get_query - read the query sequence */
get_query(filename)
	char	*filename;
{
	char	*fgets();
	register FILE	*qp;
	register char *p, *q, ch;
		
	if (strcmp(filename, "-") == 0)
		qp = fdopen(0, "r"); /* read query sequence from stdin */
	else
		qp = ckopen(filename, "r");
	for (q = n_seq; ; ) {
		p = q;
		if (fgets(p, QUERYLEN_MAX + 1 - (p - n_seq), qp) == NULL) {
			(void) fclose(qp);
			break;
		}
		if (*q == '>')	/* Optional header/title line */
			if (q == n_seq) {
				qname = str_dup(q);
				continue;
			}
			else
				break;
		while ((ch = *p++) != NULLB) {
			if (isalpha(ch))
				*q++ = TOUPPER(ch);
		}
		if (q - n_seq > QUERYLEN_MAX)
			fatal("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
	*q = NULLB;
	n_len = q - n_seq;
	qq_seq = ckalloc(n_len/3+1);

	if (str_cnt(n_seq, "ACGT")/(double)n_len < 0.85) {
		printf("\nNote:  the query sequence appears to be protein when nucleic acid is expected.\n\n");
		fflush(stdout);
	}
}

/* nucleotide sequence --> amino acid sequence translation routines */

static unsigned char	aatab[64];	/* translation table */
static unsigned char	nttab[128];

xlat_query()
{
	char	*revcomp();
	int		len;

	initgcode(&gcodes[C], aatab);

	nttab['T'] = nttab['t'] = nttab['U'] = nttab['u'] = 0;
	nttab['C'] = nttab['c'] = 1;
	nttab['A'] = nttab['a'] = 2;
	nttab['G'] = nttab['g'] = 3;

	for (frame=minframe; frame<3; ++frame) {
		sdp = &sd[frame];
		xlat_seq(n_seq+frame, n_len-frame, &sdp->seq, &sdp->len);
	}
	if (bottom)
		n_rev = revcomp(n_seq, n_len);
	for (frame=3; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		xlat_seq(n_rev+frame-3, n_len-frame+3, &sdp->seq, &sdp->len);
	}
	for (frame=minframe; frame<maxframe; ++frame) {
		sdp = &sd[frame];
		sdp->elen = (Meff != 0 ? Meff : sdp->len);
		sdp->xlen = sdp->elen;
		len = maxlen + sdp->len + 1;
		sdp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(sdp->_dlev[0])*len);
		sdp->diag_top = sdp->diag_stack =
					(DIAG_PTR)ckalloc(sizeof(sdp->diag_top[0])*len);
		sdp->diag_level = sdp->_dlev + sdp->len;
	}
}

xlat_seq(nseq, nlen, aseq, alen)
	register char	*nseq;
	int		nlen;
	char	**aseq;
	int		*alen;
{
	register char	*cp;
	register int	len = nlen/3;
	register int	i;

	if (len < 0)
		len = 0;

	cp = *aseq = ckalloc(len + 2) + 1;
	cp[-1] = NULLB; /* null prefix byte */

	while (nlen > 2) {
		nlen -= 3;
		i = nttab[*nseq++] * 4*4;
		i += nttab[*nseq++] * 4;
		i += nttab[*nseq++];
		*cp++ = aa_atob[aatab[i]];
	}
	*cp = NULLB;
	*alen = cp - *aseq;
}

/* Initialize the translation table */
initgcode(gp, xltab)
    GeneticCodePtr	gp;
    register unsigned char   *xltab;
{
	register char	*code;
    register int	i, j, k, tot;

	code = gp->code;

	for (i=0; i<4; ++i) {
		for (j=0; j<4; ++j) {
			for (k=0; k<4; ++k) {
				tot = i*4*4 + j*4 + k;
				xltab[tot] = *code++;
			}
		}
	}
}

static char	revtab[128];

char *
revcomp(s, len)
	register char	*s;
	register int	len;
{
	register char	*t;

	revtab['A'] = 'T';
	revtab['C'] = 'G';
	revtab['G'] = 'C';
	revtab['T'] = revtab['U'] = 'A';

	t = ckalloc(len+1);
	t[len] = NULLB;
	while (len-- > 0)
		t[len] = revtab[*s++];
	return t;
}

/* UNIX-specific input routine; database must be pre-processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	Int4	dbtype, dbformat, i;
	char	*dirname = CWDSTR;
	char	database[FILENAME_MAX];
	char	name[FILENAME_MAX];
	long	many, len;
	int		once=0;

	current = -1;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(database, "%s%c%s", dirname, DIRDELIMCHR, dbname);
	else
		STRNCPY(database, dbname, sizeof(database));

	(void) sprintf(name, "%s.atb", database);
	tfile = mfil_open(name, "r");
	if (tfile == NULL) {
		if ((dirname = getenv("BLASTP")) == NULL)
			dirname = BLASTP_DIR;
		if (++once == 1)
			goto RETRY;
		fatal("couldn't open database file ``%s''", name);
	}
	if (mfil_dup(tfile, (char far *)&dbtype, sizeof(dbtype)) != sizeof(dbtype))
		fatal("Database read error on file %s", name);
	if (dbtype != DB_TYPE_PRO)
		fatal("``%s'' is not a protein sequence database", dbname);
	if (mfil_dup(tfile, (char *)&dbformat, sizeof(dbformat)) != sizeof(dbformat) ||
			dbformat != AAFORMAT)
		fatal("Database format is an incompatible version; re-run ``setdb''");

	if (mfil_dup(tfile, (char far *)&i, sizeof(i)) != sizeof(i)
		|| (dbtitle = mfil_get(tfile, i)) == NULL)              
			fatal("Corrupted database file:  %s", name);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, 1);
	}

	if (mfil_dup(tfile, (char *)&nbr_seqs, sizeof(nbr_seqs)) != sizeof(nbr_seqs)
		|| mfil_dup(tfile, (char *)&totdblen, sizeof(totdblen)) != sizeof(totdblen)
		|| mfil_dup(tfile, (char *)&maxlen, sizeof(maxlen)) != sizeof(maxlen))
		fatal("corrupted database file:  %s", name);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) printf("Database:  %s\n", dbtitle);
	(void) printf("           %s sequences; %s total residues.\n",
		ld2s(nbr_seqs), ld2s(totdblen));                       
 
	many = nbr_seqs + 1;
	ss_seq = ckalloc(sizeof(*ss_seq)*(maxlen+1));

	seq_beg = (off_t far *)mfil_get(tfile, sizeof(*seq_beg)*many);
	if (seq_beg == NULL)
		fatal("improper sequence offsets in database");
	header_beg = (off_t far *)mfil_get(tfile, sizeof(*header_beg)*many);
	if (header_beg == NULL)
		fatal("improper header offsets in database");

	(void) sprintf(name, "%s.ahd", database);
	hfile = mfil_open(name, "r");

	(void) sprintf(name, "%s.bsq", database);
#ifdef USE_SHM
	/* Attach a shared memory segment for the .bsq file (if one exists) */
	if ((dbbuf = shm_attfile(name, "r")) == NULL)
#endif /* USE_SHM */
	{
		off_t	fsize;
		sfile = ckopen(name, "r");
#ifdef USE_MPROC
		dbbuf = ckalloc(fsize = sys_filesize(name));
		if (fread(dbbuf, fsize, 1, sfile) != 1)
			fatal("Error reading database file %s", name);
#else
		(void) fseek(sfile, (long)seq_beg[0], 0);
		s_seq = ckalloc(sizeof(*s_seq)*(maxlen+2));
		*s_seq++ = NULLB;
#endif
	}

	len = maxlen + q_len + 1;

	if (Neff == 0)
		Neff = totdblen;

#ifdef USE_MPROC
	if (numprocs > 1) {
		chunksize = nbr_seqs / (numprocs*50);
		if (numprocs < 4)
			chunksize /= 2;
		chunksize = MAX(chunksize, 1);
	}
	else
		chunksize = nbr_seqs;
#endif
}

/* get_seq - read a database sequence and header line, return 0 on EOF */
static int
get_seq(MP)
	MPDEF
{
	if (++MPPTR current >= MPPTR nbr_seqs)
#ifndef USE_MPROC
		return 1;
#else
	{	/* Allocate a chunk of sequences for this processor to scan */
		MPPTR current = m_next() * chunksize;
		if (MPPTR current >= nbr_seqs)
			return 1;
		MPPTR nbr_seqs = MPPTR current + chunksize;
		MPPTR nbr_seqs = MIN(MPPTR nbr_seqs, nbr_seqs);
	}
#endif
#if defined(USE_SHM) || defined(USE_MPROC)
	if (dbbuf != NULL) {
		MPPTR s_len = seq_beg[MPPTR current+1] - seq_beg[MPPTR current] - 1;
		MPPTR s_seq = dbbuf + seq_beg[MPPTR current];
	}
	else
#endif
	{
		s_len = seq_beg[current+1] - seq_beg[current];
		if (fread(s_seq, 1, s_len, sfile) != s_len)
			fatal("error reading sequence");
		--s_len;
	}
	if (MPPTR s_len > maxlen)
		fatal("Database sequence longer than expected maxlen!");
	return 0;
}

/* -----------------------   build table of critical words   -----------------*/

signed char
	wword[WORDSIZE_MAX];	/* candidate for critical word */

SCORE
	best[WORDSIZE_MAX];	/* best[i] = smallest score of wword[0]..wword[i]
				   that can be extended to a critical word */
char	*query;			/* pointer into query sequence */

static void
bld_table()
{
	register int i;
	register int	wm1 = W-1;
	register int	score;

	for (score=i=0; i<wm1; ++i)
		score += matrx[q_seq[i]][q_seq[i]];

	for (query = q_seq; query[wm1] != NULLB; query++) {
		best[wm1] = T;
		for (i = wm1; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);

		/* Check the score of a perfect match */
		score += matrx[query[wm1]][query[wm1]];
		if (score < T && score > 0) {
			/* Add perfect match even if its score is less than T */
			++numexact;
			if (dfa_add(dp, (LETTER *)query, W, (mempint)((query-q_seq+W)<<3)+frame) == NULL)
				fatal("error in dfa_add");
		}
		score -= matrx[query[0]][query[0]];
	}
	if (numwords + numexact == 0)
		fatal("no satisfying words in query sequence to compare against database");
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for wword[0]..wword[i-1] */
enumerate(i, score)
	register int		i;
	register int	score;
{
	/* possible letters by decreasing weight */
	register signed char	*o = order[query[i]];
	register int	c;
	register int	new_score;

	for (c = 0; c < alpha; ++c) {
		new_score = score + matrx[query[i]][wword[i] = o[c]];
		if (new_score < best[i])
			break;
		if (i < W-1)
			enumerate(i+1, new_score);
		else {
			++numwords;
			if (dfa_add(dp, (LETTER *)wword, W, (mempint)((query-q_seq+W)<<3)+frame) == NULL)
				fatal("error in dfa_add");
		}
	}
}

/***************************/
#include "searchx.inc"
/***************************/

static int
hsp_score_cmp(h1, h2)
	HSPPtr	*h1, *h2;
{
	return (*h2)->score - (*h1)->score;
}

static void
get_pvals(MP)
	MPDEF
{
	register int i, j;
	register HSPPtr hp;
	register double Q, R, T;
	double	C;

	++MPPTR hasHSP;
	MPPTR totHSP += MPPTR curHSP;
	/* Sort HSPs by score, highest to lowest */
	if (MPPTR curHSP > 1)
		hsort((char far *)MPPTR HSP, MPPTR curHSP, sizeof(MPPTR HSP[0]), hsp_score_cmp);

	for (i = 0; i < MPPTR curHSP; ++i) {
		hp = MPPTR HSP[i];
		sdp = &sd[hp->frame];
		Q = stoe(hp->score, sdp->Lambda, sdp->K, sdp->H, MPPTR s_len, sdp->elen);
		T = R = 1.0;
		for (j = 1; j <= i; ++j) {
			R *= (Q/j);
			T += R;
		}
		if (Q > 0.01)	/* Is limit approximation applicable? */
			hp->rawp.P_val = 1.0 - exp(-Q)*T;
		else
			hp->rawp.P_val = R * (Q/j);
	}

	/* add HSPs for this sequence to the cumulative list */
	MPPTR HSP[MPPTR curHSP-1]->next = MPPTR firstHSP;
	for (i = MPPTR curHSP-2; i >= 0; --i)
		MPPTR HSP[i]->next = MPPTR HSP[i+1];
	MPPTR firstHSP = MPPTR HSP[0];

	MPPTR curHSP = 0;
}

print_HSPs()
{
	register HSPPtr kp1, kp2;
	register int i;
	register double	R, minp, minp2;
	double	xslen;

	if (hasHSP < maxHSP)
		HSPlist = HSP;
	else
		HSPlist = (HSPPtr *)ckalloc(sizeof(HSPPtr)*hasHSP);

	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		firstHSP = kp2;
	}

	/*
		For each sequence, calculate the P_val adjusted for the total
		length of the database (DBP_val); then find the smallest DBP_val
		of its HSPs
	 */
	for (i = 0; i < hasHSP; ++i) {
		kp1 = HSPlist[i];
		xslen = seq_beg[kp1->seq_nbr+1] - seq_beg[kp1->seq_nbr];
		R = Neff/xslen;
		minp = R * kp1->rawp.P_val;
		if (minp > 0.01) /* Is limit approximation applicable? */
			minp = (double)1.0 - exp(-minp);
		kp1->DBP_val = minp;
		for (kp2 = kp1->next; kp2 != NULL; kp2 = kp2->next) {
			minp2 = R * kp2->rawp.P_val;
			if (minp2 > 0.01)	/* Is limit approximation applicable */
				minp2 = (double)1.0 - exp(-minp2);
			kp2->DBP_val = minp2;
			if (minp2 < minp)
				minp = minp2;
		}
		kp1->rawp.smallest_P = minp;
	}

	hsort((char *)HSPlist, hasHSP, sizeof(HSPlist[0]), smaller);

	/* print the matches for each sequence */
	for (i = 0; i < hasHSP; ++i)
		print_seq(HSPlist[i]);
}

static void
print_seq(hp)
	HSPPtr	hp;
{
	int		current, len, qpos, spos;
	HSPPtr	kp;
	register char	*cp;

	current = hp->seq_nbr;

	/* get header line */
	if (mfil_seek(hfile, header_beg[current], 0) == -1)
		fatal("fseek error on header file");
	len = header_beg[current+1] - header_beg[current];
	if (len > sizeof(header)-1)
		len = sizeof(header)-1;
	if (mfil_dup(hfile, header, len) != len)
		fatal("error reading header");
	header[len] = NULLB;
	(void) printf("\n\n%s\n", header);

	/* get sequence */
#ifdef USE_SHM
	if (dbbuf != NULL) {
		len = seq_beg[current+1] - seq_beg[current] - 1;
		s_seq = dbbuf + seq_beg[current];
	}
	else
#endif
	{
		len = seq_beg[current+1] - seq_beg[current];
		if (fseek(sfile, seq_beg[current], 0) == -1)
			fatal("fseek error on sequence file");
		if (fread(s_seq, 1, len, sfile) != len)
			fatal("error reading sequence");
		--len;
	}
	(void) printf("\tLength = %s\n", ld2s((long)len));
	
	for (kp = hp; kp != NULL; kp = kp->next) {
		qpos = kp->q_pos;
		spos = kp->s_pos;
		len = kp->len;
		sdp = &sd[kp->frame];
		frame = sdp->frame;
		q_seq = sdp->seq;

		/* Extract the aligned subsequences from q_seq & s_seq */

		mem_cpy(qq_seq, q_seq + qpos, len);
		qq_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = qq_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		mem_cpy(ss_seq, s_seq + spos, len);
		ss_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = ss_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		report(ss_seq, qq_seq, spos, qpos, kp->score, len, kp->DBP_val);
		(void) fflush(stdout);
	}
}

/* report - print a match */
static void
report(s, q, spos, qpos, score, len, DBP_val)
	char	*s, *q;
	int		spos, qpos, score, len;
	double	DBP_val;
{
	double	Expect;
	int		numdone = 0;
	int		mlen;

	Expect = stoe(score, sdp->Lambda, sdp->K, sdp->H, Neff, sdp->elen);

	(void) printf("\n Frame = %s%d, Score = %d, Expect = %#0.2lg, Poisson P = %#0.2lg, Len = %s\n",
			(frame<3? "+":""), (frame < 3 ? frame+1 : 2-frame),
			score, Expect, DBP_val, ld2s(len));

	if (score > sdp->highs_obs)
		sdp->highs_obs = score;

	while (len > 0) {
		mlen = MIN(len, 60);
		(void) printf("\nQuery: %5d %.60s %d\n",
				ac_to_nt1(qpos + numdone + 1, frame, n_len),
				q,
				ac_to_nt3(qpos + numdone + mlen, frame, n_len));
		printmid(q, s, len);
		(void) printf("Sbjct: %5d %.60s %d\n",
				spos + numdone + 1, s, spos + numdone + mlen);
		q += 60;
		s += 60;
		len -= 60;
		numdone += 60;
	}
}

printmid(s1, s2, len)
	register char	*s1, *s2;
	register int	len;
{
	register char	c1, c2;

	(void) printf("%13s", "");
	if (len > 60)
		len = 60;
	while (len-- > 0) {
		if ((c1 = *s1++) == (c2 = *s2++) && c1 != 'X') {
			(void) putchar(c1);
			continue;
		}
		if (matrx[aa_atob[c1]][aa_atob[c2]] > 0)
			(void) putchar('+');
		else
			(void) putchar(' ');
	}
	(void) putchar('\n');
}

/*
    ac_to_nt1
 
    Convert an amino acid coordinate number to the nucleotide number
    of the first nucleotide in the codon.
*/
long
ac_to_nt1(pos, frame, len)
	long    pos; /* amino acid offset from start of a.a. sequence buffer */
	int     frame; /* reading frame of protein sequence */
	long    len; /* length of nucleotide sequence */
{
 
	if (frame < 3)
		return pos*CODON_LEN + frame - 2;
 
	return len - pos*CODON_LEN - frame + 6;
}
 
/*
	ac_to_nt3
 
	Convert an amino acid coordinate number to the nucleotide number
	of the third nucleotide in the codon.
*/
long
ac_to_nt3(pos, frame, len)
	long    pos;    /* amino acid offset from start of a.a. sequence buffer */
	int     frame;  /* reading frame of amino acid sequence */
	long    len;    /* length of nucleotide sequence */
{
	if (frame < 3)
		return pos*CODON_LEN + frame;
	return len - pos*CODON_LEN - frame + 4;
}

/* ----------------------------  parameter calculations  ---------------------*/
/*
	Amino acid residue frequencies used by S. Altschul.
*/
double fq[21] = {0., .081, .057, .045, .054, .015, .039, .061, .068, .022,
		.057, .093, .056, .025, .040, .049, .068, .058, .013, .032, .067};

/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, Lambda, K, H)
	int		lows, highs, q_len;
	register char	*q_seq;
	double	*Lambda, *K, *H;
{
	double	fq1[DIM2(matrx)];
	long	iq1[DIM2(matrx)];
	double	*prob;
	register int	i,j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM2(matrx); ++i)
		iq1[i]=0;
	for (i=0; i<q_len; ++i)
		++iq1[q_seq[i]];
	/* Convert frequencies to floating point */
	for (i=0; i<DIM1(fq1); ++i)
		fq1[i] = iq1[i];

	/* Convert frequencies to probabilities */
	for (i=0; i<DIM1(fq1); ++i)
		fq1[i] /= q_len;

	/* Calculate probability of each score */
	prob = (double *)ckalloc((highs-lows+1) * sizeof(*prob));
	for (i=highs-lows; i>=0; --i)
		prob[i]=0.;
	for (i=minchar; i<=maxchar; ++i)
		for (j=1; j<DIM1(fq); ++j)
			prob[matrx[i][j]-lows] += fq1[i]*fq[j];

	i = karlin(lows, highs, prob, Lambda, K, H);
	free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}


/* fmaxscore(seq, len) -- returns the maximum achievable score */
/* Note:  effect of parameter X is not taken into consideration here */
int
fmaxscore(seq, len)
	register char	*seq;
	register int	len;
{
	register char	*maxseq;
	register int	sum, maxsum;

	maxseq = seq+len;
	sum = maxsum = 0;
	while (seq < maxseq) {
		sum += maxcost[*seq++];
		if (sum < 0)
			sum = 0;
		else
			if (sum > maxsum)
				maxsum = sum;
	}
	return maxsum;
}
