/*************************************************************************
*                                                                        *
*                             COPYRIGHT NOTICE                           *
*                                                                        *
* This software/database is "United States Government Work" under the	 *
* terms of the United States Copyright Act.  It was written as part of	 *
* the author's official duties as a Government employee and thus cannot	 *
* be copyrighted.  This software/database is freely available to the	 *
* public for use without a copyright notice.  Restrictions cannot be	 *
* placed on its present or future use.                                   *
*                                                                        *
* Although all reasonable efforts have been taken to insure the accuracy *
* and reliability of the software and data, the National Library of      *
* Medicine (NLM) and the U.S. Government does not and cannot warrant the *
* performance or results that may be obtained by using this software or  *
* data.  The NLM and the U.S. Government disclaims all warranties as to  *
* performance, merchantability or fitness for any particular purpose.    *
*                                                                        *
* In any work or product derived from this material, proper attribution  *
* of the author as the source of the software or data would be           *
* appreciated.                                                           *
*                                                                        *
*************************************************************************/
/*
* GBLASTA - Search a protein database for regions that approximately match
* regions of a query sequence.  The basic command syntax is
*
* 	GBLASTA database query_sequence
*
* where:
* 	database names a file containing protein sequences, separated by
* 		header lines beginning with the character '>'.  A typical
* 		sequence entry is:
*
*		>CCHU (PIR) Cytochrome c - Human | 1.0,1.0,1.0,1.0,1.0
*		GDVEKGKKIFIMKCSQCHTVEKGGKHKTGP
*		NLHGLFGRKTGQAPGYSYTAANKNKGIIWG
*		EDTLMEYLENPKKYIPGTKMIFVGIKKKEE
*		RADLIAYLKKATNE
*
* 	query_sequence names a file containing a protein sequence (where the
* 		header line is optional).
* 
* GBLASTA differs from BLASTP in that it permits gaps in alignments.
* The GBLASTA command permits optional additional arguments, such as X=50,
* that reset certain internal parameters.  The available parameters are:
*
*	E or e gives the gap-extension cost (the default is 4)
*	G or G gives the gap-open cost (the default is 4)
*	K or k gives the karlin cutoff value (the default is 100)
*	T or t gives the threshold for a word match (the default is 20)
*	X or x gives the value for terminating word extensions
*		(the default is 20)
*
* Thus a typical command is
*
*	GBLASTA AABANK SEQ K=50 X=15
*/
#include <ncbi.h>

#define HEADERLEN_MAX	200	/* maximum length of header line */
#define QUERYLEN_MAX	5000	/* maximum length of query sequence */
#define WORD_SIZE	4	/* length of critical word (do not change!) */
#ifdef sun
#define signed
#endif
#define SMALL_VALS	signed char	/* type to hold values -128 to 127 */

#define DEFAULT_E	4	/* default constant for gap extensions */
#define DEFAULT_G	4	/* default constant for gap opening */
#define DEFAULT_K	100	/* default cutoff for karlins */
#define DEFAULT_T	20	/* default threshold for word hits */
#define DEFAULT_X	20	/* default constant for hit extensions */

char
	header[HEADERLEN_MAX],
	*s_seq, *ss_seq,
	q_seq[QUERYLEN_MAX+2],
	qq_seq[QUERYLEN_MAX+3];

int
	onechar,	/* G + E, the cost of a one-character gap */
	undef,
	nohit,
	hasHSP;	/* number of sequences containing a karlin */
long
	numwords,	/* number of critical words in hash table */
	totHSP,	/* number of karlins found */
	numhits;	/* number of database words in hash table */
int
	E,		/* constant penalty for gap extension */
	G,		/* constant penalty for gap opening */
	K,		/* cutoff for karlins */
	T,		/* threshold for word hits */
	X,		/* constant for hit extensions */
	n_kars;		/* number of karlins found for current sequence */

/* data structure to hold locations of matches for current database sequence */
int
	*_dlev,		/* extent of discovered matches .. */
	*diag_level,	/* on a given diagonal */
	*diag_stack,	/* list of diagonals with matches */
	*diag_top;	/* next diagonal with a match */

typedef struct karlin {
	int seq_nbr, q_pos, s_pos, score, big_score;
	struct karlin *next;
	} Karlin, *HSPPtr;

#define MAX_KARLIN 500	/* limit on number of karlins per database sequence */

HSPPtr karl[MAX_KARLIN];

HSPPtr firstHSP = NULL;
HSPPtr *HSPlist;

main(argc, argv)
int argc;
char *argv[];
{
	char *a;

	if (WORD_SIZE != 4)
		fatal("Word size must be 4.");
	if (argc < 3)
		fatal("%s database query_sequence ...", argv[0]);
	E = DEFAULT_E;
	G = DEFAULT_G;
	K = DEFAULT_K;
	T = DEFAULT_T;
	X = DEFAULT_X;
	while (--argc > 2) {
		if (argv[argc][1] != '=')
			fatal("argument %d has improper form", argc+1);
		a = argv[argc] + 2;
		switch (argv[argc][0]) {
			case 'E':
			case 'e':
				G = atoi(a);
				break;
			case 'G':
			case 'g':
				G = atoi(a);
				break;
			case 'K':
			case 'k':
				K = atoi(a);
				break;
			case 'T':
			case 't':
				T = atoi(a);
				break;
			case 'X':
			case 'x':
				X = atoi(a);
				break;
			default:
				fatal("Options are K, M, T and X.");
		}
	}
	onechar = G + E;
	undef = G + 1;
	get_weights();		/* handle array of substitution weights */
	get_query(argv[2]);
	bld_table();		/* build table of critical words */

	init_search();
	while (get_seq(argv[1])) {
		nohit = 1;
		n_kars = 0;
		search();	/* search a database sequence */
		if (n_kars > 0)
			++hasHSP;
	}
	init_search();
	print_HSPs();

	printf("\n\nStatistics:\n");
	printf("  # of critical words = %d\n", numwords);
	printf("  # of critical hits  = %d\n", numhits);
	printf("  # of matches found  = %d\n", totHSP);
	printf("  # of db sequences with at least one match = %d\n",
		hasHSP);

	exit (0);
}

/* get_query - read the query sequence */
get_query(file_name)
char *file_name;
{
	FILE *qp, *ckopen();
	char *p, *fgets();
		
	qp = ckopen(file_name, "r");
	for (p = q_seq; ; ) {
		if (fgets(p, QUERYLEN_MAX + 2 - (p - q_seq), qp) == NULL) {
			fclose(qp);
			*p = '\0';
			break;
		}
		if (p == q_seq && *p == '>')	/* optional header line */
			continue;
		while (isupper(*p))
			++p;
		if (*p != '\n' && *p != '\0')
			fatal("Illegal character %c in query sequence.", *p);
		if (p - q_seq > QUERYLEN_MAX)
			fatal("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
}

/* UNIX-specific input routine; database must be processed by SETDB */

int
	current,		/* number of current sequence */
	sfile,			/* file descriptor for sequences */
	hfile;			/* file descriptor for headers */
long
	maxlen,			/* max length of a database sequence */
	nbr_seqs,		/* number of database sequences */
	*seq_beg, 		/* starts of sequences */
	*header_beg;		/* starts of headers */

/* get_seq - read a database sequence and header line, return 0 on EOF */
int get_seq(file_name)
char *file_name;
{
	static int tfile, init = 1, i;
	int len, many;
	char name[FILENAME_MAX], *ckalloc();

	if (init) {
		init = 0;
		current = -1;
		sprintf(name, "%s.hed", file_name);
		hfile = open(name, 0);
		sprintf(name, "%s.seq", file_name);
		sfile = open(name, 0);
		sprintf(name, "%s.tab", file_name);
		tfile = open(name, 0);
		if (hfile == -1 || tfile == -1 || sfile == -1)
			fatal("Cannot open database %s.", file_name);
		if (read(tfile, (char *)&nbr_seqs, sizeof(int)) == -1 ||
		   read(tfile, (char *)&maxlen, sizeof(int)) == -1)
			fatal("corrupted database file");
		s_seq = ckalloc(sizeof(char)*(maxlen+1));
		ss_seq = ckalloc(sizeof(char)*(maxlen+3));
		seq_beg = (long *) ckalloc(sizeof(long)*(nbr_seqs+1));
		header_beg = (long *) ckalloc(sizeof(long)*(nbr_seqs+1));
		many = sizeof(long)*(nbr_seqs+1);
		if (read(tfile, (char *)seq_beg, many) != many ||
		   read(tfile, (char *)header_beg, many) != many)
			fatal("improper compressed database file");
		len = strlen(q_seq);
		_dlev = (int *) ckalloc(sizeof(int)*(len+maxlen+1));
		for (i = 0; i <= len+maxlen; ++i)
			_dlev[i] = -1;
		diag_level = _dlev + len;
		diag_top = diag_stack =
			(int *) ckalloc(sizeof(int)*(len+maxlen+1));
	}
	if (++current >= nbr_seqs)
		return 0;
	len = seq_beg[current+1] - seq_beg[current];
	if (read(sfile, s_seq, len) != len)
		fatal("corrupted sequence file");
	s_seq[len] = '\0';
	return 1;
}

/* -----------------   handle matrix of substitution weights   ---------------*/

/*  Get_weights() read an array giving weights for amino acid substitutions. */

char achars[] = "ARNBDCQZEGHILKMFPSTWYVX";	/* amino acid names */

int alpha = 23;			/* alpabet size */

int wgts[23][23] = {	/* the PAM120 matrix */
  { 3,-3, 0, 0, 0,-3,-1,-1, 0, 1,-3,-1,-3,-2,-2,-4, 1, 1, 1,-7,-4, 0, 0},
  {-3, 6,-1,-2,-3,-4, 1,-1,-3,-4, 1,-2,-4, 2,-1,-4,-1,-1,-2, 1,-6,-3, 0},
  { 0,-1, 4, 3, 2,-5, 0, 0, 1, 0, 2,-2,-4, 1,-3,-4,-2, 1, 0,-5,-2,-3, 0},
  { 0,-2, 3, 4, 3,-6, 0, 1, 2, 0, 1,-3,-5, 0,-4,-6,-2, 0,-1,-7,-4,-3, 0},
  { 0,-3, 2, 3, 5,-7, 1, 2, 3, 0, 0,-3,-5,-1,-4,-7,-2, 0,-1,-8,-5,-3, 0},
  {-3,-4,-5,-6,-7, 9,-7,-7,-7,-5,-4,-3,-7,-7,-6,-6,-3,-1,-3,-8,-1,-2, 0},
  {-1, 1, 0, 0, 1,-7, 6, 4, 2,-3, 3,-3,-2, 0,-1,-6, 0,-2,-2,-6,-5,-3, 0},
  {-1,-1, 0, 1, 2,-7, 4, 5, 3,-2, 1,-3,-3,-1,-3,-6,-1,-2,-2,-7,-5,-3, 0},
  { 0,-3, 1, 2, 3,-7, 2, 3, 5,-1,-1,-3,-4,-1,-4,-6,-1,-1,-2,-8,-4,-3, 0},
  { 1,-4, 0, 0, 0,-5,-3,-2,-1, 5,-4,-4,-5,-3,-4,-5,-2, 1,-1,-8,-6,-2, 0},
  {-3, 1, 2, 1, 0,-4, 3, 1,-1,-4, 7,-4,-3,-2,-4,-2,-1,-2,-3,-5,-1,-3, 0},
  {-1,-2,-2,-3,-3,-3,-3,-3,-3,-4,-4, 6, 1,-2, 1, 0,-3,-2, 0,-7,-2, 3, 0},
  {-3,-4,-4,-5,-5,-7,-2,-3,-4,-5,-3, 1, 5,-4, 3, 0,-3,-4,-3,-5,-3, 1, 0},
  {-2, 2, 1, 0,-1,-7, 0,-1,-1,-3,-2,-2,-4, 5, 0,-6,-2,-1,-1,-5,-6,-4, 0},
  {-2,-1,-3,-4,-4,-6,-1,-3,-4,-4,-4, 1, 3, 0, 8,-1,-3,-2,-1,-7,-4, 1, 0},
  {-4,-4,-4,-6,-7,-6,-6,-6,-6,-5,-2, 0, 0,-6,-1, 8,-5,-3,-4,-1, 4,-3, 0},
  { 1,-1,-2,-2,-2,-3, 0,-1,-1,-2,-1,-3,-3,-2,-3,-5, 6, 1,-1,-7,-6,-2, 0},
  { 1,-1, 1, 0, 0,-1,-2,-2,-1, 1,-2,-2,-4,-1,-2,-3, 1, 3, 2,-2,-3,-2, 0},
  { 1,-2, 0,-1,-1,-3,-2,-2,-2,-1,-3, 0,-3,-1,-1,-4,-1, 2, 4,-6,-3, 0, 0},
  {-7, 1,-5,-7,-8,-8,-6,-7,-8,-8,-5,-7,-5,-5,-7,-1,-7,-2,-6,12,-1,-8, 0},
  {-4,-6,-2,-4,-5,-1,-5,-5,-4,-6,-1,-2,-3,-6,-4, 4,-6,-3,-3,-1, 8,-3, 0},
  { 0,-3,-3,-3,-3,-2,-3,-3,-3,-2,-3, 3, 1,-4, 1,-3,-2,-2, 0,-8,-3, 5, 0},
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
};

SMALL_VALS
	maxcost[128],		/* max cost of substituting for a character */
	matrx[128][128],	/* matrx[i][j] = cost of replacing i by j */
	order[128][26];		/* order[i][j] = j-th best char to sub. for i */

get_weights()
{
	int i, j;
	SMALL_VALS *m, *o, *p, *q, t;

	alpha = strlen(achars);

	/* for illegal characters, set weight to lowest possible value */
	for (i = 0; i < 128; i++) {
		m = matrx[i];
		for (j = 0; j < 128; j++)
			m[j] = -128;
		maxcost[i] = -128;
	}

	/* read substitution weights for each residue */
	for (i = 0; i < alpha; i++) {
		m = matrx[achars[i]];
		for (j = 0; j < alpha; j++) {
			m[achars[j]] = wgts[i][j];
		}
		o = order[achars[i]];
		for (j = 0; j < alpha; j++)
			o[j] = achars[j];
		/* bubble-sort residue names by cost of substitution for i-th
		   residue */
		for (q = o + (alpha-1); q > o; --q)
			for (p = o; p < q; ++p)
				if (m[*p] < m[*q]) {
					t = *p;
					*p = *q;
					*q = t;
				}
		maxcost[achars[i]] = m[*o];
	}
}

/* -----------------------   build table of critical words   -----------------*/

struct hash_node {
	int ch;			/* fourth letter of word */
	struct match *matches;	/* positions where word hits query sequence */
	struct hash_node *link;	/* next word with same first three letters */
};

struct match {
	int pos;		/* position in query sequence */
	int score;		/* score where word hits */
	struct match *succ;	/* next hit for this word */
};

/* The hash value of a four-letter word is simply an encoding of its first
three letters:
	hval = (word[0] - 'A')*1024 + (word[1] - 'A')*32 + word[2] - 'A';
This is space-inefficient since some hash values are unreachable, but the
hash value is quickly updated as we march along a database sequence.  If s
points to the last letter of the new word, then the new hash value is:
	hval = ((hval & 01777) << 5) + (*s - 'A');
*/

struct hash_node *hashtab[26426];	/* hval of "ZZZ" is 26425 */

SMALL_VALS
	best[WORD_SIZE];	/* best[i] = smallest score of word[0]..word[i]
				   that can be extended to a critical word */
char
	word[WORD_SIZE],	/* candidate for critical word */
	*query;			/* pointer into query sequence */

bld_table()
{
	int i;

	for (query = q_seq; query[3] != '\0'; query++) {
		best[3] = T;
		for (i = 3; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);
	}
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for word[0]..word[i-1] */
enumerate(i, score)
int i, score;
{
	SMALL_VALS *o = order[query[i]]; /* possible letters by decreasing weight */
	int c, new_score;

	for (c = 0; c < alpha; ++c) {
		word[i] = o[c];
		new_score = score + matrx[query[i]][word[i]];
		if (new_score < best[i])
			break;
		if (i < WORD_SIZE-1)
			enumerate(i+1, new_score);
		else
			add_word(query - q_seq + WORD_SIZE, new_score);
	}
}
    
/* add_word - add a word to the table of critical words */
add_word(pos, score)
int pos, score;
{
	struct hash_node *h;
	struct match *m;
	long hval;
	char *ckalloc();

	numwords++;
	hval = (word[0] - 'A')*1024 + (word[1] - 'A')*32 + word[2] - 'A';
	for (h = hashtab[hval]; h; h = h->link)
		if (h->ch == word[3])
			break;
	if (h == NULL) {
		h = (struct hash_node *) ckalloc (sizeof(struct hash_node));
		h->link = hashtab[hval];
		hashtab[hval] = h;
		h->ch = word[3];
		h->matches = NULL;
	}
	m = (struct match *) ckalloc (sizeof(struct match));
	m->succ = h->matches;
	h->matches = m;
	m->pos = pos;
	m->score = score;
}

/* -----------------------   search a database sequence   --------------------*/

long gapwork;

typedef struct vtx { struct vtx *list;
                     int         ycrd, xcrd;
                     int         I, D, S, M;
                     struct vtx *Itr, *Dtr, *Str;
                   };

typedef struct anc { struct vtx *list;
                     int         ycrd;
                   };

#define SEQ_MAX 5000
#define POOLMAX 10000

static struct anc _anti[2*SEQ_MAX+5], *anti = _anti+2;
static struct vtx pool[POOLMAX], *fptr;

static init_search()
{ register struct anc *f;
  register struct vtx *p, *q;

  for (f = _anti+(2*SEQ_MAX+4); f >= _anti; f--)
    { f->list =  0;
      f->ycrd = -2;
    }

  q = 0;
  for (p = pool+(POOLMAX-1); p >= pool; q = p--)
    p->list = q;
  fptr = pool;
}


search()
{
	register struct hash_node *h;
	struct match *m;
	register char *s;
	register long hval;
	int i, j;
	HSPPtr kp;

	s = s_seq;
	if (s[0] == '\0' || s[1] == '\0' || s[2] == '\0')
		return;
	hval = (s[0] - 'A')*1024 + (s[1] - 'A')*32 + s[2] - 'A';
	for (s += 3; *s; ++s) {
		for (h = hashtab[hval]; h; h = h->link)
			if (h->ch == *s) {
				for (m = h->matches; m; m = m->succ)
					noalign(s+1-s_seq, m->pos, m->score);
				break;
			}
		hval = ((hval & 01777) << 5) + (*s - 'A');
	}
	/* bubble sort karlins by score */
	for (i = n_kars-1; i > 0; --i)
		for (j = 0; j < i; ++j)
			if (karl[j]->score < karl[j+1]->score) {
				kp = karl[j];
				karl[j] = karl[j+1];
				karl[j+1] = kp;
			}

	/* add karlins for this sequence to the cumulative list */
	if (n_kars > 0) {
		karl[n_kars-1]->next = firstHSP;
		for (i = n_kars-2; i >= 0; --i)
			karl[i]->next = karl[i+1];
	}
	firstHSP = karl[0];

	while (diag_top > diag_stack)
		diag_level[*--diag_top] = -1;
}


noalign(send,qend,score) int send, qend, score;
{ register struct vtx *pb2, *pb1, *p, *q, *pa1, *pa2;
  register int k, y;
  int S, M, I, D;
  int noasgn, nodel, noins;

  struct vtx *emax, *eroot;
  struct vtx *bmax, *broot;
  int kend, kmax;
  int kbeg, kmin;
  int qbeg, sbeg;

  char *salign, *qalign;
  static char SA[2*SEQ_MAX+1], QA[2*SEQ_MAX+1];

#define ADV(p,y)				\
  while (p->list != 0 && p->list->ycrd <= y)	\
    p = p->list;

#define INSERT(p,y)		\
  ADV(p,y)			\
  if (p->ycrd != y)		\
    { q = fptr;			\
      fptr = q->list;		\
      q->list = p->list;	\
      p = p->list = q;		\
      p->ycrd = y;		\
    }

  qbeg = qend-WORD_SIZE;
  sbeg = send-WORD_SIZE;
  numhits += 1;
  if (sbeg < diag_level[sbeg-qbeg]) return;

  kbeg = qbeg + sbeg;
  p   = (struct vtx *) (anti + kbeg);
  INSERT(p,sbeg);
  broot = bmax = p;
  for (k = kbeg; anti[k].list != 0; k--)
    { pb2 = (struct vtx *) (anti + (k+2));
      pb1 = (struct vtx *) (anti + (k+1));
      q   = (struct vtx *) (anti + k);
      pa1 = (struct vtx *) (anti + (k-1));
      pa2 = (struct vtx *) (anti + (k-2));
      for (p = q->list; p != 0; p = q->list)
        { y = p->ycrd;
          ADV(pb2,y+1)
          noasgn = 1;
          if (pb2->ycrd == y+1)
            { S = pb2->S + matrx[q_seq[k-y]][s_seq[y]];
              p->Str = pb2;
              M = pb2->M;
              if (M < S)
                M = S;
              noasgn = 0;
            }
          ADV(pb1,y)
          if (noins = (pb1->ycrd == y))
            { I = pb1->S - onechar;
              p->Itr = pb1;
              if (I < pb1->I - E)
                { I = pb1->I - E;
                  p->Itr = pb1->Itr;
                }
              if (noasgn || S < I)
                { S = I;
                  p->Str = p->Itr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
               noasgn = 0;
            }
          ADV(pb1,y+1)
          if (nodel = (pb1->ycrd == y+1))
            { D = pb1->S - onechar;
              p->Dtr = pb1;
              if (D < pb1->D - E)
                { D = pb1->D - E;
                  p->Dtr = pb1->Dtr;
                }
              if (noasgn || S < D)
                { S = D;
                  p->Str = p->Dtr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
            }
          if (k == kbeg)
            S = M = 0;
          if (S >= M - X)
            { p->xcrd = k-y;
              p->S = S;
              p->M = M;
              if (nodel)
                p->D = D;
              else
                p->D = S - undef;
              if (noins)
                p->I = I;
              else
                p->I = S - undef;
#ifdef SHOW
printf("(%3d,%3d) S=%d D=%d I=%d M=%d\n",p->xcrd,p->ycrd,p->S,p->D,p->I,p->M);
fflush(stdout);
#endif
              gapwork += 1;
              if (S >= bmax->M)
                bmax = p;
              if (y > 0)
                { INSERT(pa1,y-1)
                  if (p->xcrd > 0)
                    { INSERT(pa1,y)
                      INSERT(pa2,y-1)
                    }
                }
              else if (y > 0)
                { INSERT(pa1,y-1) }
              q = p;
            }
          else
            { q->list = p->list;
              p->list = fptr;
              fptr = p;
            }
        }
    }
  kmin = k;

  kend = qend + send;
  p   = (struct vtx *) (anti + kend);
  INSERT(p,send);
  eroot = emax = p;
  for (k = kend; anti[k].list != 0; k++)
    { pb2 = (struct vtx *) (anti + (k-2));
      pb1 = (struct vtx *) (anti + (k-1));
      q   = (struct vtx *) (anti + k);
      pa1 = (struct vtx *) (anti + (k+1));
      pa2 = (struct vtx *) (anti + (k+2));
      for (p = q->list; p != 0; p = q->list)
        { y = p->ycrd - 1;
          ADV(pb2,y)
          noasgn = 1;
          if (pb2->ycrd == y)
            { S = pb2->S + matrx[q_seq[k-y-2]][s_seq[y]];
              p->Str = pb2;
              M = pb2->M;
              if (M < S)
                M = S;
              noasgn = 0;
            }
          ADV(pb1,y)
          if (nodel = (pb1->ycrd == y))
            { D = pb1->S - onechar;
              p->Dtr = pb1;
              if (D < pb1->D - E)
                { D = pb1->D - E;
                  p->Dtr = pb1->Dtr;
                }
              if (noasgn || S < D)
                { S = D;
                  p->Str = p->Dtr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
              noasgn = 0;
            }
          y += 1;
          ADV(pb1,y)
          if (noins = (pb1->ycrd == y))
            { I = pb1->S - onechar;
              p->Itr = pb1;
              if (I < pb1->I - E)
                { I = pb1->I - E;
                  p->Itr = pb1->Itr;
                }
              if (noasgn || S < I)
                { S = I;
                  p->Str = p->Itr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
            }
          if (k == kend)
            M = S = 0;
          if (S >= M - X)
            { p->xcrd = k-y;
              p->S = S;
              p->M = M;
              if (nodel)
                p->D = D;
              else
                p->D = S - undef;
              if (noins)
                p->I = I;
              else
                p->I = S - undef;
#ifdef SHOW
printf("(%3d,%3d) S=%d D=%d I=%d M=%d\n",p->xcrd,p->ycrd,p->S,p->D,p->I,p->M);
fflush(stdout);
#endif
              gapwork += 1;
              if (S >= emax->M)
                emax = p;
              if (q_seq[p->xcrd] != '\0')
                { INSERT(pa1,y)
                  if (s_seq[y] != '\0')
                    { INSERT(pa1,y+1)
                      INSERT(pa2,y+1)
                    }
                }
              else if (s_seq[y] != '\0')
                { INSERT(pa1,y+1) }
              q = p;
            }
          else
            { q->list = p->list;
              p->list = fptr;
              fptr = p;
            }
        }
    }
  kmax = k;

  if (bmax->S + score + emax->S >= K)
    { totHSP += 1;
/***************************************/
{
HSPPtr kp;
		kp = karl[n_kars++] = (Karlin *)ckalloc(sizeof(Karlin));
		kp->seq_nbr = current;
		kp->q_pos = qend;
		kp->s_pos = send;
		kp->score = score;
		kp->big_score = bmax->S + score + emax->S;
}
/***************************************/
      salign = SA;
      qalign = QA;
      for (p = bmax; p != broot; p = q)
        { q = p->Str;
          if (q->ycrd == p->ycrd)
            { k = q->xcrd - p->xcrd;
              strncpy(qalign,q_seq + p->xcrd,k);
              qalign += k;
              while (k-- > 0)
                *salign++ = '-';
            }
          else if (q->xcrd == p->xcrd)
            { k = q->ycrd - p->ycrd;
              strncpy(salign,s_seq + p->ycrd,k);
              salign += k;
              while (k-- > 0)
                *qalign++ = '-';
            }
          else
            { *salign++ = s_seq[p->ycrd];
              *qalign++ = q_seq[p->xcrd];
            }
        }
    
      *salign++ = *qalign++ = '|';
      strncpy(qalign,q_seq+qbeg,WORD_SIZE);
      strncpy(salign,s_seq+sbeg,WORD_SIZE);
      qalign += WORD_SIZE;
      salign += WORD_SIZE;
      *salign++ = *qalign++ = '|';

      p = emax;
      q = p->Str;
      while (p != eroot)
        { pb1 = q->Str;
          q->Str = p;
          p = q;
          q = pb1;
        }

      y = send - qend;
      if (diag_level[y] < 0)
        *diag_top++ = y;
      if (send > diag_level[y])
        diag_level[y] = send;
      for (p = eroot; p != emax; p = q)
        { q = p->Str;
          y = p->ycrd - p->xcrd;
          if (q->ycrd == p->ycrd)
            { k = q->xcrd - p->xcrd;
              strncpy(qalign,q_seq + p->xcrd,k);
              qalign += k;
              while (k-- > 0)
                { *salign++ = '-';
                  if (diag_level[--y] < 0)
                    *diag_top++ = y;
                  if (q->ycrd > diag_level[y])
                    diag_level[y] = q->ycrd;
                }
            }
          else if (q->xcrd == p->xcrd)
            { k = q->ycrd - p->ycrd;
              strncpy(salign,s_seq + p->ycrd,k);
              salign += k;
              while (k-- > 0)
                { *qalign++ = '-';
                  if (diag_level[++y] < 0)
                    *diag_top++ = y;
                  if (q->xcrd + y > diag_level[y])
                    diag_level[y] = q->xcrd + y;
                }
            }
          else
            { *salign++ = s_seq[p->ycrd];
              *qalign++ = q_seq[p->xcrd];
              diag_level[y] = q->ycrd;
            }
        }
      *salign = *qalign = '\0';
/*
      REPORT(QA,SA,bmax->xcrd+1,bmax->ycrd+1,
                     emax->xcrd  ,emax->ycrd  ,bmax->S+score+emax->S);
*/
    }

  for (k = kbeg; k > kmin; k--)
    { for (p = anti[k].list; p != 0; p = q)
        { q = p->list;
          p->list = fptr;
          fptr = p;
        }
      anti[k].list = 0;
    }

  for (k = kend; k < kmax; k++)
    { for (p = anti[k].list; p != 0; p = q)
        { q = p->list;
          p->list = fptr;
          fptr = p;
        }
      anti[k].list = 0;
    }
}

/* print_HSPs - sort and print all karlins */
print_HSPs()
{
	int i;
	HSPPtr kp1, kp2;
	char *ckalloc();

	if (hasHSP < MAX_KARLIN)
		HSPlist = karl;	/* use original array of HSPPtrs */
	else
		HSPlist = (HSPPtr *)ckalloc(hasHSP*sizeof(HSPPtr));
	/* break into sublists for each database sequence containing karlins */
	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    kp1->seq_nbr != kp2->seq_nbr)
				break;
		if (kp1 == NULL)
			fatal("improper karlin list");
		kp1->next = NULL;
		firstHSP = kp2;
	}

	sort_karlins();

	/* print the matches for each sequence */
	for (i = 0; i < hasHSP; ++i)
		print_seq(i);
}

/* sort_karlins - order database sequence for printing */
sort_karlins()
{
	int i;
	HSPPtr kp;

	for (i = (hasHSP/2) - 1; i >= 0; --i)
		heapify(i, hasHSP-1);
	for (i = hasHSP-1; i > 0; --i) {
		kp = HSPlist[0];
		HSPlist[0] = HSPlist[i];
		HSPlist[i] = kp;
		if (i > 1)
			heapify(0, i-1);
	}
}

/* heapify - core procedure for heapsort */
heapify(i, last)
int i, last;
{
	int lim = (last-1)/2, left_son, small_son;
	HSPPtr kp;

	while (i <= lim) {
		left_son = 2*i + 1;
		if (left_son == last)
			small_son = left_son;
		else
			small_son = smaller(left_son, left_son+1);
		if (smaller(i, small_son) == small_son) {
			kp = HSPlist[i];
			HSPlist[i] = HSPlist[small_son];
			HSPlist[small_son] = kp;
			i = small_son;
		} else
			break;
	}
}

/* smaller - determine which of two sequences should be printed first */
int smaller(i, j)	/* if tie, return first arg */
int i, j;
{
	HSPPtr ki = HSPlist[i], kj = HSPlist[j];

	/* print one with larger score first */
	return ((ki->big_score <= kj->big_score) ? i : j);
}

/* print_seq - display karlins for the i-th sequence */
print_seq(i)
int i;
{
	int current, len;
	HSPPtr kp;

	current = HSPlist[i]->seq_nbr;

	/* get header line */
	lseek(hfile, header_beg[current], 0);
	len = header_beg[current+1] - header_beg[current];
	read(hfile, header, len);
	if (--len > 78)
		len = 78;
	header[len] = '\0';
	printf("\n\n%s\n", header);

	/* get sequence */
	lseek(sfile, seq_beg[current], 0);
	len = seq_beg[current+1] - seq_beg[current];
	if (read(sfile, s_seq, len) != len)
		fatal("cannot read %d characters of sequence", len);
	s_seq[len] = '\0';
	
	/* print karlins */
	for (kp = HSPlist[i]; kp; kp = kp->next)
		align(kp->s_pos, kp->q_pos, kp->score);
}

align(send,qend,score) int send, qend, score;
{ register struct vtx *pb2, *pb1, *p, *q, *pa1, *pa2;
  register int k, y;
  int S, M, I, D;
  int noasgn, nodel, noins;

  struct vtx *emax, *eroot;
  struct vtx *bmax, *broot;
  int kend, kmax;
  int kbeg, kmin;
  int qbeg, sbeg;

  char *salign, *qalign;
  static char SA[2*SEQ_MAX+1], QA[2*SEQ_MAX+1];

#define ADV(p,y)				\
  while (p->list != 0 && p->list->ycrd <= y)	\
    p = p->list;

#define INSERT(p,y)		\
  ADV(p,y)			\
  if (p->ycrd != y)		\
    { q = fptr;			\
      fptr = q->list;		\
      q->list = p->list;	\
      p = p->list = q;		\
      p->ycrd = y;		\
    }

  qbeg = qend-WORD_SIZE;
  sbeg = send-WORD_SIZE;

  kbeg = qbeg + sbeg;
  p   = (struct vtx *) (anti + kbeg);
  INSERT(p,sbeg);
  broot = bmax = p;
  for (k = kbeg; anti[k].list != 0; k--)
    { pb2 = (struct vtx *) (anti + (k+2));
      pb1 = (struct vtx *) (anti + (k+1));
      q   = (struct vtx *) (anti + k);
      pa1 = (struct vtx *) (anti + (k-1));
      pa2 = (struct vtx *) (anti + (k-2));
      for (p = q->list; p != 0; p = q->list)
        { y = p->ycrd;
          ADV(pb2,y+1)
          noasgn = 1;
          if (pb2->ycrd == y+1)
            { S = pb2->S + matrx[q_seq[k-y]][s_seq[y]];
              p->Str = pb2;
              M = pb2->M;
              if (M < S)
                M = S;
              noasgn = 0;
            }
          ADV(pb1,y)
          if (noins = (pb1->ycrd == y))
            { I = pb1->S - onechar;
              p->Itr = pb1;
              if (I < pb1->I - E)
                { I = pb1->I - E;
                  p->Itr = pb1->Itr;
                }
              if (noasgn || S < I)
                { S = I;
                  p->Str = p->Itr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
               noasgn = 0;
            }
          ADV(pb1,y+1)
          if (nodel = (pb1->ycrd == y+1))
            { D = pb1->S - onechar;
              p->Dtr = pb1;
              if (D < pb1->D - E)
                { D = pb1->D - E;
                  p->Dtr = pb1->Dtr;
                }
              if (noasgn || S < D)
                { S = D;
                  p->Str = p->Dtr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
            }
          if (k == kbeg)
            S = M = 0;
          if (S >= M - X)
            { p->xcrd = k-y;
              p->S = S;
              p->M = M;
              if (nodel)
                p->D = D;
              else
                p->D = S - undef;
              if (noins)
                p->I = I;
              else
                p->I = S - undef;
#ifdef SHOW
printf("(%3d,%3d) S=%d D=%d I=%d M=%d\n",p->xcrd,p->ycrd,p->S,p->D,p->I,p->M);
fflush(stdout);
#endif
              gapwork += 1;
              if (S >= bmax->M)
                bmax = p;
              if (y > 0)
                { INSERT(pa1,y-1)
                  if (p->xcrd > 0)
                    { INSERT(pa1,y)
                      INSERT(pa2,y-1)
                    }
                }
              else if (y > 0)
                { INSERT(pa1,y-1) }
              q = p;
            }
          else
            { q->list = p->list;
              p->list = fptr;
              fptr = p;
            }
        }
    }
  kmin = k;

  kend = qend + send;
  p   = (struct vtx *) (anti + kend);
  INSERT(p,send);
  eroot = emax = p;
  for (k = kend; anti[k].list != 0; k++)
    { pb2 = (struct vtx *) (anti + (k-2));
      pb1 = (struct vtx *) (anti + (k-1));
      q   = (struct vtx *) (anti + k);
      pa1 = (struct vtx *) (anti + (k+1));
      pa2 = (struct vtx *) (anti + (k+2));
      for (p = q->list; p != 0; p = q->list)
        { y = p->ycrd - 1;
          ADV(pb2,y)
          noasgn = 1;
          if (pb2->ycrd == y)
            { S = pb2->S + matrx[q_seq[k-y-2]][s_seq[y]];
              p->Str = pb2;
              M = pb2->M;
              if (M < S)
                M = S;
              noasgn = 0;
            }
          ADV(pb1,y)
          if (nodel = (pb1->ycrd == y))
            { D = pb1->S - onechar;
              p->Dtr = pb1;
              if (D < pb1->D - E)
                { D = pb1->D - E;
                  p->Dtr = pb1->Dtr;
                }
              if (noasgn || S < D)
                { S = D;
                  p->Str = p->Dtr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
              noasgn = 0;
            }
          y += 1;
          ADV(pb1,y)
          if (noins = (pb1->ycrd == y))
            { I = pb1->S - onechar;
              p->Itr = pb1;
              if (I < pb1->I - E)
                { I = pb1->I - E;
                  p->Itr = pb1->Itr;
                }
              if (noasgn || S < I)
                { S = I;
                  p->Str = p->Itr;
                }
              if (noasgn || M < pb1->M)
                M = pb1->M;
            }
          if (k == kend)
            M = S = 0;
          if (S >= M - X)
            { p->xcrd = k-y;
              p->S = S;
              p->M = M;
              if (nodel)
                p->D = D;
              else
                p->D = S - undef;
              if (noins)
                p->I = I;
              else
                p->I = S - undef;
#ifdef SHOW
printf("(%3d,%3d) S=%d D=%d I=%d M=%d\n",p->xcrd,p->ycrd,p->S,p->D,p->I,p->M);
fflush(stdout);
#endif
              gapwork += 1;
              if (S >= emax->M)
                emax = p;
              if (q_seq[p->xcrd] != '\0')
                { INSERT(pa1,y)
                  if (s_seq[y] != '\0')
                    { INSERT(pa1,y+1)
                      INSERT(pa2,y+1)
                    }
                }
              else if (s_seq[y] != '\0')
                { INSERT(pa1,y+1) }
              q = p;
            }
          else
            { q->list = p->list;
              p->list = fptr;
              fptr = p;
            }
        }
    }
  kmax = k;

if(bmax->S + score + emax->S < K)
printf("HELP: score = %d but limit = %d\n",  bmax->S + score + emax->S, K);
  if (bmax->S + score + emax->S >= K)
    { salign = SA;
      qalign = QA;
      for (p = bmax; p != broot; p = q)
        { q = p->Str;
          if (q->ycrd == p->ycrd)
            { k = q->xcrd - p->xcrd;
              strncpy(qalign,q_seq + p->xcrd,k);
              qalign += k;
              while (k-- > 0)
                *salign++ = '-';
            }
          else if (q->xcrd == p->xcrd)
            { k = q->ycrd - p->ycrd;
              strncpy(salign,s_seq + p->ycrd,k);
              salign += k;
              while (k-- > 0)
                *qalign++ = '-';
            }
          else
            { *salign++ = s_seq[p->ycrd];
              *qalign++ = q_seq[p->xcrd];
            }
        }
    
      *salign++ = *qalign++ = '|';
      strncpy(qalign,q_seq+qbeg,WORD_SIZE);
      strncpy(salign,s_seq+sbeg,WORD_SIZE);
      qalign += WORD_SIZE;
      salign += WORD_SIZE;
      *salign++ = *qalign++ = '|';

      p = emax;
      q = p->Str;
      while (p != eroot)
        { pb1 = q->Str;
          q->Str = p;
          p = q;
          q = pb1;
        }

      y = send - qend;
      for (p = eroot; p != emax; p = q)
        { q = p->Str;
          y = p->ycrd - p->xcrd;
          if (q->ycrd == p->ycrd)
            { k = q->xcrd - p->xcrd;
              strncpy(qalign,q_seq + p->xcrd,k);
              qalign += k;
              while (k-- > 0)
                *salign++ = '-';
            }
          else if (q->xcrd == p->xcrd)
            { k = q->ycrd - p->ycrd;
              strncpy(salign,s_seq + p->ycrd,k);
              salign += k;
              while (k-- > 0)
                *qalign++ = '-';
            }
          else
            { *salign++ = s_seq[p->ycrd];
              *qalign++ = q_seq[p->xcrd];
            }
        }
      *salign = *qalign = '\0';
      REPORT(QA,SA,bmax->xcrd+1,bmax->ycrd+1,
                     emax->xcrd  ,emax->ycrd  ,bmax->S+score+emax->S);
    }

  for (k = kbeg; k > kmin; k--)
    { for (p = anti[k].list; p != 0; p = q)
        { q = p->list;
          p->list = fptr;
          fptr = p;
        }
      anti[k].list = 0;
    }

  for (k = kend; k < kmax; k++)
    { for (p = anti[k].list; p != 0; p = q)
        { q = p->list;
          p->list = fptr;
          fptr = p;
        }
      anti[k].list = 0;
    }
}

int REPORT(qa,sa,qb,sb,qe,se,score)
char *qa, *sa;
int qb, sb, qe, se, score;
{ register int len;

#ifdef OVERLAP
  if (qb > 2 && sb > 2 || qe < qlen-2 && se < slen-2)
    { totmatch -= 1;
      return;
    }
#endif

  printf("\n  Score  = %4d,   Query range = %4d,%4d\n",score,qb,qe);
  len = strlen(sa);
  printf("  Length = %4d, Subject range = %4d,%4d\n",len,sb,se);
  while (len > 0)
    { printf("\nQuery: %.60s\n",qa);
      printf("Sbjct: %.60s\n",sa);
      qa  += 60;
      sa  += 60;
      len -= 60;
    }
}
