/**************** Statistical Significance Parameter Subroutine ****************

        Version 1.0     February 2, 1990
        Version 1.2     July 6,     1990

        Program by:     Stephen Altschul

        Address:        National Center for Biotechnology Information
                        National Library of Medicine
                        National Institutes of Health
                        Bethesda, MD  20894

        Internet:       altschul@ncbi.nlm.nih.gov

See:    Karlin, S. & Altschul, S.F. "Methods for Assessing the Statistical
        Significance of Molecular Sequence Features by Using General Scoring
        Schemes,"  Proc. Natl. Acad. Sci. USA 87 (1990), 2264-2268.

        Computes the parameters lambda and K for use in calculating the
        statistical significance of high-scoring segments or subalignments.

        The scoring scheme must be integer valued.  A positive score must be
        possible, but the expected (mean) score must be negative.

        A program that calls this routine must provide the value of the lowest
        possible score, the value of the greatest possible score, and a pointer
        to an array of probabilities for the occurence of all scores between
        these two extreme scores.  For example, if score -2 occurs with
        probability 0.7, score 0 occurs with probability 0.1, and score 3
        occurs with probability 0.2, then the subroutine must be called with
        low = -2, high = 3, and pr pointing to the array of values
        { 0.7, 0.0, 0.1, 0.0, 0.0, 0.2 }.  The calling program must also provide
        pointers to lambda and K; the subroutine will then calculate the values
        of these two parameters.  In this example, lambda=0.330 and K=0.154.

        The parameters lambda and K can be used as follows.  Suppose we are
        given a length N random sequence of independent letters.  Associated
        with each letter is a score, and the probabilities of the letters
        determine the probability for each score.  Let S be the aggregate score
        of the highest scoring contiguous segment of this sequence.  Then if N
        is sufficiently large (greater than 100), the following bound on the
        probability that S is greater than or equal to x applies:
        
                P( S >= x )   <=   1 - exp [ - KN exp ( - lambda * x ) ].
        
        In other words, the p-value for this segment can be written as
        1-exp[-KN*exp(-lambda*S)].

        This formula can be applied to pairwise sequence comparison by assigning
        scores to pairs of letters (e.g. amino acids), and by replacing N in the
        formula with N*M, where N and M are the lengths of the two sequences
        being compared.

        In addition, letting y = KN*exp(-lambda*S), the p-value for finding m
        distinct segments all with score >= S is given by:

                               2             m-1           -y
                1 - [ 1 + y + y /2! + ... + y   /(m-1)! ] e

        Notice that for m=1 this formula reduces to 1-exp(-y), which is the same
        as the previous formula.

*******************************************************************************/
#include <ncbi.h>
#include "blast.h"

#define MAXIT	150	/* Maximum no. of iterations */

	/* For greater accuracy, set SUMLIMIT to 0.00001 */
	/* For higher speed, use SUMLIMIT to 0.0001 */
#define SUMLIMIT 0.0001

int
karlin(low, high, prob, lambda, K, H)
	SCORE	low;		/* Lowest score (must be negative)    */
	SCORE	high;		/* Highest score (must be positive)   */
	double	*prob;		/* Probabilities for various scores   */
	double	*lambda;	/* Pointer to parameter Lambda        */
	double	*K;			/* Pointer to parameter K             */
	double	*H;			/* Pointer to relative entropy parameter H */
{
	int		range;
	register int	i, j;
	Int4	lo, hi, first, last;
	register double	sum;
	double	up, new, Sum, av;
	double	*p0 = NULL, *P0 = NULL;
	double	*p, *P, *ptrP, *ptr1, *ptr2, *ptr1e;
	int		retcode = 0;

	/* Check that scores and their associated probabilities are valid */

	if (low >= 0) {
		fprintf(stderr, "The lowest score in the substitution matrix must be negative.\n");
		goto ErrExit;
	}
	if (high <= 0) {
		fprintf(stderr, "The highest score in the substitution matrix must be positive.\n");
		goto ErrExit;
	}
	for (i=range=high-low; i > -low && prob[i] == 0.; --i)
		;
	if (i <= -low) {
		fprintf(stderr, "A positive score is impossible with the given substitution matrix and residue composition.\n");
		goto ErrExit;
	}
	for (sum=0., i=0; i <= range; sum += prob[i++])
		if (prob[i] < 0.) {
			fprintf(stderr, "Negative probabilities are disallowed.\n");
			goto ErrExit;
		}

	p0 = p = (double *)ckalloc0(sizeof(*p)*(range+1));
	P0 = P = (double *)ckalloc0(MAXIT*sizeof(*P)*(range+1));

	if (sum < 0.99995 || sum > 1.00005)
		fprintf(stderr, "Probabilities sum to %.5f and will be normalized to 1.\n",
			sum);
	for (Sum=low, i=0; i <= range; ++i)
		Sum += i * (p[i] = prob[i]/sum);
	if (Sum >= 0.) {
		fprintf(stderr, "Invalid (non-negative) expected score:  %#0.2lg", Sum);
		goto ErrExit;
	}

	/* Calculate the parameter Lambda */

	up = 0.5;
	do {
		up *= 2;
		ptr1 = p;
		for (sum=0, i=low; i<=high; ++i)
			sum += *ptr1++ * exp(up*i);
	} while (sum < 1.0);

	for (*lambda=0., j=0; j<25; ++j) {
		new = (*lambda+up)/2.0;
		ptr1 = p;
		for (sum=0., i=low; i<=high; ++i)
			sum += *ptr1++ * exp(new*i);
		if (sum>1.0)
			up = new;
		else
			*lambda = new;
	}

	/* Calculate the relative entropy of the p's and q's, the parameter H */

	ptr1=p;
	for (av=0., i=low; i<=high; ++i)
		av += *ptr1++ *i*exp(*lambda*i);
	*H = *lambda * av/log(2.0);

	/* Calculate the parameter K */
	if (low == -1 || high == 1) {
		*K = (high==1 ? av : Sum*Sum/av);
		*K *= 1.0 - exp(- *lambda);
		goto OKExit;
	}
	Sum = 0.;
	lo = hi = 0;
	for (*P=sum=1., j=1; j<=MAXIT && sum>SUMLIMIT; Sum += sum/=j++) {
		first = last = range;
		for (ptrP=P+(hi+=high)-(lo+=low); ptrP>=P; *ptrP-- =sum) {
			ptr1 = ptrP-first;
			ptr1e = ptrP-last;
			ptr2 = p+first;
			for (sum=0.; ptr1>=ptr1e; )
				sum += *ptr1-- * *ptr2++;
			if (first)
				--first;
			if (ptrP-P <= range)
				--last;
		}
		for (sum=0., i=lo; i; ++i)
			sum += *++ptrP * exp(*lambda*i);
		for (; i <= hi; ++i)
			sum += *++ptrP;
	}
	if (j > MAXIT)
		fprintf(stderr, "Value for K may be too large due to insufficient iterations.\n");
	for (i=low; p[i-low] == 0.0; ++i)
		;
	for (j = -i; i < high && j > 1; )
		if (p[++i-low])
			j = fct_gcd(j, i);

	*K = (j*exp(-2.0*Sum))/(av*(1.0-exp(- *lambda*j)));

OKExit:
	if (p0 != NULL)
		free((char *)p0);
	if (P0 != NULL)
		free((char *)P0);
	return retcode;

ErrExit:
	retcode = 1;
	goto OKExit;
}
