/*
* PAM - Calculate a log-odds matrix for given PAM distance.
* The basic command syntax is
*
* 	pam <PAM distance> [-s#]
*
* where	PAM distance is an integer from 2 to 511.
*
* The PAM matrix is a "log-odds" matrix, and is calculated as the natural
* logarithm of target over background amino acid pair frequencies, divided
* by <scale>.  The default value for scale is (ln 2)/2 = 0.3466.  This may
* be changed by using the command line option -s<scale>.  For significance
* calculations, <scale> is approximately equal to the parameter "lambda"
* (Karlin & Altschul, PNAS 87:2264-2268).  The slight difference is due to
* round-off errors, and to the use of PAM matrices with proteins that have
* amino acid frequencies different than those used by Dayhoff et al.
* 
* Relevant literature:
*
* Dayhoff, M. O., Schwartz, R. M. & Orcutt, B. C. (1978).
* A model of evolutionary change in proteins. In Atlas of Protein Sequence
* and Structure, Vol. 5, Suppl. 3, Ed. M. O. Dayhoff, pp. 345-352.
* Natl. Biomed. Res. Found., Washington.
*
*/
#include <ncbi.h>
#include <ncbilib.h>

/*	One letter amino acid code	*/

char aa[] = {"ARNDCQEGHILKMFPSTWYVBZX"};

#define STOP_CODON_CHR	'*'

/*	Dayhoff amino acid background frequencies	*/

double fq[20] = {
0.087, 0.041, 0.040, 0.047, 0.033, 0.038, 0.050, 0.089, 0.034, 0.037,
0.085, 0.081, 0.015, 0.040, 0.051, 0.070, 0.058, 0.010, 0.030, 0.065};

/*	Dayhoff 1 PAM stochastic amino acid transition matrix	*/

int mdm[400] = {
9867,  2,  9, 10,  3,  8, 17, 21,  2,  6,  4,  2,  6,  2, 22, 35, 32, 0,  2, 18,
 1, 9913,  1,  0,  1, 10,  0,  0, 10,  3,  1, 19,  4,  1,  4,  6,  1, 8,  0,  1,
 4,  1, 9822, 36,  0,  4,  6,  6, 21,  3,  1, 13,  0,  1,  2, 20,  9, 1,  4,  1,
 6,  0, 42, 9859,  0,  6, 53,  6,  4,  1,  0,  3,  0,  0,  1,  5,  3, 0,  0,  1,
 1,  1,  0,  0, 9973,  0,  0,  0,  1,  1,  0,  0,  0,  0,  1,  5,  1, 0,  3,  2,
 3,  9,  4,  5,  0, 9876, 27,  1, 23,  1,  3,  6,  4,  0,  6,  2,  2, 0,  0,  1,
10,  0,  7, 56,  0, 35, 9865,  4,  2,  3,  1,  4,  1,  0,  3,  4,  2, 0,  1,  2,
21,  1, 12, 11,  1,  3,  7, 9935,  1,  0,  1,  2,  1,  1,  3, 21,  3, 0,  0,  5,
 1,  8, 18,  3,  1, 20,  1,  0, 9912,  0,  1,  1,  0,  2,  3,  1,  1, 1,  4,  1,
 2,  2,  3,  1,  2,  1,  2,  0,  0, 9872,  9,  2, 12,  7,  0,  1,  7, 0,  1, 33,
 3,  1,  3,  0,  0,  6,  1,  1,  4, 22, 9947,  2, 45, 13,  3,  1,  3, 4,  2, 15,
 2, 37, 25,  6,  0, 12,  7,  2,  2,  4,  1, 9926, 20,  0,  3,  8, 11, 0,  1,  1,
 1,  1,  0,  0,  0,  2,  0,  0,  0,  5,  8,  4, 9874,  1,  0,  1,  2, 0,  0,  4,
 1,  1,  1,  0,  0,  0,  0,  1,  2,  8,  6,  0,  4, 9946,  0,  2,  1, 3, 28,  0,
13,  5,  2,  1,  1,  8,  3,  2,  5,  1,  2,  2,  1,  1, 9926, 12,  4, 0,  0,  2,
28, 11, 34,  7, 11,  4,  6, 16,  2,  2,  1,  7,  4,  3, 17, 9840, 38, 5,  2,  2,
22,  2, 13,  4,  1,  3,  2,  2,  1, 11,  2,  8,  6,  1,  5, 32, 9871, 0,  2,  9,
 0,  2,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1,  0,  1,  0, 9976, 1,  0,
 1,  0,  3,  0,  3,  0,  1,  0,  4,  1,  1,  0,  0, 21,  0,  1,  1, 2, 9945,  1,
13,  2,  1,  1,  3,  2,  2,  3,  3, 57, 11,  1, 17,  1,  3,  2, 10, 0, 2, 9901};

double	M[8][DIM1(fq)][DIM1(fq)];
double	pam[2][DIM1(fq)][DIM1(fq)];
double	PAM[DIM1(aa)-1][DIM1(aa)-1];
double	minpam;

main(argc, argv)
	int		argc;
	char	*argv[];
{
	int		i, j, k, l, n, index, q, qq, first;
	double	t, sc;

	if (argc < 2) {
		fprintf(stderr, "\nUsage:  pam pam-distance [-s#]\n");
		fprintf(stderr, "where pam-distance is an integer from 2 to 511\n");
		fprintf(stderr, "and # is the optional scale for the log-odds matrix.\n");
		exit(1);
	}

	sc=log(2.)/2.;
	for (i=1;i<argc;++i)
		if (argv[i][0]=='-')
			switch (argv[i][1]) {
					case 'S':
                	case 's':
						sc=atof(argv[i]+2);
                       	break;
                	default  :
                       	break;
        	}
		else
			n = atoi(argv[i]);

	if (n<2) {
		fprintf(stderr,"PAM distance must be greater than 1.\n");
		exit(1);
	}
	if (n>511) {
		fprintf(stderr,"PAM distance must be less than 512.\n");
		exit(1);
	}

	/* Normalize background frequencies */

	for (t=i=0;i<DIM1(fq);++i)
		t+=fq[i];
	for (i=0;i<DIM1(fq);++i)
		fq[i]/=t;

	/* Calculate transition matricies for powers of 2 */

	for (i=k=0;i<DIM3(M);++i)
		for (j=0;j<DIM2(M);++j)
			M[0][j][i] = mdm[k++]/10000.;

	for (i=qq=first=index=1;(index*=2)<=n;++i)
		for (j=0;j<DIM2(M);++j)
			for (k=0;k<DIM3(M);++k)
				for (M[i][j][k]=l=0;l<DIM3(M);++l)
					M[i][j][k]+=M[i-1][j][l]*M[i-1][l][k];

	/* Calculate transition matrix for PAM distance requested */

	for (;n;n-=index) {
		for (;index>n;index/=2)
			--i;
		if (first)
			for (j=first=q=0;j<DIM2(M);++j)
				for (k=0;k<DIM3(M);++k)
					pam[0][j][k]=M[i][j][k];
		else
			for (j=0;j<DIM2(pam);++j)
				for (k=0;k<DIM3(pam);++k)
					for (pam[q][j][k]=l=0;l<DIM3(pam);++l)
						pam[q][j][k]+=pam[qq][j][l]*M[i][l][k];
		q=1-q;
		qq=1-qq;
	}

	/* Calculate symmetric log-odds PAM matrix */

	for (k=0;k<DIM1(fq);++k)
		for (j=0;j<=k;++j) 
			PAM[k][j] = PAM[j][k]
					= log(pam[qq][j][k]*pam[qq][k][j]/(fq[k]*fq[j]))/(sc*2);

	/* Estimate PAM scores for non-standard B,Z and X symbols */

	for (j=0;j<DIM1(fq);++j)
		PAM[j][20] = PAM[20][j]
				= log((pam[qq][j][2]+pam[qq][j][3])
					*(pam[qq][2][j]+pam[qq][3][j])
					/((fq[2]+fq[3])*fq[j]))/(sc*2);

	PAM[20][20]
		= log((pam[qq][2][2]+pam[qq][2][3]+pam[qq][3][2]+pam[qq][3][3])
				/(fq[2]+fq[3]))/sc;

	for (j=0;j<DIM1(fq);++j)
		PAM[j][21] = PAM[21][j]
			= log((pam[qq][j][5]+pam[qq][j][6])*(pam[qq][5][j]+pam[qq][6][j])
				/((fq[5]+fq[6])*fq[j]))/(sc*2);

	PAM[20][21] = PAM[21][20]
			= log((pam[qq][2][5]+pam[qq][2][6]+pam[qq][3][5]+pam[qq][3][6])
			*(pam[qq][5][2]+pam[qq][5][3]+pam[qq][6][2]+pam[qq][6][3])
			/((fq[2]+fq[3])*(fq[5]+fq[6])))/(sc*2);

	PAM[21][21] = log((pam[qq][5][5]+pam[qq][5][6]+pam[qq][6][5]+pam[qq][6][6])
			/(fq[5]+fq[6]))/sc;

	for (j=0;j<DIM1(PAM);++j)
		PAM[22][j] = PAM[j][22] = -1.0;

	/* Print PAM matrix */

	minpam = PAM[0][0];
	for (i=k=0;k<DIM1(fq);++k)
		for (j=0;j<=k;++j) {
			if (PAM[j][k] <= -9.5)
				i=1;
			if (PAM[j][k] < minpam)
				minpam = PAM[j][k];
		}

	if (i) {
		for (j=0;j<DIM1(aa)-1;++j)
			printf("   %c", aa[j]);
		printf("   %c", STOP_CODON_CHR);
		putchar('\n');
		for (j=0;j<DIM1(PAM);++j) {
			for (k=0;k<DIM1(PAM);++k)
				printf("%4ld", fct_nint(PAM[j][k]));
			printf("%4ld", fct_nint(minpam));
			putchar('\n');
		}
		for (j=0; j<DIM1(PAM); ++j)
			printf("%3ld", fct_nint(minpam));
		printf("%3d\n", 1);
	}
	else {
		for (j=0;j<DIM1(aa)-1;++j)
			printf("  %c", aa[j]);
		printf("  %c", STOP_CODON_CHR);
		putchar('\n');
		for (j=0;j<DIM1(PAM);++j) {
			for (k=0;k<DIM1(PAM);++k)
				printf("%3ld", fct_nint(PAM[j][k]));
			printf("%3ld", fct_nint(minpam));
			putchar('\n');
		}
		for (j=0; j<DIM1(PAM); ++j)
			printf("%3ld", fct_nint(minpam));
		printf("%3d\n", 1);
	}
}
