/*************************************************************************
*                                                                        *
*                             COPYRIGHT NOTICE                           *
*                                                                        *
* This software/database is "United States Government Work" under the	 *
* terms of the United States Copyright Act.  It was written as part of	 *
* the author's official duties as a Government employee and thus cannot	 *
* be copyrighted.  This software/database is freely available to the	 *
* public for use without a copyright notice.  Restrictions cannot be	 *
* placed on its present or future use.                                   *
*                                                                        *
* Although all reasonable efforts have been taken to insure the accuracy *
* and reliability of the software and data, the National Library of      *
* Medicine (NLM) and the U.S. Government does not and cannot warrant the *
* performance or results that may be obtained by using this software or  *
* data.  The NLM and the U.S. Government disclaims all warranties as to  *
* performance, merchantability or fitness for any particular purpose.    *
*                                                                        *
* In any work or product derived from this material, proper attribution  *
* of the author as the source of the software or data would be           *
* appreciated.                                                           *
*                                                                        *
*************************************************************************/

#include <ncbi.h>

#ifdef SYSV_SHM_AVAIL

#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>

#define PROJECTID	5	/* the project ID used in calls to ftok() */

/*
 shmalloc -- allocate a shared memory segment; give other processes access.
 */
char *
shmalloc(theKey, mode, size)
	key_t	theKey;
	int		mode;
    size_t	size;
{
	char	*module = "shmalloc";
    extern char *shmat();
    int     shmid;
    struct shmid_ds shmds;
    char    *cp;

    shmid = shmget(theKey, size, IPC_CREAT);
    if (shmid == -1) {
        perror(module);
        return NULL;
    }

    shmctl(shmid, IPC_STAT, &shmds);
    shmds.shm_perm.mode |= mode;
    shmctl(shmid, IPC_SET, &shmds);

	/* Lock the segment in core */
	shmctl(shmid, SHM_LOCK, &shmds);

    cp = shmat(shmid, NULL, 0);
    if (cp == NULL || cp == (char *)-1) {
        perror(module);
        return NULL;
    }
	return cp;
}


/*
	shmexists -- return TRUE if a shared memory segment
	for the specified key exists; otherwise return FALSE.
*/
int
shmexists(theKey, size)
	key_t	theKey;
	unsigned int	size;
{
	int		shmid;

	shmid = shmget(theKey, size, 0);
	if (shmid == -1)
		return FALSE;
	return TRUE;
}


/*
	Create a new shared memory segment and read the entire file into it.
	Returns pointer to the shared memory segment on success; NULL on failure.
*/
char *
shm_loadfile(filename, mode, dolock)
	char	*filename;
	int		mode;
	Boolean	dolock;
{
	struct stat	sbuf;
	key_t	key;
	register int	shmid, semid;
    struct shmid_ds shmds;
	struct sembuf	sops;
	register char	*cp;
	FILE	*fp;

	if (filename == NULL || *filename == NULLB)
		return NULL;

	if (stat(filename, &sbuf) == -1) {
		perror(filename);
		return NULL;
	}

	key = ftok(filename, PROJECTID);
	if (key == -1) {
		perror(filename);
		return NULL;
	}

	/* Create a semaphore corresponding to "key" */
	semid = semget(key, 1, IPC_EXCL|IPC_CREAT|0666);
	if (semid == -1) {
		perror(filename);
		return NULL;
	}

	/* Does a shared memory segment corresponding to "key" already exist? */
	shmid = shmget(key, sbuf.st_size, IPC_EXCL|IPC_CREAT|0666);
	if (shmid == -1) {
		perror(filename);
		/* Destroy the semaphore just created */
		semctl(semid, 1, IPC_RMID, NULL);
		return NULL;
	}
    cp = shmat(shmid, NULL, 0);
	if (cp == NULL || cp == (char *)-1) {
		perror(filename);
		/* Destroy the shared memory segment just created */
		shmctl(shmid, IPC_RMID, NULL);
		/* Destroy the semaphore just created */
		semctl(semid, 1, IPC_RMID, NULL);
		return NULL;
	}

	fp = fopen(filename, "r");
	if (fp == NULL) {
		perror(filename);
		/* Destroy the shared memory segment just created */
		shmctl(shmid, IPC_RMID, NULL);
		/* Destroy the semaphore just created */
		semctl(semid, 1, IPC_RMID, NULL);
		return NULL;
	}

	if (fread(cp, sbuf.st_size, 1, fp) != 1) {
		perror(filename);
		fclose(fp);
		/* Destroy the shared memory segment just created */
		shmctl(shmid, IPC_RMID, NULL);
		/* Destroy the semaphore just created */
		semctl(semid, 1, IPC_RMID, NULL);
		return NULL;
	}
	fclose(fp);

	if (mode != 0666) {
		if (shmctl(shmid, IPC_STAT, &shmds) != 0)
			perror("shmctl1");
    	shmds.shm_perm.mode &= ~0x01ff;
    	shmds.shm_perm.mode |= (mode & 0x01ff);
	    if (shmctl(shmid, IPC_SET, &shmds) != 0)
			perror("shmctl2");
	}

	if (dolock == TRUE)
		shmctl(shmid, SHM_LOCK);

	sops.sem_num = 0;
	sops.sem_op = 16000;
	sops.sem_flg = 0;
	/* Make the semaphore available to 32,000 other processes */
	if (semop(semid, &sops, 1) == -1)
		perror(filename);

	return cp;
}


/*
	Attach to an existing shared memory segment that contains
	a copy of the specified file.

	Segment is attached RDONLY if mode == "r"
	Segment is attached RDWR if mode == "w"
*/
char *
shm_attfile(filename, mode)
	char	*filename;
	char	*mode; /* "r" or "w" */
{
	static char	*module = "shm_attfile";
	struct stat	sbuf;
	key_t	key;
	struct sembuf	sops;
	int		semid, shmid, shmflg;
	char	*cp;

	if (filename == NULL || mode == NULL ||
			(strcmp(mode, "r") && strcmp(mode, "w")))
		return NULL;

	key = ftok(filename, PROJECTID);
	if (key == -1)
		return NULL;

	semid = semget(key, 1, 0666);
	if (semid == -1)
		return NULL;
	sops.sem_num = 0;
	sops.sem_op = -1;
	sops.sem_flg = SEM_UNDO;
	if (semop(semid, &sops, 1) == -1)
		return NULL;

	if (stat(filename, &sbuf) == -1)
		return NULL;

	shmid = shmget(key, sbuf.st_size, 0);
	if (shmid == -1)
		return NULL;

	if (*mode == 'r')
		shmflg = SHM_RDONLY;
	else
		shmflg = 0;
	cp = shmat(shmid, NULL, shmflg);
	if (cp == NULL || cp == (char *)-1) {
		perror(module);
		return NULL;
	}
	return cp;
}


/*
	shm_detach() -- detach the shared memory segment pointed to by 'sp'
*/
int
shm_detach(sp)
	char	*sp;
{
	return shmdt(sp);
}


/*
	shm_dropfile -- remove any existing shared memory segment that keys
	to the specified file.  Warning:  all processes including the local
	process must be assured of being detached from the segment before
	this function is called, or they will crash upon their next attempt
	to reference storage within the segment.
*/
int
shm_dropfile(filename)
	char	*filename;
{
	key_t	key;
	int		semid, shmid;
	struct sembuf	sops;
	int		i;

	key = ftok(filename, PROJECTID);
	if (key == -1)
		return -1;

	semid = semget(key, 1, 0666);
	if (semid == -1)
		return -1;
	sops.sem_num = 0;
	sops.sem_op = -16000;
	sops.sem_flg = 0;
	if (semop(semid, &sops, 1) == -1)
		return -1;

	shmid = shmget(key, 1, 0);
	if (shmid == -1) {
		semctl(semid, IPC_RMID, 0);
		return -1;
	}

	i = shmctl(shmid, IPC_RMID, 0);
	if (semctl(semid, IPC_RMID, 0) == -1 || i == -1)
		return -1;
	return 0;
}

#endif /* SYSV_SHM_AVAIL */
