/**************************************************************************
*                                                                         *
*                             COPYRIGHT NOTICE                            *
*                                                                         *
* This software/database is categorized as "United States Government      *
* Work" under the terms of the United States Copyright Act.  It was       *
* produced as part of the author's official duties as a Government        *
* employee and thus can not be copyrighted.  This software/database is    *
* freely available to the public for use without a copyright notice.      *
* Restrictions can not be placed on its present or future use.            *
*                                                                         *
* Although all reasonable efforts have been taken to ensure the accuracy  *
* and reliability of the software and data, the National Library of       *
* Medicine (NLM) and the U.S. Government do not and can not warrant the   *
* performance or results that may be obtained by using this software,     *
* data, or derivative works thereof.  The NLM and the U.S. Government     *
* disclaim any and all warranties, expressed or implied, as to the        *
* performance, merchantability or fitness for any particular purpose or   *
* use.                                                                    *
*                                                                         *
* In any work or product derived from this material, proper attribution   *
* of the author(s) as the source of the software or data would be         *
* appreciated.                                                            *
*                                                                         *
**************************************************************************/
/*
TBLASTN - Search a nucleotide database for regions that approximately match
regions of a protein query sequence.  The basic command syntax is

 	TBLASTN database query_sequence

	query_sequence names a file containing a protein sequence (where the
	> header line is optional).

The TBLASTN command permits optional additional arguments, such as X=50,
that reset certain internal parameters.  The available parameters are:

	E or e is the _expected_ no. of HSPs to produce (which influences S)
		(default is 25).
	S or s gives the High-scoring Segment Pair (HSP) cutoff score
		(default is calculated from E).
	T or t gives the threshold for a word match (default is calculated,
		but is usually 17 or 18 with the PAM120 matrix).
	X or x gives the value for terminating word extensions (default is 20)
	M or m is the name of a file containing substitution costs matrix
		(default is "PAM120")
	Y or y is the effective length of the querY sequence for statistical
		significance calculations (default uses the actual query length)
	Z or z is the effective databaZe length for statistical significance
		calculations (default uses the actual database length)
	C or c is the identifying number of the genetic code to use.  The
		default value is 0, which represents the standard genetic code.
		See gcodes.h for the list of pre-defined genetic codes.
	P or p is the number of processors to utilize when the program
		is compiled to run on a multi-processor computer.

Thus a typical command is

	tblastn genbank myseq e=0.5 x=25

Literature:

    Karlin & Altschul, "Methods for assessing the statistical significance
        of molecular sequence features by using general scoring schemes,"
        Proc. Natl. Acad. Sci. USA 87 (1990) 2264-2268;
    Altschul, Gish, Miller, Myers and Lipman, "Basic local alignment search
        tool," J. Mol. Biol. 214 (1990) in press.

 (Note:  set tabstops every 4 columns for a better appearance).
*/
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>

#define CKTIME	60	/* Check cpu time expended at this interval (in seconds) */
#define MAXTIME	900	/* Maximum cpu time (in seconds) that can be expended */

#define EXTERN
#define TBLASTN
#include "blast.h"
#include "aabet.h"	/* the alphabet */
#include "ntbet.h"
#include "gcodes.h"

/* default directory containing nucleotide sequence databases */
#ifndef BLASTN_DIR
#define BLASTN_DIR	"/usr2/db"
#endif

#define HEADERLEN_MAX	200	/* maximum length of header line */
#define WORDSIZE_MIN	4	/* minimum length of critical word */
#define WORDSIZE_MAX	7	/* maximum length of critical word */
#define CDBLEN_MAX	15000000	/* max. length of compressed database */

#define DEFAULT_C	0	/* default genetic code is Standard */
#define DEFAULT_E	25.	/* default Expected no. of HSPs */
#define DEFAULT_S	1	/* default cutoff for reporting HSPs */
#define DEFAULT_M	"PAM120"	/* default scoring matrix */
#define DEFAULT_T	15	/* default threshold for neighborhood word hits */
#define DEFAULT_X	20	/* default score drop-off during hit extensions */
#define DEFAULT_E2	0.1	/* default secondary Expect cutoff */
#define DEFAULT_S2	0	/* default secondary cutoff score */
#define DEFAULT_W	4	/* default word size */

		/* Parameters for statistics */
#define DEFAULT_L	0.336
#define DEFAULT_K	0.179
static double	Lambda = DEFAULT_L,
				K = DEFAULT_K,
				H;		/* Relative entropy of the p's and q's */

char
	header[HEADERLEN_MAX],
	dbfname[FILENAME_MAX],
	name[FILENAME_MAX],
	*qname,
	*dbtitle,
#ifdef USE_SHM
	*dbbuf,			/* Memory-resident database */
#endif
	*d_seq,
	*s_seq0, *s_seq, *ss_seq,
	_q_seq[QUERYLEN_MAX+2],	/* null prefix + query sequence + null terminator */
	*q_seq = _q_seq+1,	/* query sequence + null terminator */
	qq_seq[QUERYLEN_MAX+1],	/* binary-encoded query sequence + null terminator */
	*in_seq,		/* input buffer */
	M[FILENAME_MAX] = DEFAULT_M;	/* file containing the substitution scores */

unsigned long
		q_len,		/* length of query sequence */
		totdblen,	/* total no. of residues in the database */
		Meff,		/* effective length of query sequence */
		Neff;		/* effective length of database */

long	line_len,
		c_len,
		clean_count,
		*cseq_beg;

Int4	dbtype, dbformat;

PACK_TYPE	*_c_seq, *c_seq;

int		s_len, d_len;		/* length of current database sequence */

Uint4
	maxlen,		/* max length of a database sequence */
	numwords,	/* number of critical words in hash table */
	numexact;	/* number of exact words that did not score at least T */

long	*score_hist;

SCORE
	score_hist_dim,
	high_score,
	highs_dummy,
	highs_exp,	/* expected high score */
	highs_obs;	/* observed high score */

SCORE
	S = DEFAULT_S,		/* cutoff score for High-scoring Segment Pairs (HSP) */
	S2 = DEFAULT_S2;	/* secondary cutoff score once at least 1 HSP found */

int
	S_set,
	S2_set,
	T = DEFAULT_T,		/* threshold for word hits */
	T_set,
	X = DEFAULT_X,		/* constant for word-hit extensions */
	minusX = -DEFAULT_X, /* -X */
	X_set,
	W = DEFAULT_W,		/* word size */
	W_set,
	C = DEFAULT_C,		/* genetic code */
	C_set,
	E_set,
	E2_set,
	M_set;

double
	E = DEFAULT_E,		/* Expected number of HSPs achieving cutoff score */
	E2 = DEFAULT_E2;		/* Secondary Expect value */

DIAG_TYPE
	*_dlev,		/* flags for which diagonals... */
	*diag_level;	/* have been rigorously examined */

DIAG_PTR
	diag_stack,	/* list of diagonals with matches */
	diag_top;	/* next diagonal with a match */

HSPPtr	firstHSP = NULL;
HSPPtr	*HSPlist;

static int
	maxscore,		/* maximum achievable score */
	current;		/* number of current sequence */

static FILE
	*sfile,			/* file pointer for sequences */
	*cfile;			/* file pointer for compressed sequences */
static MFILE
	*tfile,			/* file pointer for table */
	*hfile;			/* memory file pointer for header lines */

static int
	highs,		/* highest score in the PAM matrix */
	lows;		/* lowest score in the PAM matrix */

static off_t
	nbr_seqs,		/* number of sequences in database */
	*seq_beg, 		/* starts of sequences in file or memory-resident dbase */
	*header_beg;	/* starts of headers */

static char
	*ambiguity_ray;

DFA		*dp;	/* automaton for neighborhood word matching */

int		job_strands = TOP_STRAND|BOTTOM_STRAND;
int		strand;

int		alpha;			/* alpabet size */
int		minchar, maxchar;	/* min. and max. letter values */
static signed char
	/* array of amino acid names, plus null terminator */
	achars[ALPHASIZE_MAX+1],
	/* order[i][j] = j-th best char to sub. for i */
	order[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

char	gcode[64], revgcode[64];	/* genetic codes */
int		frame;	/* current reading frame */

SCORE
	/* max cost of substituting for a character */
	maxcost[ALPHAVAL_MAX+1],
	/* matrx[i][j] = cost of replacing i by j */
	matrx[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];

static void	bld_table PROTO(());
static void	print_seq PROTO((HSPPtr hp));
static void	search_aa PROTO((MPPROTO));
static void	get_pvals PROTO((MPPROTO));
static void	report PROTO((far char *s, far char *q,
							int spos, int qpos, int frame,
							int score, int len,
							double DBP_val, long nseqlen));
static void	initdb PROTO((far char *dbname));
static int	get_seq PROTO((MPPROTO));
static void	dosearch PROTO(());


#ifdef USE_MPROC
MPContext	mpc[PROC_MAX];
int		maxprocs, maxpprocs, numprocs = PROC_DEFAULT;
#else
int		numprocs = 1;
#endif

Uint4	dfanstates, dfastatesize, dfasize, dfaextent;

Cputime_t	t0, t1, t2, t3;

#ifndef USE_MPROC
char	*version = "1.1.7 [6-Dec-90]";
#else
char	*version = "1.1.7MP [6-Dec-90]";
#endif
/*
1.1.5	11-24-90	When at least one HSP is found involving a given database
					sequence, it is scanned again using a lower cutoff score.
					The amount the cutoff score is reduced is the V parameter.
1.1.6	12-4-90		Fixed BLAST algorithm to halt hit extensions properly
1.1.7	12-6-90		Moved many statistical calculations into library functions
*/

main(argc, argv)
	int		argc;
	char	**argv;
{
	char	o, *a;
	int		i;
	Boolean	secondary;

	sys_times(&t0);

	banner(argv[0], version);

	if (argc < 3)
		fatal("Usage:  %s database query_sequence [options]\nValid %s options are E, S, T, E2, S2, X, M, C, W, Y, and Z.", module, module);
	while (--argc > 2) {
		secondary = FALSE;
		if (argv[argc][1] == '=')
			o = argv[argc][0];
		else
			if (argv[argc][0] == '-')
				o = argv[argc][1];
			else
				if (argv[argc][1] == '2' && argv[argc][2] == '=') {
					o = argv[argc][0];
					secondary = TRUE;
				}
				else
					fatal("Argument %d (%s) has improper form", argc+1, argv[argc]);
		a = argv[argc] + 2;
		if (secondary)
			++a;
		o = TOUPPER(o);
		switch (o) {
			case 'P':
#ifdef USE_MPROC
				numprocs = atoi(a);
				if (numprocs <= 0)
					fatal("Bad value for P (the number of processors)");
				if (numprocs > PROC_MAX)
					fprintf(stderr, "\n%s:  Pmax = %d\n", module, PROC_MAX);
#else
				fprintf(stderr, "\n%s:  Parameter P is ignored--\n       this program has not been compiled to use multiprocessing\n",
						module);
#endif
				break;
			case 'M':
				if (str_len(a) >= FILENAME_MAX)
					fatal("PAM file name is too long.");
				str_cpy(M, a);
				if (str_cmp(M, DEFAULT_M) != 0)
					M_set = 1;
				break;
			case 'W':
				if ((W = atoi(a)) <= 0)
					fatal("Bad value for W.");
				if (W > WORDSIZE_MAX)
					fatal("W is too big--must be less than %d.", WORDSIZE_MAX+1);
				W_set = 1;
				break;
			case 'S':
				if (secondary) {
					if ((S2 = atol(a)) == 0)
						fatal("Bad value for S2.");
					S2_set = 1;
					break;
				}
				if ((S = atol(a)) == 0)
					fatal("Bad value for S.");
				S_set = 1;
				break;
			case 'E':
				if (secondary) {
					if (sscanf(a, "%lg", &E2) != 1 || E2 > 1000.0)
						fatal("Bad value for E2.");
					E2_set = 1;
					break;
				}
				if (sscanf(a, "%lg", &E) != 1 || E <= 0.0 || E > 1000.0)
					fatal("Bad value for E.");
				E_set = 1;
				break;
			case 'T':
				if ((T = atoi(a)) < 1)
					fatal("Bad value for T.");
				T_set = 1;
				break;
			case 'X':
				if ((X = atoi(a)) < 1 || X > 100000)
					fatal("Bad value for X.");
				X_set = 1;
				minusX = -X;
				break;
			case 'Y':
				if ((Meff = atoi(a)) == 0)
					fatal("Bad value for Y.");
				break;
			case 'Z':
				if ((Neff = atoi(a)) == 0)
					fatal("Bad value for Z.");
				break;
			case 'C':
				if (isalpha(*a)) {
					display_gcodes();
					exit(0);
				}
				C = atoi(a);
				if (C < 0 || C >= DIM1(gcodes)) {
					printf("Bad value for C\n");
					display_gcodes();
					exit(0);
				}
				break;
			default:
				fatal("Valid options are E, S, T, E2, S2, X, M, C, W, Y, and Z.");
		}
	}

	HSP = (struct hsp **)ckalloc(sizeof(struct hsp)*maxHSP);

	InitGeneticCode(&gcodes[C], gcode, revgcode);

	/* read file of substitution weights */
	get_weights(M, AAID_MAX+1, aa_atob, &alpha, &minchar, &maxchar, achars,
					order, &lows, &highs, maxcost, matrx);

	get_query(argv[2]);
	if (qname != NULL)
		(void) printf("Query:  %s\n", qname);
	
#ifdef USE_MPROC
	(void) prctl(PR_SETEXITSIG, SIGTERM);

	maxprocs = prctl(PR_MAXPROCS);/* max. processes per user */
	maxpprocs = prctl(PR_MAXPPROCS); /* max. parallel processes */
	numprocs = MIN(numprocs, PROC_MAX);
	numprocs = MAX(1, numprocs);
	numprocs = MIN(numprocs, maxprocs);
	numprocs = MIN(numprocs, maxpprocs - PROC_RESERVE);
	numprocs = MAX(1, numprocs);
#endif

#ifdef CKTIME
	/* Poll the cpu time that has been consumed every CKTIME seconds */
	(void) signal(SIGALRM, sigalrm);
	(void) alarm(CKTIME);
#endif

	initdb(argv[1]);

	(void) fflush(stdout);

	/*
	It would be more accurate to call lambdak() for every sequence in
	the database, but then the program would run much slower.
	Another suggestion is to call lambdak() before reporting
	the probability of an HSP containing part of a specific database
	sequence.  For the sake of speed, though, lambdak() is called only once.
	*/
	if (lambdak(lows, highs, q_seq, q_len, &Lambda, &K, &H) != 0)
		fatal("Could not calculate Lambda, K, and H");

	if (!T_set) {
		/* Ad hoc relationship for the PAM120 substitution matrix only */
		T = ceil((3.4 + 2.6*H)/Lambda - 0.5);
#ifdef USE_MPROC
		T -= 2; /* extra sensitivity when the processing power is available */
#endif
	}

	/* Determine the cutoff score, S */
	if (S_set && !E_set)
		E = 0.;
	Neff *= 2; /* searching both strands, so effective length is doubled */
	cutoffs(&S, &E, &highs_exp, Lambda, K, H, Neff, Meff);

	/* Determine the secondary cutoff score, S2, to use */
	if ((E2_set && !S2_set && E2 <= 0.) || (S2_set && S2 > S))
		E2 = 0., S2 = S; /* S2 will be unused */
	else {
		/* Assume an "average" protein's length is 300 a.a. */
		cutoffs(&S2, &E2, &highs_dummy, Lambda, K, H, 300, 300);
		/* Adjust S2 to be in line with S, as necessary */
		S2 = MAX(S2, 1);
		if (S2 > S)
			S2 = S;
		E2 = stoe(S2, Lambda, K, H, 300, 300);
	}

	score_hist_dim = etos(0.01, Lambda, K, H, Neff, Meff) + 1;
	score_hist = (long *)ckalloc0(sizeof(score_hist[0])*(score_hist_dim+1));

	/* The following relation is safe to use, because should the score
	ever reach S during hit extension, the entire diagonal will be
	rigorously searched ignoring X.
	*/
	X = MIN(X, S);
	minusX = -X;


	maxscore = fmaxscore(q_seq, q_len);
	if (maxscore < S) {
		(void) printf("The maximum achievable score with the query sequence is only %d,\n",
				maxscore);
		(void) printf("whereas the cutoff score for reporting HSPs is %d.\n", S);
		(void) printf("You might lower the cutoff, use a different scoring scheme,\n");
		(void) printf("or increase the length of the query sequence.\n");
		exit(1);
	}

	sys_times(&t1);
	bld_table();		/* build table of critical (neighborhood) words */
	if (dfa_close(dp) != 0)
		fatal("error in dfa_close");
	sys_times(&t2);

#ifdef USE_MPROC
	for (i=0; i<numprocs; ++i) {
		hspinit(&mpc[i]);
		initmpc(i, &mpc[i]);
	}
	m_set_procs(numprocs);
	m_fork(dosearch);
	m_kill_procs();
#else
	hspinit();
	dosearch();
#endif

#ifdef CKTIME
	(void) alarm(0);
#endif

	dfanstates = dfa_nstates(dp);
	dfastatesize = dfa_statesize(dp);
	dfasize = dfa_size(dp);
	dfaextent = dfa_extent(dp);
	dfa_destruct(dp);

#ifdef USE_MPROC
	/* Merge the results from the individual processors */
	{
		register int	i, j;
		register struct hsp	*hp;

		for (i=0; i<numprocs; ++i) {
			if (mpc[i].firstHSP != NULL) {
				if (firstHSP == NULL)
					firstHSP = mpc[i].firstHSP;
				else {
					/* Find the tail of the linked list */
					hp = mpc[i].firstHSP;
					while (hp->next != NULL)
						hp = hp->next;
					hp->next = firstHSP;
					firstHSP = mpc[i].firstHSP;
				}
			}
			hasHSP += mpc[i].hasHSP;
			totHSP += mpc[i].totHSP;
			overcnt += mpc[i].overcnt;
			numhits += mpc[i].numhits;
			numfailed += mpc[i].numfailed;
			if (i > 0)
				for (j=0; j <= score_hist_dim; ++j)
					score_hist[j] += mpc[i].score_hist[j];
		}
	}
#endif

	sys_times(&t3);

	print_hist(score_hist_dim, score_hist, S, T, Lambda, K, H,
					Meff, Neff, nbr_seqs);

	print_HSPs();

	print_parms();

	print_stats();

	exit(0);
	/*NOTREACHED*/
}


print_parms()
{
	(void) printf("\n\nParameters:\n");
	(void) printf("  E = %#0.2lg, S = %s, T = %d, X = %d, W = %d, M = %s\n",
		E, lu2s(S), T, X, W, M);
	(void) printf("  E2 = %#0.2lg, S2 = %s\n", E2, (S2 != S ? lu2s(S2) : "not used"));
	(void) printf("  C = %d (%s genetic code)\n", C, gcodes[C].name);
}

print_stats()
{
	(void) printf("\nStatistics:\n");
	(void) printf("  Lambda = %#0.3lg, Lambda/ln2 = %#0.3lg, K = %#0.3lg, H = %#0.3lg\n",
				Lambda, Lambda/log(2.), K, H);
	(void) printf("  Expected/Observed high score = %s/%s\n",
			ld2s(highs_exp), ld2s(highs_obs));
	if (S > highs_exp && highs_obs == 0)
		(void) printf("  NOTE:  the cutoff score is greater than the expected high score\n");
	if (highs_exp > maxscore)
		(void) printf("  NOTE:  expected high score is greater than the max. achievable score %d\n",
				maxscore);
	(void) printf("  Database:  %s\n", dbtitle);
	(void) printf("  # of residues in query:  %s aa", lu2s(q_len));
	if (Meff != q_len)
		(void) printf("  (Y = %s)\n", lu2s(Meff));
	else
		(void) putchar('\n');
	(void) printf("  # of neighborhood words in query:  %s\n", lu2s(numwords));
	(void) printf("  # of exact words scoring below T:  %s\n", lu2s(numexact));
	(void) printf("  # of residues in database:  %s bp", lu2s(totdblen));
	if (Neff != 2*totdblen)
		(void) printf("  (Z = %s)\n", lu2s(Neff/2));
	else
		(void) putchar('\n');
	(void) printf("  # of word hits against database:  %s\n", lu2s((long)numhits));
	(void) printf("  # of failed hit extensions:  %s\n", lu2s((long)numfailed));
	(void) printf("  # of successful extensions:  %s\n", lu2s((long)totHSP+overcnt));
	(void) printf("  # of HSPs reported:  %s\n", lu2s((long)totHSP));
	(void) printf("  # of overlapping HSPs discarded:  %s\n",
				lu2s((long)overcnt));
	(void) printf("  # of sequences in db:  %s\n", lu2s((long)nbr_seqs));
	(void) printf("  # of db sequences with at least one match:  %s\n",
				lu2s((long)hasHSP));

	(void) printf("No. of states in DFA:  %s (%s KB)\n",
			lu2s(dfanstates),
			lu2s(HOWMANY(dfanstates*dfastatesize, KBYTE)));
	(void) printf("Total size of DFA:  %s KB (%s KB)\n",
			lu2s(HOWMANY(dfasize, KBYTE)),
			lu2s(HOWMANY(dfaextent, KBYTE)));
	(void) printf("Time to generate neighborhood:  ");
	sys_showtime(&t1, &t2);
	(void) printf("No. of processors used:  %d\n", numprocs);
	(void) printf("Time to search database:  ");
	sys_showtime(&t2, &t3);
	(void) printf("Total cpu time:  ");
	sys_times(&t3);
	sys_showtime((Cputime_t *)NULL, &t3);
	exit(0);
}



static void
dosearch()
{
	int		frame;
	int		strand;
#ifdef USE_MPROC
	int		id;
	MPDEF

	id = m_get_myid();
	mp = &mpc[id];
	if (id == 0)
		signal(SIGTERM, sigterm);
#endif

	while (get_seq(MP) == 0) {
		MPPTR high_score = 0;
		if (job_strands & TOP_STRAND) {
			strand = TOP_STRAND;
			for (frame = 1; frame <= 3; ++frame) {
				translate(MPC frame);
				search_aa(MP);	/* search the database */
			}
		}
		if (job_strands & BOTTOM_STRAND) {
			strand = BOTTOM_STRAND;
			for (frame = -1; frame >= -3; --frame) {
				btranslate(MPC frame);
				search_aa(MP);	/* search the database */
			}
		}

		if (MPPTR high_score < score_hist_dim)
			++MPPTR score_hist[MPPTR high_score];
		else
			++MPPTR score_hist[score_hist_dim];
		if (MPPTR high_score > highs_obs) {
			M_LOCK;
			if (MPPTR high_score > highs_obs)
				highs_obs = MPPTR high_score;
			M_UNLOCK;
		}

		if (MPPTR curHSP > 0)
			get_pvals(MP);
	}
}


#ifdef USE_MPROC
/* Initialize the processor context pointed to by mp */
initmpc(id, mp)
	int		id;
	MPContextPtr	mp;
{
	int		len;
	register int	i;

	mp->id = id;

	mp->current = mp->nbr_seqs = -1;

	len = maxlen + q_len + 1;
	mp->_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	mp->diag_level = mp->_dlev + q_len;
	mp->diag_top = mp->diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);
	mp->s_seq0 = ckalloc(maxlen+4);
	*mp->s_seq0++ = NULLB;
	mp->s_seq = mp->s_seq0;
	mp->s_seq[maxlen/3 - 1] = mp->s_seq[2*maxlen/3 - 1] = NULLB;
	if (id == 0)
		mp->score_hist = score_hist;
	else
		mp->score_hist
				= (long *)ckalloc0(sizeof(mp->score_hist[0])*(score_hist_dim+1));
}
#endif


display_gcodes()
{
	register int	i;

	printf("The available genetic codes are:\n");
	for (i=0; i<DIM1(gcodes); ++i)
		printf("  %d. %s\n", i, gcodes[i].name);
	printf("Specify a particular code by its number (0 is the default)\n");
}

translate(MPC frame)
	MPDEF
	int		frame;
{
	MPPTR frame = frame;
	if (frame == 1) {
		register int	i;
		register char	ch;
		register char	*sp1, *sp2, *sp3;
		register char	*cp;
		register char	*cend;
		char	*spray[3];
		int		j;

		sp1 = MPPTR s_seq = MPPTR s_seq0;
		sp2 = sp1 + maxlen/3;
		sp3 = sp2 + maxlen/3;

		cp = MPPTR d_seq;
		cend = cp + MPPTR d_len/4;
		ch = *cp;
		i = (ch>>6)&3;
		i <<= 2;
		i |= (ch>>4)&3;
		j = MPPTR d_len%4;

#define XBASE0(p)	i<<=2, i&=0x3f, i|=(ch>>6)&3, *p++ =gcode[i]
#define XBASE1(p)	i<<=2, i&=0x3f, i|=(ch>>4)&3, *p++ =gcode[i]
#define XBASE2(p)	i<<=2, i&=0x3f, i|=(ch>>2)&3, *p++ =gcode[i]
#define XBASE3(p)	i<<=2, i&=0x3f, i|=ch&3, *p++ =gcode[i]

		if (cp < cend) {
			for (;;) {
				XBASE2(sp1);
				XBASE3(sp2);
				if (++cp >= cend)
					goto Break1;
				ch = *cp;
				XBASE0(sp3);
				XBASE1(sp1);
				XBASE2(sp2);
				XBASE3(sp3);
				if (++cp >= cend)
					goto Break2;
				ch = *cp;
				XBASE0(sp1);
				XBASE1(sp2);
				XBASE2(sp3);
				XBASE3(sp1);
				if (++cp >= cend)
					goto Break3;
				ch = *cp;
				XBASE0(sp2);
				XBASE1(sp3);
			}
Break1:
			spray[0] = sp3;
			spray[1] = sp1;
			spray[2] = sp2;
			goto Break;
Break2:
			spray[0] = sp1;
			spray[1] = sp2;
			spray[2] = sp3;
			goto Break;
Break3:
			spray[0] = sp2;
			spray[1] = sp3;
			spray[2] = sp1;
Break:
			if (j == 0)
				goto Terminate;
			ch = *cp;
			XBASE0(spray[0]);
			if (j > 1)
				XBASE1(spray[1]);
			if (j > 2)
				XBASE2(spray[2]);
		}
		else {
			if (j > 2)
				XBASE2(sp1);
			spray[0] = sp1;
			spray[1] = sp2;
			spray[2] = sp3;
		}

Terminate:
		*spray[0] = *spray[1] = *spray[2] = NULLB;
		s_len = MPPTR d_len/3;
		return;
	}
	if (frame == 2) {
		MPPTR s_seq = MPPTR s_seq0 + maxlen/3;
		MPPTR s_len = (MPPTR d_len-1)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
	else { /* frame == 3 */
		MPPTR s_seq = MPPTR s_seq0 + 2*maxlen/3;
		MPPTR s_len = (MPPTR d_len-2)/CODON_LEN;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
}

btranslate(MPC frame)
	MPDEF
	int		frame;
{
	MPPTR frame = frame;
	if (frame == -1) {
		register int	i;
		register unsigned char	ch;
		register char	*sp1, *sp2, *sp3;
		register char	*cp = (char *)MPPTR d_seq;
		register char	*cbeg;
		char	*spray[3];
		int		j;

		sp1 = spray[0] = MPPTR s_seq = MPPTR s_seq0;
		sp2 = spray[1] = MPPTR s_seq + maxlen/3;
		sp3 = spray[2] = MPPTR s_seq + 2*maxlen/3;
		if (MPPTR d_len == 0) {
			*sp1 = *sp2 = *sp3 = NULLB;
			MPPTR s_len = 0;
			return;
		}

		cbeg = MPPTR d_seq;
		cp = MPPTR d_seq + (MPPTR d_len+3)/4 - 1;
		j = MPPTR d_len%4;
		/* skip over pad bits */
		ch = (unsigned char)*cp;
		if (j == 0)
			j = 4;
		else
			ch >>= 2*(4-j);

#define XBASE(p)	i<<=2, i&=0x3c, i|=ch&3, ch>>=2, *p++ =revgcode[i]

		/* Finish off the terminal byte */
		i = ch&3;
		if (j > 1) {
			ch >>= 2;
			i <<= 2;
			i |= ch&3;
			ch >>= 2;
			if (j > 2) {
				XBASE(sp1);
				if (j == 4) {
					XBASE(sp2);
				}
			}
		}

		if (cp > cbeg) {
			ch = (unsigned char)*--cp;
			/* Start work on the penultimate byte */
			switch (j) {
				case 1:
					i <<= 2;
					i |= ch&3;
					ch >>= 2;
					XBASE(sp1);
					spray[0] = sp2;
					spray[1] = sp3;
					spray[2] = sp1;
					break;
				case 2:
					XBASE(sp1);
					XBASE(sp2);
					spray[0] = sp3;
					spray[1] = sp1;
					spray[2] = sp2;
					break;
				case 3:
					XBASE(sp2);
					XBASE(sp3);
					spray[0] = sp1;
					spray[1] = sp2;
					spray[2] = sp3;
					break;
				case 4:
					XBASE(sp3);
					XBASE(sp1);
					spray[0] = sp2;
					spray[1] = sp3;
					spray[2] = sp1;
					break;
			}
			sp1 = spray[0];
			sp2 = spray[1];
			sp3 = spray[2];
			/* Finish the penultimate byte and process all further bytes */
			for (;;) {
				XBASE(sp1);
				XBASE(sp2);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp3);
				XBASE(sp1);
				XBASE(sp2);
				XBASE(sp3);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp1);
				XBASE(sp2);
				XBASE(sp3);
				XBASE(sp1);
				if (--cp < cbeg)
					break;
				ch = (unsigned char)*cp;
				XBASE(sp2);
				XBASE(sp3);
			}
		}

		*sp1 = *sp2 = *sp3 = NULLB;
		MPPTR s_len = MPPTR d_len/3;
		return;
	}
	if (frame == -2) {
		MPPTR s_seq = MPPTR s_seq0 + maxlen/3;
		MPPTR s_len = (MPPTR d_len-1)/3;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
	else { /* frame == -3 */
		MPPTR s_seq = MPPTR s_seq0 + 2*maxlen/3;
		MPPTR s_len = (MPPTR d_len-2)/3;
		if (MPPTR s_len < 0)
			MPPTR s_len = 0;
	}
}

/* get_query - read the query sequence */
get_query(filename)
char *filename;
{
	register FILE	*qp;
	register char *p, *q, ch;
		
	if (str_cmp(filename, "-") == 0)
		qp = fdopen(0, "r"); /* read query sequence from stdin */
	else
		qp = ckopen(filename, "r");
	for (q = q_seq; ; ) {
		p = q;
		if (fgets(p, QUERYLEN_MAX + 1 - (p - q_seq), qp) == NULL) {
			(void) fclose(qp);
			break;
		}
		if (*q == '>')	/* Optional header/title line */
			if (q == q_seq) {
				qname = str_dup(q);
				continue;
			}
			else
				break;
		while ((ch = *p++) != NULLB) {
			if (aa_atob[ch] <= AAID_MAX)
				*q++ = ch;
		}
		if (q - q_seq > QUERYLEN_MAX)
			fatal("query sequence exceeds %d characters",
				QUERYLEN_MAX);
			
	}
	*q = NULLB;
	q_len = q - q_seq;
	if (q_len < W)
		fatal("query sequence is shorter than %d, the Word length.", W);
	if (Meff == 0)
		Meff = q_len;

	if (str_cnt(q_seq, "ACGTacgt")/(double)q_len > 0.85) {
		printf("\nNote:  the query sequence appears to be nucleic acid when protein is expected.\n\n");
		fflush(stdout);
	}

	q = q_seq;
	while ((ch = *q) != NULLB)
		*q++ = aa_atob[ch];
}

/* -----------------------   initialize the genetic code   -------------------*/
InitGeneticCode(gp, gcode, revgcode)
	GeneticCodePtr	gp;
	char	gcode[64];
	char	revgcode[64];
{
	register int	i, j, k, ii, jj, kk;
	register char	*res1, *res2, *res3, *code;
	static char	ntbet[] = "TCAG";

	code = gp->code;

	for (res1 = ntbet; *res1 != NULLB; ++res1) {
		i = nt_atob[*res1];
		ii = res1-ntbet;
		for (res2 = ntbet; *res2 != NULLB; ++res2) {
			j = nt_atob[*res2];
			jj = res2-ntbet;
			for (res3 = ntbet; *res3 != NULLB; ++res3) {
				k = nt_atob[*res3];
				kk = res3-ntbet;
				revgcode[4*4*(3-i)+4*(3-j)+(3-k)]
					= gcode[4*4*i+4*j+k] = aa_atob[code[4*4*ii+4*jj+kk]];
			}
		}
	}
}

/* -----------------------   read the compressed database  -------------------*/
/* UNIX-specific input routine; database must be processed by SETDB */

static void
initdb(dbname)
	char	*dbname;
{
	long	many, len;
	long	cleaned;
	int		once = 0;
	char	*dirname = CWDSTR;
	Int4	i;
#ifdef USE_SHM
	int		dbexists;
#endif

	current = -1;

RETRY:
	if (*dirname != NULLB)
		(void) sprintf(dbfname, "%s%c%s", dirname, DIRDELIMCHR, dbname);
	else
		STRNCPY(dbfname, dbname, sizeof(dbfname));
	sfile = fopen(dbfname, "r");
	if (sfile == NULL) {
		if ((dirname = getenv("BLASTN")) == NULL)
			dirname = BLASTN_DIR;
		if (++once == 1)
			goto RETRY;
		fatal("database not found");
	}

	(void) sprintf(name, "%s.ntb", dbfname);
	tfile = mfil_open(name, "r");
	if (tfile == NULL)
		fatal("Couldn't find database file %s", name);
	(void) mfil_dup(tfile, (char far *)&dbtype, sizeof(dbtype));
	if (dbtype != DB_TYPE_NUC)
		fatal("``%s'' is not a nucleotide sequence database", dbname);
	(void) mfil_dup(tfile, (char far *)&dbformat, sizeof(dbformat));
	if (dbformat != NTFORMAT)
		fatal("Incorrect database format");

	if (mfil_dup(tfile, (char far *)&i, sizeof(i)) != sizeof(i)
			|| (dbtitle = mfil_get(tfile, i)) == NULL)
		fatal("Corrupted database file:  %s", name);
	if (i%4 != 0) {                                      
		i = 4 - (i%4);
		mfil_seek(tfile, i, 1);
	}

	if (mfil_dup(tfile, (char *)&line_len, sizeof(line_len)) != sizeof(line_len)
		|| mfil_dup(tfile, (char *)&nbr_seqs, sizeof(nbr_seqs)) != sizeof(nbr_seqs)
		|| mfil_dup(tfile, (char *)&maxlen, sizeof(maxlen)) != sizeof(maxlen)
		|| mfil_dup(tfile, (char *)&totdblen, sizeof(totdblen)) != sizeof(totdblen)
		|| mfil_dup(tfile, (char *)&c_len, sizeof(c_len)) != sizeof(c_len)
		|| mfil_dup(tfile, (char *)&clean_count, sizeof(clean_count)) != sizeof(clean_count))
		fatal("corrupted database tab file");

	/* Skip over the list of cleaned 8-mers */
	(void) mfil_seek(tfile, clean_count * sizeof(cleaned), 1);

	if (dbtitle[0] == NULLB)
		dbtitle = str_dup(dbname);
	(void) printf("Database:  %s\n", dbtitle);
	(void) printf("           %s sequences; %s total residues.\n",
			ld2s(nbr_seqs), ld2s(totdblen));                       
 
	many = nbr_seqs + 1;
	cseq_beg = (long far *)mfil_get(tfile, sizeof(*cseq_beg)*many);
	seq_beg = (long far *)mfil_get(tfile, sizeof(*seq_beg)*many);
	header_beg = (long far *)mfil_get(tfile, sizeof(*header_beg)*many);
	ambiguity_ray = (char far *)mfil_get(tfile, nbr_seqs/CHAR_BIT + 1);
	if (cseq_beg == NULL || seq_beg == NULL
			|| header_beg == NULL || ambiguity_ray == NULL)
		fatal("Corrupted database tab file");

	(void) sprintf(name, "%s.csq", dbfname);
#ifdef USE_SHM
	/* Attach a shared memory segment for the .csq file (if one exists) */
	if ((c_seq = (PACK_TYPE *)shm_attfile(name, "r")) == NULL)
#endif
	{
		cfile = ckopen(name, "r");
#ifdef USE_MPROC
		if (c_len > CDBLEN_MAX)
			fatal("database is too large; raise CDBLEN_MAX and recompile");
		c_seq = (PACK_TYPE *)ckalloc(c_len);
		dbbuf = (char *)c_seq;
		if (fread((char *)c_seq, c_len, 1, cfile) != 1)
			fatal("Cannot read compressed database sequence file.");
#else
		d_seq = ckalloc(maxlen/(CHAR_BIT/NBPN) + 2 + NSENTINELS);
		(void) fseek(cfile, cseq_beg[0]/(CHAR_BIT/NBPN), 0);
#endif
	}

	c_len -= NSENTINELS; /* Adjust length for the NT_MAGIC_BYTEs at end */

	s_seq0 = ckalloc(maxlen+6);
	/* Store NULL prefix sentinel bytes for 3 reading frames */
	*s_seq0++ = NULLB;
	s_seq = s_seq0;
	s_seq[maxlen/3 - 1] = s_seq[2*maxlen/3 - 1] = NULLB;

	len = maxlen + q_len + 1;
	_dlev = (DIAG_TYPE *)ckalloc0(sizeof(*_dlev)*len);
	diag_level = _dlev + q_len;

	diag_top = diag_stack = (DIAG_PTR)ckalloc(sizeof(*diag_top)*len);

	if (Neff == 0)
		Neff = totdblen;

#ifdef USE_MPROC
	if (numprocs > 1) {
		chunksize = nbr_seqs / (numprocs*50);
		if (numprocs < 4)
			chunksize /= 2;
		chunksize = MAX(chunksize, 1);
	}
	else
		chunksize = nbr_seqs;
#endif
}

static int
get_seq(MP)
	MPDEF
{
	register int	pad;

	if (++MPPTR current >= MPPTR nbr_seqs)
#ifndef USE_MPROC
		return 1;
#else
	{	/* Allocate a chunk of sequences for this process to scan */
		MPPTR current = m_next() * chunksize;
		if (MPPTR current >= nbr_seqs)
			return 1;
		MPPTR nbr_seqs = MPPTR current + chunksize;
		MPPTR nbr_seqs = MIN(MPPTR nbr_seqs, nbr_seqs);
	}
#endif
#if defined(USE_SHM) || defined(USE_MPROC)
	if (c_seq != NULL)
		MPPTR d_seq = (char *)c_seq + (cseq_beg[MPPTR current]>>NBPN);
	else
#endif
	{
		c_len = cseq_beg[current+1]/(CHAR_BIT/NBPN)
					- cseq_beg[current]/(CHAR_BIT/NBPN);
		if (fread(d_seq, c_len, 1, cfile) != 1)
			fatal("error reading compressed database sequence");
	}
	MPPTR d_len = ((cseq_beg[MPPTR current+1]>>NBPN)
					- (cseq_beg[MPPTR current]>>NBPN) - NSENTINELS)<<NBPN;
	pad = cseq_beg[MPPTR current] & BITMASK(NBPN);
	if (pad != 0)
		MPPTR d_len -= (4-pad);
	return 0;
}

/* -----------------------   build table of critical words   -----------------*/

signed char
	wword[WORDSIZE_MAX];	/* candidate for critical word */

SCORE
	best[WORDSIZE_MAX];	/* best[i] = smallest score of wword[0]..wword[i]
				   that can be extended to a critical word */
char	*query;			/* pointer into query sequence */
void	addpat();

static void
bld_table()
{
	register int i;
	register int	score;
	register int	wm1 = W - 1;

	dp = dfa_open(NULL, NULL,
				(LETTER *)AAID_MIN, AAID_MAX-AAID_MIN+1, WORDSIZE_MAX,
				ckalloc, 50*KBYTE, NULL, (DFA_HALTNONALPHAS|DFA_NOXLTAB));
	if (dp == NULL) {
		dfa_perror(dp, module);
		fatal("bld_table");
	}

	for (score=i=0; i<wm1; ++i)
		score += matrx[q_seq[i]][q_seq[i]];

	for (query = q_seq; query[wm1] != NULLB; query++) {
		best[wm1] = T;
		for (i = wm1; i > 0; --i)
			best[i-1] = best[i] - maxcost[query[i]];
		enumerate(0, 0);

		/* Check the score of a perfect match */
		score += matrx[query[wm1]][query[wm1]];
		if (score < T && score > 0) {
			/*
			Add a perfect match, even if its score is less than T.
			This was not done by the program when the JMB paper
			was written, but was a suggested enhancement which should
			yield better (lower) figures for the number of MSPs
			that are overlooked on average.
			*/
			++numexact;
			addpat(query);
		}
		score -= matrx[query[0]][query[0]];
	}
	if (numwords + numexact == 0)
		fatal("no satisfying words in query sequence to compare against database");
}

/* enumerate - try all letters in position i of candidate for a critical word;
	score is for wword[0]..wword[i-1] */
enumerate(i, score)
	register int		i;
	register int	score;
{
	/* possible letters by decreasing weight */
	register signed char	*o = order[query[i]];
	register int	c;
	register int	new_score;

	for (c = 0; c < alpha; ++c) {
		new_score = score + matrx[query[i]][wword[i] = o[c]];
		if (new_score < best[i])
			break;
		if (i < W-1)
			enumerate(i+1, new_score);
		else {
			++numwords;
			addpat(wword);
		}
	}
}

void
addpat(aword)
	register char	*aword;
{
	if (dfa_add(dp, aword, W, (mempint)(query+W)) == NULL)
		fatal("error in dfa_add");
}

/*****************************/
#include "searcha.inc"
/*****************************/

static int
hsp_score_cmp(h1, h2)
	struct hsp	**h1, **h2;
{
	return (*h2)->score - (*h1)->score;
}

static void
get_pvals(MP)
	MPDEF
{
	register int i, j;
	register HSPPtr	hp;
	register double	Q, R, T;

	MPPTR totHSP += MPPTR curHSP;
	++MPPTR hasHSP;

	/* Sort HSPs by score, highest to lowest */
	if (MPPTR curHSP > 1)
		hsort((char far *)MPPTR HSP, MPPTR curHSP, sizeof(MPPTR HSP[0]), hsp_score_cmp);

	for (i = 0; i <  MPPTR curHSP; ++i) {
		hp =  MPPTR HSP[i];
		Q = stoe(hp->score, Lambda, K, H, MPPTR s_len, Meff);
		T = R = 1.0;
		for (j = 1; j <= i; ++j) {
			R *= (Q/j);
			T += R;
		}
		if (Q > 0.01)	/* Is limit approximation applicable? */
			hp->rawp.P_val = 1.0 - exp(-Q)*T;
		else
			hp->rawp.P_val = R * (Q/j);
	}

	/* add HSPs for this sequence to the cumulative list */
	MPPTR HSP[MPPTR curHSP-1]->next = MPPTR firstHSP;
	for (i = MPPTR curHSP-2; i >= 0; --i)
		MPPTR HSP[i]->next = MPPTR HSP[i+1];
	MPPTR firstHSP = MPPTR HSP[0];

	MPPTR curHSP = 0;
}

print_HSPs()
{
	register HSPPtr kp1, kp2;
	register int i;
	register double	R, minp, minp2;
	int		max_match;
	double	xslen;

	if (hasHSP < maxHSP)
		HSPlist = HSP;
	else
		HSPlist = (HSPPtr *)ckalloc(sizeof(HSPPtr)*hasHSP);

	for (i = 0; i < hasHSP; ++i) {
		HSPlist[i] = firstHSP;
		for (kp1 = firstHSP; kp1 != NULL; kp1 = kp2)
			if ((kp2 = kp1->next) == NULL ||
			    kp1->seq_nbr != kp2->seq_nbr)
				break;
		kp1->next = NULL;
		firstHSP = kp2;
	}

	/*
		For each sequence, calculate the P_val adjusted for the total
		length of the database (DBP_val); then find the smallest DBP_val
		of its HSPs
	 */
	for (i = 0; i < hasHSP; ++i) {
		kp1 = HSPlist[i];
		xslen = seq_beg[kp1->seq_nbr+1] - seq_beg[kp1->seq_nbr];
		R = Neff/xslen;
		minp = R * kp1->rawp.P_val;
		if (minp > 0.01) /* Is limit approximation applicable? */
			minp = (double)1.0 - exp(-minp);
		kp1->DBP_val = minp;
		for (kp2 = kp1->next; kp2 != NULL; kp2 = kp2->next) {
			minp2 = R * kp2->rawp.P_val;
			if (minp2 > 0.01)	/* Is limit approximation applicable */
				minp2 = (double)1.0 - exp(-minp2);
			kp2->DBP_val = minp2;
			if (minp2 < minp)
				minp = minp2;
		}
		kp1->rawp.smallest_P = minp;
	}

	(void) hsort((char *)HSPlist, hasHSP, sizeof(HSPlist[0]), smaller);

	max_match = q_len;
	ss_seq = ckalloc((size_t)(max_match+1));
	max_match *= CODON_LEN;
	in_seq = ckalloc((size_t)(max_match + max_match/line_len + 2));

	/* print the matches for each sequence */
	for (i = 0; i < hasHSP; ++i)
		print_seq(HSPlist[i]);
}

static void
print_seq(hp)
	HSPPtr	hp;
{
	long	nseqlen;
	int		current, len, n, qpos, spos, npos, nlen;
	int		in_len, line;
	register char	*p, *q;
	HSPPtr	kp;
	register char	*cp;

	current = hp->seq_nbr;

	/* get header line */
	if (hfile == NULL) {
		(void) sprintf(name, "%s.nhd", dbfname);
		hfile = mfil_open(name, "r");
		if (hfile == NULL)
			fatal("Couldn't find database file %s", name);
	}
	(void) mfil_seek(hfile, header_beg[current], 0);
	len = header_beg[current+1] - header_beg[current];
	if (len > sizeof(header)-1)
		len = sizeof(header)-1;
	(void) mfil_dup(hfile, header, len);
	header[len] = NULLB;
	(void) printf("\n\n%s\n", header);
	fflush(stdout);

	nseqlen = ((cseq_beg[current+1]>>NBPN) - (cseq_beg[current]>>NBPN) - NSENTINELS) << NBPN;
	n = cseq_beg[current] & BITMASK(NBPN);
	if (n != 0)
		nseqlen -= (4-n);

	(void) printf("\tLength = %s\n", ld2s((long)nseqlen));

	for (kp = hp; kp != NULL; kp = kp->next) {
		qpos = kp->q_pos;
		spos = kp->s_pos;
		len = kp->len;
		frame = kp->frame;

		nlen = len * CODON_LEN;
		if (frame > 0)
			npos = spos*CODON_LEN + frame - 1;
		else
			npos = nseqlen - spos*CODON_LEN + frame - nlen + 1;

		line = npos/line_len;
		/* n = no. of nucleotides remaining to be read */
		n = nseqlen - npos;
		if (n < nlen)
			printf("woops! (%d)\n", nlen), nlen = n; /* woops! */
		in_len = nlen + nlen/line_len + (npos%line_len == 0 ? 0 : 1);
		(void) fseek(sfile, seq_beg[current] + npos + line, 0);
		in_len = fread(in_seq, 1, in_len, sfile);
		in_seq[in_len] = NULLB;
		s_seq = s_seq0;
		for (p = s_seq, q = in_seq; p < s_seq+nlen; ++q) {
			if (*q != '\n') {
				if ((*p++ = *q) == NULLB)
					fatal("cannot remove null bytes (%d vs. %d)", nlen, p-s_seq);
			}
		}
		s_seq[nlen] = NULLB;
		if (frame > 0)
			trans2(s_seq, ss_seq, len);
		else
			rtrans2(s_seq, ss_seq, len);

		/* Extract the aligned subsequence from q_seq */

		mem_cpy(qq_seq, q_seq + qpos, len);
		qq_seq[len] = NULLB;
		/* Convert to printable ASCII */
		cp = qq_seq;
		while (*cp != NULLB) {
			*cp = aa_btoa[*cp];
			++cp;
		}

		report(ss_seq, qq_seq, spos, qpos, frame, kp->score, len, kp->DBP_val, nseqlen);
		(void) fflush(stdout);
	}
}

trans2(ntq, aaq, n)
	register char	*ntq;	/* input nucleotide sequence */
	register char	*aaq;	/* output amino acid sequence */
	int		n;			/* length of translated protein */
{
	register char	*aaqmax = aaq+n;
	register int	i, j;

	while (aaq < aaqmax) {
		i = nt_atob[*ntq++];
		if (i > 3) {
			ntq += 2;
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		j = nt_atob[*ntq++];
		if (j > 3) {
			++ntq;
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		i = (i<<2)|j;
		j = nt_atob[*ntq++];
		if (j > 3) {
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		*aaq++ = aa_btoa[gcode[(i<<2)|j]];
	}
	*aaq = NULLB;
}

rtrans2(ntq, aaq, n)
	register char	*ntq, *aaq;
	int		n;
{
	register char	*aaqmax = aaq+n;
	register int	i, j;

	ntq += CODON_LEN*n;
	while (aaq < aaqmax) {
		i = nt_atob[*--ntq];
		if (i > 3) {
			ntq -= 2;
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		j = nt_atob[*--ntq];
		if (j > 3) {
			--ntq;
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		i = (i<<2)|j;
		j = nt_atob[*--ntq];
		if (j > 3) {
			*aaq++ = UNKNOWN_AA_CHR;
			continue;
		}
		*aaq++ = aa_btoa[revgcode[(i<<2)|j]];
	}
	*aaq = NULLB;
}

/* report - print a match */
static void
report(s, q, spos, qpos, frame, score, len, DBP_val, seqlen)
	char	*s, *q;
	int		spos, qpos, frame, score, len;
	double	DBP_val;
	long	seqlen;
{
	double	Expect;
	int		numdone = 0;
	int		mlen;

	Expect = stoe(score, Lambda, K, H, Neff, Meff);

	(void) printf("\n  Score = %d, Expect = %#0.2lg, Poisson P = %#0.2lg, frame = %s%d, len = %s\n",
			score, Expect, DBP_val, (frame>0?"+":""), frame, ld2s(len));

	while (len > 0) {
		mlen = MIN(len, 60);
		(void) printf("\nQuery:%6d %.60s %d\n",
				qpos + numdone + 1,
				q,
				qpos + numdone + mlen);
		printmid(q, s, len);
		(void) printf("Sbjct:%6ld %.60s %ld\n",
				ac_to_nt1(spos + numdone, frame, seqlen),
				s,
				ac_to_nt3(spos + numdone + mlen - 1, frame, seqlen));
		q += 60;
		s += 60;
		len -= 60;
		numdone += 60;
	}
}

printmid(s1, s2, len)
	register char	*s1, *s2;
	register int	len;
{
	register char	c1, c2;

	(void) printf("%13s", "");
	if (len > 60)
		len = 60;
	while (len-- > 0) {
		if ((c1 = *s1++) == (c2 = *s2++) && c1 != UNKNOWN_AA_CHR) {
			(void) putchar(c1);
			continue;
		}
		if (matrx[aa_atob[c1]][aa_atob[c2]] > 0)
			(void) putchar('+');
		else
			(void) putchar(' ');
	}
	(void) putchar('\n');
}

/*
	ac_to_nt1

	Convert an amino acid coordinate number to the nucleotide number
	of the first nucleotide in the codon.
*/
long
ac_to_nt1(pos, frame, len)
	long	pos; /* amino acid offset from start of a.a. sequence buffer */
	int		frame; /* reading frame of protein sequence */
	long	len; /* length of nucleotide sequence */
{

	if (frame > 0)
		return pos*CODON_LEN + frame;

	return len - pos*CODON_LEN + frame + 1;
}

/*
	ac_to_nt3

	Convert an amino acid coordinate number to the nucleotide number
	of the third nucleotide in the codon.
*/
long
ac_to_nt3(pos, frame, len)
	long	pos;	/* amino acid offset from start of a.a. sequence buffer */
	int		frame;	/* reading frame of amino acid sequence */
	long	len;	/* length of nucleotide sequence */
{
	if (frame > 0)
		return pos*CODON_LEN + frame + 2;
	return len - pos*CODON_LEN + frame - 1;
}

/* ----------------------------  parameter calculations  ---------------------*/
/*
	Amino acid residue frequencies used by S. Altschul.
*/
double fq[21] = {0.,
		.081, .057, .045, .054, .015, .039, .061, .068, .022, .057,
		.093, .056, .025, .040, .049, .068, .058, .013, .032, .067 };

/* Version of lambdak() specifically for a.a. alphabet used by BLASTP */
int
lambdak(lows, highs, q_seq, q_len, Lambda, K, H)
	Int4	lows, highs;
	Uint4	q_len;
	register char	*q_seq;
	double	*Lambda, *K, *H;
{
	double	fq1[AAID_MAX+1];
	Uint4	iq1[AAID_MAX+1];
	double	*prob;
	Uint4	range;
	register Uint4	ui;
	register int	i, j;

	if (lows >= 0 || highs <= 0)
		return 1;

	/* Tally the frequency of occurrence of each letter in the query */
	for (i=0; i<DIM1(iq1); ++i)
		iq1[i]=0;
	for (ui=0; ui<q_len; ++ui)
		++iq1[q_seq[ui]];

	/* Convert frequencies to probabilities */
	for (i=AAID_MIN; i<=AAID_MAX; ++i)
		fq1[i] = (double)iq1[i] / q_len;

	/* Calculate probability of each score */
	range = highs - lows;
	prob = (double *)ckalloc(sizeof(*prob)*(range+1));
	for (ui=0; ui<=range; ++ui)
		prob[ui] = 0.;

	for (i=minchar; i<=maxchar; ++i)
		for (j=AAID_MIN; j<DIM1(fq); ++j)
			prob[matrx[i][j]-lows] += fq1[i]*fq[j];

	i = karlin(lows, highs, prob, Lambda, K, H);
	free((char *)prob);
	if (i != 0) {
		*K = *Lambda = -1.;
		return i;
	}
	return 0;
}


/* fmaxscore(seq, len) -- returns the maximum achievable score */
/* Note:  effect of parameter X is not taken into consideration here */
int
fmaxscore(seq, len)
	register char	*seq;
	register Uint4	len;
{
	register char	*maxseq;
	register int	sum, maxsum;

	maxseq = seq+len;
	sum = maxsum = 0;
	while (seq < maxseq) {
		sum += maxcost[*seq++];
		if (sum < 0)
			sum = 0;
		else
			if (sum > maxsum)
				maxsum = sum;
	}
	return maxsum;
}
