/* Utility functions for the BLAST program family */
#include <ncbi.h>
#include <ncbilib.h>
#include <signal.h>
#include "blast.h"

/* stoe() -- given a score, return the associated Expect value */
double
stoe(S, Lambda, K, H, dblen, qlen)
	SCORE	S;
	double	Lambda, K, H; /* parameters for Karlin statistics */
	unsigned long	dblen;	/* length of database */
	unsigned long	qlen;	/* length of query sequence */
{
	double	C, E, xqlen = qlen;

	C = K * xqlen * dblen; /* Frequently used constant */

	E = C * exp(-Lambda * S);
#ifdef REFINED_STATS
	E *= MAX(xqlen - Lambda*S/H, 1.) / xqlen;
#endif /* REFINED_STATS */

	return E;
}

/* etos() -- given an Expect value, return the associated cutoff score */
SCORE
etos(E, Lambda, K, H, dblen, qlen)
	double	E, Lambda, K, H; /* parameters for Karlin statistics */
	unsigned long	dblen;	/* length of database */
	unsigned long	qlen;	/* length of query sequence */
{
	SCORE	S, es;
	double	C, f, x, prod1, prod2, xqlen = qlen;
	long	i;
	long	lo, hi;

	C = K * xqlen * dblen; /* Frequently used constant */

	S = ceil( log(C/E) / Lambda );

#ifdef REFINED_STATS
	/* Binary search to find appropriate cutoff */
	hi = S;
	lo = S = 1;
	while (lo < hi) {
		S = lo + (hi - lo)/2;
		x = C * exp(-Lambda * S);
		f = MAX(xqlen - Lambda*S/H, 1.) / xqlen;
		prod1 = x*f;
		x = C * exp(-Lambda * (S+1));
		f = MAX(xqlen - Lambda*(S+1)/H, 1.) / xqlen;
		prod2 = x*f;
		if (prod1 <= E) {
			if (prod1 == E)
				break;
			hi = S;
		}
		else {
			lo = S+1;
			if (prod2 <= E) {
				S = lo;
				break;
			}
		}
	}
#endif /* REFINED_STATS */

	return S;
}


/* shigh() --  return the expected high score */

SCORE
ehighs(Lambda, K, H, dblen, qlen)
	double	Lambda, K, H; /* parameters for Karlin statistics */
	unsigned long	dblen;	/* length of database */
	unsigned long	qlen;	/* length of query sequence */
{
	SCORE	highs, s;
	double	C, f, xqlen = qlen;

	C = K * xqlen * dblen; /* Frequently used constant */

	/* Calculate the expected high score */
	highs = log(C)/Lambda;
#ifdef REFINED_STATS
	do {
		f = MAX(xqlen - Lambda*highs/H, 1.) / xqlen;
		s = log(C * f)/Lambda;
	} while (s < highs && (highs = s) > 0);
#endif

	return highs;
}

/* ckoverlap - remove overlapping HSPs */
/*
   ckoverlap discards HSPs that overlap.  An HSP is said to "contain"
another, if at least one of the segments of the segment pair spans the
entirety of the corresponding segment in the other HSP.

The overlap algorithm ensures that two HSPs will never appear together
in the output if one of them "contains" the other.  An old HSP may be
replaced by a new HSP, if either HSP contains the other AND:  (a) the
new HSP has a higher score; or (b) the new HSP has the same score but
is shorter.  Conversely, the new HSP will be discarded, if either HSP
contains the other AND:  (a) the old HSP has a higher score; or (b)
the old HSP has the same score but is shorter.

In any case, a newly found HSP is never used to discard an old HSP until
it is known that the new HSP should not itself be discarded due to some
other old HSP.

At present, the pruning of HSPs occurs as the HSPs are found, rather than
accumulating all HSPs until the end of the search and trying to optimize
the process in some other manner.
*/
void
ckoverlap(MPC hp)
	MPDEF
	register struct hsp far	*hp;
{
	register struct hsp far	*hp2;
	register int	i;
	register Int4	score;
	register struct hsp far	**hpp;
	register int	firstov = -1, ovcnt;

	score = hp->score;
	for (i=0; i<MPPTR curHSP; ++i) {
		hp2 = MPPTR HSP[i];
		if ((*ovlap)(hp, hp2) == TRUE) {
			if (score > hp2->score) {
				if (firstov < 0)
					firstov = i;
				continue;
			}
			++MPPTR overcnt;
			return;
		}
	}
	if (firstov < 0) {
		/* Save the new HSP and return */
		(*hspsave)(MPC hp);
		return;
	}

	/* Save the new HSP in existing storage, overwriting an overlapping HSP */
	mem_cpy((char far *)(MPPTR HSP[firstov]), (char far *)hp, hspsize);
	hpp = &MPPTR HSP[i = firstov+1];
	ovcnt = 0;

	/* Discard any other HSPs that overlap with the new HSP */
	for (; i<MPPTR curHSP; ++i) {
		hp2 = MPPTR HSP[i];
		if ((*ovlap)(hp, hp2) == TRUE) {
			++ovcnt;
			free((char far *)hp2);
			continue;
		}
		*hpp++ = hp2;
	}
	MPPTR curHSP -= ovcnt;
	MPPTR overcnt += ovcnt+1;
}

hspinit(MP)
	MPDEF
{
	MPPTR HSP = (struct hsp **)ckalloc(sizeof(HSPPtr)*maxHSP);
}

/* ghspsave -- general routine for saving an HSP */
void
ghspsave(MPC hp)
	MPDEF
	HSPPtr	hp;
{
	if (MPPTR curHSP >= maxHSP)
		fatal("Too many matches; raise HSP_MAX and recompile,\nuse a shorter query sequence, or raise the cutoff score");
	MPPTR HSP[MPPTR curHSP] = (struct hsp *)ckalloc(hspsize);
	mem_cpy((char far *)MPPTR HSP[MPPTR curHSP++], (char far *)hp, hspsize);
}

/* smaller -- comparison function used to sort HSPs by probability */
int
smaller(pki, pkj)
	HSPPtr	*pki, *pkj;
{
	register HSPPtr	ki = *pki, kj = *pkj;
	register long	idiff;
	register double	xdiff;

	if ( (ki->rawp.smallest_P <= 0.0 && kj->rawp.smallest_P <= 0.0) ||
			(ki->rawp.smallest_P >= 1.0 && kj->rawp.smallest_P >= 1.0) ) {
		idiff = kj->score - ki->score;
		if (idiff != 0)
			return idiff;
	}
	else {
		xdiff = ki->rawp.smallest_P - kj->rawp.smallest_P;
		if (xdiff != 0.)
			return (xdiff > 0.) ? 1 : -1;
	}
	return ki->seq_nbr - kj->seq_nbr;
}

/*
	banner -- display a program banner consisting of program name
	and version on stdout.
*/
banner(fullname, version)
	register char	*fullname;
	char	*version;
{
	register char	*cp;

	module = cp = basename(fullname);
	while (isalpha(*cp) || isdigit(*cp))
		putchar(TOUPPER(*cp)), ++cp;

	printf(" %s\n\n", version);
}


/* fatal - print message and die */
#ifdef STDARG_AVAIL
fatal(char *format, ...)
#else
fatal(format, va_alist)
	char	*format;
	va_dcl
#endif
{
	va_list	args;

	(void) fflush(stdout);

	if (*module != NULLB)
		(void) fprintf(stderr, "\n%s:  ", module);

#ifdef STDARG_AVAIL
	va_start(args, format);
#else
	va_start(args);
#endif
	(void) vfprintf(stderr, format, (va_list)args);
	va_end(args);
	(void) fputc('\n', stderr);
	(void) fflush(stderr);
	(void) signal(SIGTERM, SIG_IGN);
	(void) kill(0, SIGTERM);
	exit(1);
	/*NOTREACHED*/
}

	
/* ckopen - open file; check for success */
FILE *
ckopen(name, mode)
	char *name, *mode;
{
	FILE *fp;

	if (strcmp(name, "-") == 0) {
		if (*mode == 'r')
			fp = fdopen(0, mode);
		else
			fp = fdopen(1, mode);
	}
	else
		fp = fopen(name, mode);
	if (fp == NULL)
		fatal("Cannot open file \"%s\".", name);
	return fp;
}

/* ckalloc - allocate space; check for success */
char far *
ckalloc(amount)
	size_t	amount;
{
	char far *p;
#ifdef MAXBYTES
	static unsigned	long totbytes = 0;

	M_LOCK;
	if ((totbytes += amount) > MAXBYTES)
		fatal("Memory usage exceeded limit (%s bytes)", lu2s(MAXBYTES));
	M_UNLOCK;
#endif

	if (amount == 0)
		return NULL;

	if (amount > MAXALLOC)
		fatal("ckalloc: Memory request exceeded MAXALLOC(%ld)", (long)MAXALLOC);

	if ((p = malloc(amount)) == NULL)
		fatal("ckalloc: Ran out of memory", amount);

	return p;
}


/* ckalloc0 - allocate space and initialize to zero */
char far *
ckalloc0(amount)
	size_t	amount;
{
	register char far	*p;

	if (amount == 0)
		return NULL;

	p = ckalloc(amount);
	mem_zero(p, amount);
	return p;
}

/* ckrealloc - re-allocate storage */
char far *
ckrealloc(p, newsize)
	char far	*p;
	size_t	newsize;
{
	if (p == NULL)
		return ckalloc(newsize);
	p = realloc(p, newsize);
	if (p == NULL)
		fatal("Ran out of memory during realloc().");
	return p;
}

Signal_t
sigterm(sig)
	int		sig;
{
	signal(SIGTERM, SIG_IGN);
	kill(0, SIGTERM);
	fprintf(stderr, "%s:  killed by signal #%d\n", module, sig);
	_exit(1);
}


#if defined(OS_UNIX) || defined(OS_VMS)
Signal_t
sigalrm()
{
	Cputime_t	t;
	long	t_user, t_system;

	if (maxtime == 0)
		return;

	(void) signal(SIGALRM, SIG_IGN);

	sys_times(&t);

#ifdef OS_UNIX
	t_user = t.user;
	t_system = t.system;
#endif
#ifdef OS_VMS
	t_user = t.proc_user_time;
	t_system = t.proc_system_time;
#endif

	if (t_user + t_system > maxtime*CLK_TCK)
		fatal("cpu time limit (%d seconds) exceeded.", maxtime);
	(void) signal(SIGALRM, sigalrm);
	(void) alarm(cktime);
}
#endif /* OS_UNIX || OS_VMS */
