#include <ncbi.h>
#include "blast.h"

/* -----------------   read file of substitution weights   ----------------*/

/*  get_weights() reads a file giving weights for amino acid substitutions.
The proper form of the file is illustrated by the following "PAM120" matrix.
The first row gives amino acid abbreviations, which must be single capital
letters.  Weights should be small integers (e.g., between -128 and 127).

  A  R  N  B  D  C  Q  Z  E  G  H  I  L  K  M  F  P  S  T  W  Y  V  X  
  3 -3  0  0  0 -3 -1 -1  0  1 -3 -1 -3 -2 -2 -4  1  1  1 -7 -4  0 -1
 -3  6 -1 -2 -3 -4  1 -1 -3 -4  1 -2 -4  2 -1 -4 -1 -1 -2  1 -6 -3 -1
  0 -1  4  3  2 -5  0  0  1  0  2 -2 -4  1 -3 -4 -2  1  0 -5 -2 -3 -1
  0 -2  3  4  3 -6  0  1  2  0  1 -3 -5  0 -4 -6 -2  0 -1 -7 -4 -3 -1
  0 -3  2  3  5 -7  1  2  3  0  0 -3 -5 -1 -4 -7 -2  0 -1 -8 -5 -3 -1
 -3 -4 -5 -6 -7  9 -7 -7 -7 -5 -4 -3 -7 -7 -6 -6 -3 -1 -3 -8 -1 -2 -1
 -1  1  0  0  1 -7  6  4  2 -3  3 -3 -2  0 -1 -6  0 -2 -2 -6 -5 -3 -1
 -1 -1  0  1  2 -7  4  5  3 -2  1 -3 -3 -1 -3 -6 -1 -2 -2 -7 -5 -3 -1
  0 -3  1  2  3 -7  2  3  5 -1 -1 -3 -4 -1 -4 -6 -1 -1 -2 -8 -4 -3 -1
  1 -4  0  0  0 -5 -3 -2 -1  5 -4 -4 -5 -3 -4 -5 -2  1 -1 -8 -6 -2 -1
 -3  1  2  1  0 -4  3  1 -1 -4  7 -4 -3 -2 -4 -2 -1 -2 -3 -5 -1 -3 -1
 -1 -2 -2 -3 -3 -3 -3 -3 -3 -4 -4  6  1 -2  1  0 -3 -2  0 -7 -2  3 -1
 -3 -4 -4 -5 -5 -7 -2 -3 -4 -5 -3  1  5 -4  3  0 -3 -4 -3 -5 -3  1 -1
 -2  2  1  0 -1 -7  0 -1 -1 -3 -2 -2 -4  5  0 -6 -2 -1 -1 -5 -6 -4 -1
 -2 -1 -3 -4 -4 -6 -1 -3 -4 -4 -4  1  3  0  8 -1 -3 -2 -1 -7 -4  1 -1
 -4 -4 -4 -6 -7 -6 -6 -6 -6 -5 -2  0  0 -6 -1  8 -5 -3 -4 -1  4 -3 -1
  1 -1 -2 -2 -2 -3  0 -1 -1 -2 -1 -3 -3 -2 -3 -5  6  1 -1 -7 -6 -2 -1
  1 -1  1  0  0 -1 -2 -2 -1  1 -2 -2 -4 -1 -2 -3  1  3  2 -2 -3 -2 -1
  1 -2  0 -1 -1 -3 -2 -2 -2 -1 -3  0 -3 -1 -1 -4 -1  2  4 -6 -3  0 -1
 -7  1 -5 -7 -8 -8 -6 -7 -8 -8 -5 -7 -5 -5 -7 -1 -7 -2 -6 12 -1 -8 -1
 -4 -6 -2 -4 -5 -1 -5 -5 -4 -6 -1 -2 -3 -6 -4  4 -6 -3 -3 -1  8 -3 -1
  0 -3 -3 -3 -3 -2 -3 -3 -3 -2 -3  3  1 -4  1 -3 -2 -2  0 -8 -3  5 -1
 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1 -1

*/


int
get_weights(mname, lettermax, xltab, asize, minchar, maxchar,
			achars, order, lows, highs, maxcost, matrx)
	char	*mname;	/* Name of file containing the substitution matrix */
	int		lettermax;	/* Max binary letter value used to screen input file */
	int		xltab[128];	/* ASCII to binary translation table */
	int		*asize;	/* No. of letters in the alphabet */
	int		*minchar, *maxchar;	/* Min. and max. letter values */
	signed char
		/* array of amino acid names, plus null terminator */
		(*achars)[ALPHASIZE_MAX+1],
		/* order[i][j] = j-th best char to sub. for i */
		(*order)[ALPHAVAL_MAX+1][ALPHASIZE_MAX];

	SCORE
		*lows,			/* lowest substitution value */
		*highs,			/* highest substitution value */
		/* max cost of substituting for a character */
		(*maxcost)[ALPHAVAL_MAX+1],
		/* matrx[i][j] = cost of replacing i by j */
		(*matrx)[ALPHAVAL_MAX+1][ALPHAVAL_MAX+1];
{
	FILE	*fp;
	int		i, j;
	int		score;
	SCORE	*m;
	signed char	*o, *p, *q, t;

	fp = ckopen(mname, "r");

	/* read residue names */
	*minchar = INT_MAX;
	*maxchar = INT_MIN;
	p = (*achars);
	while ((i = fgetc(fp)) != '\n' && i != 0) {
		if ((j = xltab[i]) < lettermax) {
			if (p - (*achars) >= DIM1(*achars)-1)
				fatal("Too many amino acid names in %s.", mname);
			if (j > *maxchar)
				*maxchar = j;
			if (j < *minchar)
				*minchar = j;
			*p++ = j;
		}
		else {
			if (isspace(i))
				continue;
			if (isprint(i))
				fatal("%s:  Bad a.a. residue: ``%c.''", mname, i);
			else
				fatal("%s:  Bad a.a. residue: #%d.", mname, i);
		}
	}

	*asize = p - (*achars);
	if (*asize >= DIM1(*matrx))
		fatal("too many amino acid names in %s.", mname);

	/* For illegal characters, set weight to lowest possible value. */
	for (i=0; i<DIM1(*maxcost); ++i)
		(*maxcost)[i] = SCORE_MIN/2;
	for (i=0; i<DIM1(*matrx); ++i)
		for (j=0; j<DIM2(*matrx); ++j)
			(*matrx)[i][j] = SCORE_MIN/2;

	/* read substitution weights for each residue */
	*lows = INT_MAX;
	*highs = INT_MIN;
	for (i = 0; i < *asize; ++i) {
		m = (*matrx)[(*achars)[i]];
		for (j = 0; j < *asize; ++j) {
			if (fscanf(fp, " %d", &score) != 1)
				fatal("Conversion error reading PAM file ``%s'', on line %d.",
							mname, i+2);
			m[(*achars)[j]] = score;
			if (*lows > score)
				*lows = score;
			if (*highs < score)
				*highs = score;
		}
		o = (*order)[(*achars)[i]];
		for (j = 0; j < *asize; ++j)
			o[j] = (*achars)[j];
		/* bubble-sort residue names by cost of substitution for i-th residue */
		for (q = o + (*asize-1); q > o; --q)
			for (p = o; p < q; ++p)
				if (m[*p] < m[*q]) {
					t = *p;
					*p = *q;
					*q = t;
				}
		(*maxcost)[(*achars)[i]] = m[*o];
	}
	(void) fclose(fp);
}
